/*
    alu.c
    This implements the virtual machine's ALU.
*/

/*
    Replace the subword of width 'wi' at rightmost bit 'd' of int 'dst' with
    the corresponding subword at rightmost bit 's' of int/long 'src'.
    No operators in arguments, please: I put NO redundant parens here.
    This is designed to be maximally efficient: d, s, and wi are constants.
    GCC's assembler is merely: shift, xor, and, xor.
*/
#define w(dst, d, src, s, wi) \
    dst ^= (1 << wi) - 1 << d & (dst ^ (d > s ? src << d - s : src >> s - d));

// This acts like w(), but supports a long 'dst'.
#define wl(dst, d, src, s, wi) \
    dst ^= (1L << wi) - 1 << d & \
    (dst ^ (d > s ? (long) src << d - s : (long) src >> s - d));

/*
    Here are the firmware pointers. We can't free() these pointers,
    because the discrete event simulation may also be using them.
*/
static int *alpha0, *alpha1, *alpha2, *alpha3, *alpha4, *alpha5;
static int *beta0, *beta1, *beta2, *beta3, *beta4, *beta5;
static int *gamma0, *gamma1, *gamma2, *gamma3, *gamma4, *gamma5;
static int *theta, *zeta;
static long *decoder0, *decoder1;

// input and output connections for all those RAMs
static int a0i, a0o, a1i, a1o, a2i, a2o, a3i, a3o, a4i, a4o, a5i, a5o;
static int b0i, b0o, b1i, b1o, b2i, b2o, b3i, b3o, b4i, b4o, b5i, b5o;
static int g0i, g0o, g1i, g1o, g2i, g2o, g3i, g3o, g4i, g4o, g5i, g5o;
static int thi, tho, zei, zeo;              // theta, flags

// left and right register outputs, 36 bits each
static long lro, rro;

// opcode (9 bits), control decoder output words
static int opc;
static long d00, d01, d02, d03, d10, d11, d12, d13;

// flags input and output, with NRTZI in bits 01234
static int fli, flo;

// glue logic inputs and outputs
// Limit one gate per integer bank, to assure no timing races.
static int and1a, and1b, and1y;             // T input enabler
static int xor1a, xor1b, xor1y;             // T input flipper
static int and8i = bit7, and8o;             // 8-input AND for Z
static int xor2a, xor2b, xor2y;             // Z output flipper
static int and2a, and2b, and2y;             // for passing gamma's N
static int or1a, or1b, or1y;                // for substituting zeta's N

// hook to show the ALU panel, possibly do some tracing
int alu_trace;                  // 0 nothing, 1 panel, 2 detail

void alu_setup(void)
{
    int width, rows;

    if (zeta) return;           // may already be set up

    get_firmware("alpha0", &width, &rows, (void **) &alpha0);
    get_firmware("alpha1", &width, &rows, (void **) &alpha1);
    get_firmware("alpha2", &width, &rows, (void **) &alpha2);
    get_firmware("alpha3", &width, &rows, (void **) &alpha3);
    get_firmware("alpha4", &width, &rows, (void **) &alpha4);
    get_firmware("alpha5", &width, &rows, (void **) &alpha5);
    get_firmware("beta0", &width, &rows, (void **) &beta0);
    get_firmware("beta1", &width, &rows, (void **) &beta1);
    get_firmware("beta2", &width, &rows, (void **) &beta2);
    get_firmware("beta3", &width, &rows, (void **) &beta3);
    get_firmware("beta4", &width, &rows, (void **) &beta4);
    get_firmware("beta5", &width, &rows, (void **) &beta5);
    get_firmware("gamma0", &width, &rows, (void **) &gamma0);
    get_firmware("gamma1", &width, &rows, (void **) &gamma1);
    get_firmware("gamma2", &width, &rows, (void **) &gamma2);
    get_firmware("gamma3", &width, &rows, (void **) &gamma3);
    get_firmware("gamma4", &width, &rows, (void **) &gamma4);
    get_firmware("gamma5", &width, &rows, (void **) &gamma5);
    get_firmware("theta", &width, &rows, (void **) &theta);
    get_firmware("zeta", &width, &rows, (void **) &zeta);
    get_firmware("D0", &width, &rows, (void **) &decoder0);
    get_firmware("D1", &width, &rows, (void **) &decoder1);
}

// This works exclusively via global variables, because it's meant to
// operate on the /state/ of the hardware. It will grow to encompass most
// or all of the CPU.
static void alu_sim(void)
{
    long *dd;

    // control decoder RAM actuates
    dd = decoder0 + 4 * opc;
    d00 = *dd++, d01 = *dd++, d02 = *dd++, d03 = *dd;
    dd = decoder1 + 4 * opc;
    d10 = *dd++, d11 = *dd++, d12 = *dd++, d13 = *dd;

    // feed back flags from prior operation
    fli = flo;

    // AND XOR the previous T bit to theta.
    // There's ample time for this while the register file and alpha work.
    w(and1a, 0, fli, 2, 1)
    w(and1b, 0, d01, 29, 1)             // sync w/ decoder.c t_flip_allow
    and1y = and1a & and1b;              // actuate AND
    w(xor1a, 0, and1y, 0, 1)
    w(xor1b, 0, d01, 28, 1)             // sync w/ decoder.c t_flip_allow
    xor1y = xor1a ^ xor1b;              // actuate XOR
    w(thi, 12, xor1y, 0, 1)

    // operation to alpha
    w(a0i, 12, d00, 0, 6)   w(a1i, 12, d00, 0, 6)
    w(a2i, 12, d00, 0, 6)   w(a3i, 12, d00, 0, 6)
    w(a4i, 12, d00, 0, 6)   w(a5i, 12, d00, 0, 6)

    // operation to beta
    w(b0i, 12, d01, 0, 6)   w(b1i, 12, d01, 0, 6)   w(b2i, 12, d01, 0, 6)
    w(b3i, 12, d01, 0, 6)   w(b4i, 12, d01, 0, 6)   w(b5i, 12, d01, 0, 6)

    // operation to gamma
    w(g0i, 13, d02, 1, 5)  w(g1i, 13, d02, 1, 5)  w(g2i, 13, d02, 1, 5)
    w(g3i, 13, d02, 1, 5)  w(g4i, 13, d02, 1, 5)  w(g5i, 13, d02, 1, 5)

    // operation to theta and flags
    w(thi, 13, d01, 6, 5)
    w(zei, 11, d02, 6, 5)

    // left operand to alpha
    w(a0i, 0, lro,  0, 6)
    w(a1i, 0, lro,  6, 6)
    w(a2i, 0, lro, 12, 6)
    w(a3i, 0, lro, 18, 6)
    w(a4i, 0, lro, 24, 6)
    w(a5i, 0, lro, 30, 6)

    // right operand to alpha, beta, gamma
    w(a0i, 6, rro,  0, 6)   w(b0i, 6, rro,  0, 6)   w(g0i, 6, rro,  0, 6)
    w(a1i, 6, rro,  6, 6)   w(b1i, 6, rro,  6, 6)   w(g1i, 6, rro,  6, 6)
    w(a2i, 6, rro, 12, 6)   w(b2i, 6, rro, 12, 6)   w(g2i, 6, rro, 12, 6)
    w(a3i, 6, rro, 18, 6)   w(b3i, 6, rro, 18, 6)   w(g3i, 6, rro, 18, 6)
    w(a4i, 6, rro, 24, 6)   w(b4i, 6, rro, 24, 6)   w(g4i, 6, rro, 24, 6)
    w(a5i, 6, rro, 30, 6)   w(b5i, 6, rro, 30, 6)   w(g5i, 6, rro, 30, 6)

    // alpha RAMs actuate
    a0o = alpha0[a0i];
    a1o = alpha1[a1i];
    a2o = alpha2[a2i];
    a3o = alpha3[a3i];
    a4o = alpha4[a4i];
    a5o = alpha5[a5i];

    // alpha --> transpose --> beta
T:  w(b0i, 0, a0o, 0, 1)    w(b1i, 0, a0o, 1, 1)    w(b2i, 0, a0o, 2, 1)
    w(b0i, 1, a1o, 0, 1)    w(b1i, 1, a1o, 1, 1)    w(b2i, 1, a1o, 2, 1)
    w(b0i, 2, a2o, 0, 1)    w(b1i, 2, a2o, 1, 1)    w(b2i, 2, a2o, 2, 1)
    w(b0i, 3, a3o, 0, 1)    w(b1i, 3, a3o, 1, 1)    w(b2i, 3, a3o, 2, 1)
    w(b0i, 4, a4o, 0, 1)    w(b1i, 4, a4o, 1, 1)    w(b2i, 4, a4o, 2, 1)
    w(b0i, 5, a5o, 0, 1)    w(b1i, 5, a5o, 1, 1)    w(b2i, 5, a5o, 2, 1)

    w(b3i, 0, a0o, 3, 1)    w(b4i, 0, a0o, 4, 1)    w(b5i, 0, a0o, 5, 1)
    w(b3i, 1, a1o, 3, 1)    w(b4i, 1, a1o, 4, 1)    w(b5i, 1, a1o, 5, 1)
    w(b3i, 2, a2o, 3, 1)    w(b4i, 2, a2o, 4, 1)    w(b5i, 2, a2o, 5, 1)
    w(b3i, 3, a3o, 3, 1)    w(b4i, 3, a3o, 4, 1)    w(b5i, 3, a3o, 5, 1)
    w(b3i, 4, a4o, 3, 1)    w(b4i, 4, a4o, 4, 1)    w(b5i, 4, a4o, 5, 1)
    w(b3i, 5, a5o, 3, 1)    w(b4i, 5, a5o, 4, 1)    w(b5i, 5, a5o, 5, 1)

    // beta RAMs actuate
    b0o = beta0[b0i];
    b1o = beta1[b1i];
    b2o = beta2[b2i];
    b3o = beta3[b3i];
    b4o = beta4[b4i];
    b5o = beta5[b5i];

    // carry and propagate from alpha to theta
    w(thi, 0, a0o, 6, 1)    w(thi,  6, a0o, 7, 1)
    w(thi, 1, a1o, 6, 1)    w(thi,  7, a1o, 7, 1)
    w(thi, 2, a2o, 6, 1)    w(thi,  8, a2o, 7, 1)
    w(thi, 3, a3o, 6, 1)    w(thi,  9, a3o, 7, 1)
    w(thi, 4, a4o, 6, 1)    w(thi, 10, a4o, 7, 1)
    w(thi, 5, a5o, 6, 1)    w(thi, 11, a5o, 7, 1)

    // theta RAM actuates
    tho = theta[thi];

    // beta --> transpose --> gamma
    w(g0i, 0, b0o, 0, 1)    w(g1i, 0, b0o, 1, 1)    w(g2i, 0, b0o, 2, 1)
    w(g0i, 1, b1o, 0, 1)    w(g1i, 1, b1o, 1, 1)    w(g2i, 1, b1o, 2, 1)
    w(g0i, 2, b2o, 0, 1)    w(g1i, 2, b2o, 1, 1)    w(g2i, 2, b2o, 2, 1)
    w(g0i, 3, b3o, 0, 1)    w(g1i, 3, b3o, 1, 1)    w(g2i, 3, b3o, 2, 1)
    w(g0i, 4, b4o, 0, 1)    w(g1i, 4, b4o, 1, 1)    w(g2i, 4, b4o, 2, 1)
    w(g0i, 5, b5o, 0, 1)    w(g1i, 5, b5o, 1, 1)    w(g2i, 5, b5o, 2, 1)

    w(g3i, 0, b0o, 3, 1)    w(g4i, 0, b0o, 4, 1)    w(g5i, 0, b0o, 5, 1)
    w(g3i, 1, b1o, 3, 1)    w(g4i, 1, b1o, 4, 1)    w(g5i, 1, b1o, 5, 1)
    w(g3i, 2, b2o, 3, 1)    w(g4i, 2, b2o, 4, 1)    w(g5i, 2, b2o, 5, 1)
    w(g3i, 3, b3o, 3, 1)    w(g4i, 3, b3o, 4, 1)    w(g5i, 3, b3o, 5, 1)
    w(g3i, 4, b4o, 3, 1)    w(g4i, 4, b4o, 4, 1)    w(g5i, 4, b4o, 5, 1)
    w(g3i, 5, b5o, 3, 1)    w(g4i, 5, b5o, 4, 1)    w(g5i, 5, b5o, 5, 1)

    // carry decisions from theta to gamma
    w(g0i, 12, tho, 0, 1)   w(g1i, 12, tho, 1, 1)   w(g2i, 12, tho, 2, 1)
    w(g3i, 12, tho, 3, 1)   w(g4i, 12, tho, 4, 1)   w(g5i, 12, tho, 5, 1)

    // flag parameters to zeta (already got control decoder flag scheme)
    w(zei,  0, fli, 0, 4)
    w(zei,  4, a5o, 8, 5)
    w(zei,  9, tho, 6, 1)
    w(zei, 10, tho, 7, 1)

    // gamma RAMs actuate
    g0o = gamma0[g0i];
    g1o = gamma1[g1i];
    g2o = gamma2[g2i];
    g3o = gamma3[g3i];
    g4o = gamma4[g4i];
    g5o = gamma5[g5i];

    // zeta RAM actuates
    zeo = zeta[zei];

    // 8-input AND has one input tied to Vdd. Connect remainder.
    w(and8i, 0, g0o, 12, 1)
    w(and8i, 1, g1o, 12, 1)
    w(and8i, 2, g2o, 12, 1)
    w(and8i, 3, g3o, 12, 1)
    w(and8i, 4, g4o, 12, 1)
    w(and8i, 5, g5o, 12, 1)
    w(and8i, 6, zeo, 3, 1)              // "Z allowed"
    and8o = and8i == 255;               // AND8 actuates to give Z

    // There's an XOR after AND8 to override zero flag.
    w(xor2a, 0, and8o, 0, 1)
    w(xor2b, 0, zeo, 6, 1)
    xor2y = xor2a ^ xor2b;              // actuate XOR2

    // AND+OR chooses between flag RAM's N flag and copy of gamma35
    w(and2a, 0, g5o, 13, 1)             // gamma's N
    w(and2b, 0, zeo, 5, 1)              // use gamma's N
    and2y = and2a & and2b;              // actuate AND2
    w(or1a, 0, and2y, 0, 1)
    w(or1b, 0, zeo, 0, 1)               // zeta's N (0 if using gamma's N)
    or1y = or1a | or1b;                 // actuate OR1

    // output flags
    w(flo, 0, or1y, 0, 1)
    w(flo, 1, zeo, 1, 2)
    w(flo, 3, xor2y, 0, 1)
    w(flo, 4, zeo, 4, 1)
}

/* show ALU instrument panel */
static void glow(long y)
{
    int a_op = 0, b_op = 0, g_op = 0, t_op = 0;
    int t_cooked_in = 0;

    w(a_op, 0, d00,  0, 6)      // XXX Keep in sync with actual ALU.
    w(b_op, 0, d01,  0, 6)
    w(g_op, 1, d02,  1, 5)
    w(g_op, 0, xor1y, 0, 1)
    w(t_op, 1, d01, 7, 5)       // MWA 3 May 2021 think this is 1, 7, 5 - checking
    w(t_cooked_in, 0, xor1y, 0, 1)

    print36("\n        L " RED, lro, BAW "   ");
    printf("  " CYA "%c" BAW " -> " CYA "%c" BAW "\n",
        flo & bit1 ? 'T' : 't', xor1y & bit0 ? 'T' : 't');
    print36("        R " RED, rro, BAW "\n");

    printf(" %c  alpha ", alpha_emoji(a_op));
    print6(GRE,  a5o, " "); print6(NULL, a4o, " "); print6(NULL, a3o, " ");
    print6(NULL, a2o, " "); print6(NULL, a1o, " "); print6(NULL, a0o, BAW);
    print6("   p " CYA, 63 & thi >> 6, BAW "\n");

    if (alu_trace > 1) {
        printf("    trans ");
        print6(PUR, b5i, " ");print6(NULL, b4i, " "); print6(NULL, b3i, " "); 
        print6(NULL, b2i, " "); print6(NULL, b1i, " "); print6(NULL, b0i, BAW "\n");  
    }

    printf(" %c   beta ", beta_emoji(b_op));
    print6(GRE,  b5o, " "); print6(NULL, b4o, " "); print6(NULL, b3o, " ");
    print6(NULL, b2o, " "); print6(NULL, b1o, " "); print6(NULL, b0o, BAW);
    print6("   c " CYA, 63 & thi, BAW "\n");

    if (alu_trace > 1) {
        printf("  untrans ");
        print6(PUR, g5i, " ");print6(NULL, g4i, " "); print6(NULL, g3i, " "); 
        print6(NULL, g2i, " "); print6(NULL, g1i, " "); print6(NULL, g0i, BAW "\n");  
    }

    printf(" %c  gamma ", gamma_emoji(g_op));
    print6(GRE,  g5o, " "); print6(NULL, g4o, " "); print6(NULL, g3o, " ");
    print6(NULL, g2o, " "); print6(NULL, g1o, " "); print6(NULL, g0o, BAW);
    print6("   d " CYA, 63 & tho, BAW);
    printf(" %c\n", theta_emoji(t_op + t_cooked_in));

    printf("    flags " CYA "%.1s %.1s %.1s %.1s %.1s" BAW "\n",
        "Ii" + !(flo & bit4),
        "Nn" + !(flo & bit0), "Rr" + !(flo & bit1),
        "Tt" + !(flo & bit2), "Zz" + !(flo & bit3));

    printf(" unsigned " CYA "%li" BAW "\n", y);
    printf("   signed " CYA "%li" BAW "\n", y & pole ? y | comp36 : y);

    if (alu_trace > 1) {
#define what(huh) printf(#huh " %o\n", huh);
        puts("");
        what(fli)
        what(flo)
        what(and1a)
        what(and1b)
        what(and1y)
        what(and8i)
        what(and8o)
        what(xor1a)
        what(xor1b)
        what(xor1y)
        what(and2a)
        what(and2b)
        what(and2y)
        what(or1a)
        what(or1b)
        what(or1y)
        what(a5o)
        what(zei)
        what(zeo)
        what(thi)
        what(tho)
#undef what
    }
}

long alu(int opcode, long l, long r, int *flags)
{
    long y = 0;

    opc = opcode;
    lro = l;
    rro = r;
    flo = flags ? *flags : 0;           // immediately fed to fli
    alu_sim();

    wl(y,  0, g0o, 0, 6)
    wl(y,  6, g1o, 0, 6)
    wl(y, 12, g2o, 0, 6)
    wl(y, 18, g3o, 0, 6)
    wl(y, 24, g4o, 0, 6)
    wl(y, 30, g5o, 0, 6)

    if (alu_trace)
        glow(y);

    if (flags) *flags = flo;
    return y;
}

/*  FOOTNOTES

I:  Unused labels do not generate code with -O0 or -O3,
    so they are harmless for annotation of statements.

T:  Our "sliding word" transpose() using a long generates fewer instructions,
    but does not include packing and unpacking the long. I suspect this
    construct is faster, based on looking at -O0 and -O3 assembly output
    from both.

*/
