/*
    asm.c
    This assembles Dauug|36 instructions and optionally interprets
    the machine code in a small VM for software testing.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "../consts/opcode.h"
#include "../consts/slots.h"
#include "../firmware/firmware.h"
#include "../misc/misc.h"
#include "./context.h"

#include "errors.c"
#include "syms.c"
#include "lex.c"
#include "syntax.c"
#include "alu.c"

// Mini-disassembler for listings
void asm_disassemble(char *prefix, long code, char *suffix)
{
    int opc = code >> 27 & 511;
    int l = code >> 18 & 511;
    int r = code >> 9 & 511;
    int d = code & 511;

    printf("%s%7s %i %i %i%s", prefix, opcode_name(opc), l, r, d, suffix);
}

// Reset program. Provided automatically after assembly only.
void asm_reset(void *context)
{
    X = context;

    preset_regs(X->sym_root);           // zero registers & preset constants
    X->ip = 0;
    X->flags = 0;
    X->n_stack = 0;
}

// Assemble from source already attached to X.
// If any errors, eMsg_head will have the list.
static void assemble_from_context(void)
{
    int pass, parsing, code_alloced = 0;
    long code;
    char *s;
    struct eMsg **last_err;

    X->n_regs = 0;

    for (pass = 1; pass < 3; ++pass) {
        parsing = 1;
        X->n_code = 0;
        s = X->src;

        while (parsing) {
            last_err = X->eMsg_tail;    // Notice if error message issues.
            code = -1;                  // This only changes if parse succeeds.

            if (newline(&s)) {          // Blank lines are legal and frequent.
                if (!*s) {
                    code = (long) oc_HALT << 27;
                    parsing = 0;        // how parsing terminates
                }
            }

            else if (label(&s, pass))
                ;                       // keep parsing after label

            else if (assignment_statement(&s, pass, &code))
                ;                       // something = ...

            else if (keyword_statement(&s, pass, &code))
                ;                       // keyword ...

            else {
                if (last_err == X->eMsg_tail)
                    note_parse_error(s, "syntax error", NULL, NULL);
                flush_eol(&s);          // An error message has issued.
            }

            if (code != -1) {           // code output?
                if (X->code) {          // pass 2?
                    X->code[X->n_code] = code;
                    if (X->list)
                        asm_disassemble("", code, "\n");
                }
                ++X->n_code;            // Keep count during both passes.
            }
        }

        if (pass == 1) {
            number_regs(X->sym_root);
            if (X->eMsg_head)
                break;
            X->code = trackmem(NULL, X->n_code * sizeof(*X->code));
            X->regs = trackmem(NULL, X->n_regs * sizeof(*X->regs));
            code_alloced = X->n_code;
        }
    }

    if (!X->eMsg_head) {
        if (X->n_code > code_alloced)
            ng("!assembler overrun %i %i", X->n_code, code_alloced);
        asm_reset(X);                   // ready to execute immediately
    }
}

// Deallocate assembler context.
void asm_dealloc(void *context)
{
    X = context;
    if (!X) return;
    clear_symbol_table(&X->sym_root);
    clear_parse_errors();
    trackmem(X->code, 0);
    trackmem(X->regs, 0);
    if (X->free_src) {
        trackmem(X->src, 0);
        X->src = NULL;
    }
    trackmem(X, 0);
    X = NULL;
}

// Run program (with lots of sanity checking).
void asm_run(void *context)
{
    // representation of conditions for call, jump, return
    enum pseudoflags {
        q_n = 1, q_r = 2, q_t = 4, q_z = 8,
        q_p = 1 << 16, q_not_r = 1 << 17, q_not_t = 1 << 18,
        q_call = 1 << 19, q_jump = 1 << 20, q_return = 1 << 21
    };

    // representation of when said conditions apply
    static struct {
        int j_op;
        int when;
    } jump_if_table[] = {
        oc_J   , q_t | q_not_t,
        oc_J_lt, q_n,
        oc_J_le, q_n | q_z,
        oc_J_eq, q_z,
        oc_J_ne, q_n | q_p,
        oc_J_ge, q_z | q_p,
        oc_J_gt, q_p,
        oc_J_r , q_r,
        oc_J_nr, q_not_r,
        oc_J_t , q_t,
        oc_J_nt, q_not_t,
    };

    long inst, y, l, r;                 // instruction, register *values*
    int opcode, yn, ln, rn;             // opcode, register *numbers*
    int maxreg;                         // a bound
    int flags;                          // pseudoflags from ALU
    int i;                              // loop

    X = context;
    maxreg = X->n_regs;

    // Make sure some hooligan didn't set invalid register bits.
    for (ln = 0; ln < maxreg; ++ln)
        if (X->regs[ln] & comp36)
            ng("!user goofed up register %i", ln);

    // run now
    while (1) {
        // Make sure IP points to code (no crazy branch).
        if (X->ip < 0 || X->ip >= X->n_code) {
            ng("asm_run branch outside code (%i)", X->ip);
            return;                     // bad, but no exit()
        }

        inst = X->code[X->ip++];        // read inst
        opcode = inst >> 27 & 511;      // decode opcode

        if (opcode < oc_NO_MORE_ALU) {  // This is an ALU operation.

            ln = inst >> 18 & 511;      // decode register numbers
            rn = inst >> 9 & 511;
            yn = inst & 511;

            if (yn >= maxreg || ln >= maxreg || rn >= maxreg)
                ng("!asm_run decoded invalid register");

            l = X->regs[ln];            // fetch register values
            r = X->regs[rn];

            // Do the ALU operation here.
            y = alu(opcode, l, r, &X->flags);
            if (y & comp36)
                ng("!asm_run ALU returned extraneous bits");

            // Write the result back.
            switch (opcode) {
/*
    TODO This switch block needs to be replaced by using the decoded
    instruction. But as of 29 April the hardware design does not
    accommodate reading G without writing the result to L and R.
    The gamma lockout has to be expanded, and simulation will be
    needed to ensure nothing breaks when that change is made.
*/
                case oc_CRF:
                case oc_BOUND_u:
                case oc_BOUND_s:
                case oc_CMP_uu:
                case oc_CMP_us:
                case oc_CMP_su:
                case oc_CMP_ss:
                    break;
                default:
                    X->regs[yn] = y;
            }

            // Trace the XL operation as diagnostic output.
            if (opcode == oc_XL)
                trace_reg(X->sym_root, ln);

            continue;
        }

        // This is a non-ALU operation.
        switch (opcode) {
            case oc_HALT:                       // HALT
                return;

            case oc_NOP:                        // NOP
                break;

            case oc_C:                          // CALL
                if (X->n_stack >= max_call_depth) {
                    ng("call depth too deep");  // bad, but no exit()
                    return;
                }
                X->stack[X->n_stack++] = X->ip;
                X->ip = inst & 0777777777;
                break;

            case oc_R:                          // RETURN
                if (X->n_stack <= 0) {
                    ng("return without call");  // bad, but no exit()
                    return;
                }
                X->ip = X->stack[--X->n_stack];
                break;

            default:                            // any of several JUMPs

                // Compute pseudoflags.
                flags = X->flags;
                ifnot (flags & (q_n | q_z))
                    flags |= q_p;
                ifnot (flags & q_t)
                    flags |= q_not_t;
                ifnot (flags & q_r)
                    flags |= q_not_r;

                for (i = 0; i < CARD(jump_if_table); i++) {
                    if (opcode != jump_if_table[i].j_op)
                        continue;
                    goto yes;
                }
                ng("!bonkers opcode %i in asm_run()", opcode);
yes:
                if (flags & jump_if_table[i].when)
                    X->ip = inst & 0777777777;
                break;
        }
    }
}

/*
    Assemble from source code. Returns an opaque pointer.

    flags:

        b   Print a banner message before doing anything.
        c   Show changes to registers after execution.
        d   Deallocate everything and return NULL when done.
        f   Supplied 'source' is a filename, not a buffer.
        l   Provide a listing.
        r   Run what is assembled.

    TODO If assembly fails in Pass 1, memory won't be attached to
    the context, and execution will segfault.
    XXX DON'T FIX UNTIL PROGRAM CAN BE RUN FOR TESTING.
*/
void *asm_assemble(char *source, char *flags, char *banner)
{
    char *fName;

    if (!flags)
        flags = "";

    // set up context to work in
    X = trackmem(NULL, sizeof(*X));
    X->eMsg_tail = &X ->eMsg_head;
    X->src = source;
    X->list = !!strchr(flags, 'l');

    // print signon message if called for
    if (strchr(flags, 'b'))
        puts(banner);

    // load file if called for
    if (strchr(flags, 'f')) {
        if (X->src = slurp(fName = X->src))
            X->free_src = 1;
        else {
            ng("cannot open assembler source '%s'", fName);
            X->src = "";                        // bad, but no exit()
        }
    }

    // assemble and reset
    if (X->list)
        printf("\n--- code ---\n");
    assemble_from_context();

    // list symbols
    if (X->list && X->sym_root) {
        printf("\n--- symbols ---\n");
        show_symbol_table(X->sym_root);
    }

    // list assembly errors
    if (X->eMsg_head) {
        printf("\n--- errors ---\n");
        show_parse_errors(X->src);

        // XXX DON'T FIX UNTIL PROGRAM CAN BE RUN FOR TESTING.
        // XXX Kludge: Assembly errors terminate the program.
        // Otherwise, this would segfault due to regs & code
        // not attached to the context. Want to straighten all
        // this out someday.
        ng("!terminating");
    }

    // run
    else if (strchr(flags, 'r')) {
        printf("\n--- running ---\n");
        asm_run(X);
        if (strchr(flags, 'c')) {       // show regs that changed
            printf("\n--- changed registers ---\n");
            show_changed_regs(X->sym_root);
            printf("\n");
        }
    }

    // dealloc
    if (strchr(flags, 'd')) {
        asm_dealloc(X);
        X = NULL;
    }

    return X;
}

// Return pointer to specified register. Will ng() if no such name.
// asm_run() will ng() if the register is abused by caller.
long *asm_get_register(void *context, char *name)
{
    static long bogus_register;

    X = context;
    struct sym *s = find_symbol(name, NULL, 0, 1);
    if (!strchr("su", s->what)) {
        ng("%s is not a register", name);
        return &bogus_register;                 // bad, but no exit()
    }
    if (s->number < 0 || s->number >= X->n_regs) {
        ng("%s is for invalid register %i", name, s->number);
        return &bogus_register;                 // bad, but no exit()
    }
    return X->regs + s->number;
}

// Returns pointer to flags.
int *asm_get_flags(void *context)
{
    X = context;
    return &X->flags;
}

/*
    Given a code RAM in 'out' and its size in 'rows' from netsim, this
    function copies assembled code to the RAM so it can be tested.
    Returns the number of words copied.
*/
int asm_fetch_code(void *context, long *out, int rows)
{
    int i;

    X = context;

    if (X->n_code > rows)
        ng("too much code: %i pigeons into %i holes", rows, X->n_code);
    else
        rows = X->n_code;

    for (i = 0; i < rows; i++)
        out[i] = X->code[i];

    return rows;
}
