/*
    errors.c
    This de-duplicates and reports parser error messages for the assembler.
*/

// linked list of parse errors
struct eMsg {
    struct eMsg *next;
    char *where;                    // place in asm where it happend
    char *explain;                  // what to say to the user
    char *vi;                       // start of offending identifier
    char *de;                       // end of offending identifier
};

// Add a parser error for later output to user.
static void note_parse_error(char *where, char *explain, char *vi, char *de)
{
    struct eMsg *m;

    // Deduplicate. O(n^2).
    for (m = X->eMsg_head; m; m = m->next)
        if (m->where == where && m->explain == explain && m->vi == vi)
            return;

    // Good to go.
    m = trackmem(NULL, sizeof(struct eMsg));
    m->where = where;
    m->explain = explain;
    m->vi = vi;
    m->de = de;
    *X->eMsg_tail = m;
    X->eMsg_tail = &m->next;
}

// Clear previous parser errors.
static void clear_parse_errors(void)
{
    struct eMsg *m;

    while (X->eMsg_head) {
        m = X->eMsg_head->next;
        trackmem(X->eMsg_head, 0);
        X->eMsg_head = m;
    }

    X->eMsg_tail = &X->eMsg_head;
}

// Find the line number of position 'at' within a multi-line 'text'.
// This would return 1 + (number of newlines) if not found, but the
// surrounding code always has 'at' within 'text'.
static int count_line_number(char *text, char *at)
{
    int n;

    for (n=1; text != at; ++text)
        if (*text == '\n')
            ++n;
    return n;
}

// Document assembler errors to stderr. Requires pointer to source.
static void show_parse_errors(char *start)
{
    struct eMsg *m;
    int lineNo, len;

    for (m = X->eMsg_head; m; m = m->next) {
        lineNo = count_line_number(start, m->where);
        fprintf(stderr, "line %i", lineNo);
        if (m->vi) {
            len = m->de ? m->de - m->vi : strlen(m->vi);
            fprintf(stderr, " (%.*s)", len, m->vi);
        }
        fprintf(stderr, ": %s\n", m->explain);
    }
}
