/*
    lex.c
    Assembler terminals and low-level parsing
*/

/*
    This swallows spaces, tabs, and comments from ; to end of line.
    It does NOT eat wild newlines, as we are an assembler.

    It also swallows parenthesized text, including newlines, where
    the number of consecutive closing parentheses must match the
    number which opened. So (in-line comment) as well as (multi-line
    comment) and ((comment with encapsulated parenthetical remarks
    (and other interesting discourses) get swallowed too)).

    This routine only needs to be called within terminal symbols;
    e.g., by code that examines **s.
*/
static void eatBlanks(char **s)
{
    int lParens, rParens, saw_semi = 0;
    char *began_on;

    while (**s && **s != '\n') {

        if (**s == ' ' || **s == '\t' || saw_semi) {
            ++*s;
            continue;
        }

        if (**s == ';') {
            saw_semi = 1;                   // detect ; comment
            continue;
        }

        lParens = 0;                        // detect stream of (
        while (**s == '(')
            ++*s, ++lParens;
        if (!lParens)
            break;                          // NOT blank, tab, semi, (
        began_on = *s;                      // line comment began on

        while (1) {
            while (**s && **s != ')')       // skip to first )
                ++*s;

            rParens = 0;                    // count number of )
            while (**s == ')')
                ++*s, ++rParens;

            if (rParens == lParens)         // end of (( comment ))
                break;

            if (!**s) {                     // error: EOF in (( comment
                note_parse_error
                    (began_on, "end of file in ( comment", NULL, NULL);
                break;
            }
        }
    }
}

/*
    Detect newline and advance parser past it.
    An implicit newline is at the end of the input.
*/
static int newline(char **ss)
{
    char *s = *ss;

    eatBlanks(&s);
    if (*s) {
        if (*s != '\n')
            return 0;
        ++s;
    }
    *ss = s;
    return 1;
}

/*
    Flush through newline. Used to continue assembly after bad syntax.
*/
static void flush_eol(char **ss)
{
    while (**ss) {
        eatBlanks(ss);          // pass multiline comment after bad syntax

        if (isgraph(**ss))
            ++*ss;              // pass suspicious character

        if (**ss == '\n') {
            ++*ss;              // pass newline and stop
            break;
        }
    }                           // also stop at null
}

/*
    Advance past an arbitrary symbol, like "<>", if present.

    By "symbol" here we mean a short non-empty string of characters,
    probably without any numbers or letters. So this word "symbol" doesn't
    have the same meaning as, say, an ASCII symbol (which is not a string),
    nor does it mean an entry in the assembler's symbol table (which
    provides a means to look up registers by name).
*/
static int symbol(char **ss, char *sym)
{
    char *s = *ss;
    eatBlanks(&s);
    while (*sym && *s == *sym)
        ++s, ++sym;
    if (*sym) return 0;
    *ss = s;
    return 1;
}

/*
    This detects an identifier, returns its start position, and
    advances 's' just past its end.

    Identifiers start with a letter and may contain letters, digits,
    apostrophes, and underscores.
*/
static char *identifier(char **ss)
{
    char *s = *ss, *start;

    eatBlanks(&s);
    start = s;
    if (!isalpha(*s))
        return NULL;

    while (isalnum(*s) || *s == '\'' || *s == '_' || *s == '.')
        ++s;
    *ss = s;
    return start;
}

// Advance past a keyword if present.
static int keyword(char **ss, char *kw)
{
    char *s = *ss;  
    char *id = identifier(&s);
    int len;

    if (!id) return 0;                  // no word here
    len = strlen(kw);
    if (len != s - id) return 0;        // wrong length
    if (memcmp(id, kw, len)) return 0;  // not the word we seek

    *ss = s;
    return 1;
}

/*
    Parse an integer within the 36-bit signed and unsigned range,
    which is -34_359_738_368 through +68_719_476_735.
    Anything beyond this range will report as an error.
    Digits, +, -, and _ are allowed in appropriate places.

    XXX This does not match the assembler reference manual,
        in that optional radixes are not supported other than octal,
        which works like C instead of per the manual.
*/
static int scan_number(char **ss, long *num)
{
    char *s = *ss;

    int ten = 10;
    int neg = 0;
    long n = 0;                     // Presumably 64 bits. Need at least 41.

    if (*s == '+')
        ++s;
    else if (*s == '-')
        ++s, neg = 1;

    if (!isdigit(*s))
        return 0;

    if (*s == '0')
        ten = 8;

    // There is at least a one-digit number.
    for (; ; ++s) {
        if (*s == '_')
            continue;
        if (!isdigit(*s))
            break;
        n = ten * n + *s - '0';
        if (n > mask36) {           // This test needs 41 bits for n.
            note_parse_error(s, "number is too large", NULL, NULL);
            return 0;
        }
    }

    if (neg) {
        if (n > pole) {
            note_parse_error(s, "number is too small", NULL, NULL);
            return 0;
        }
        n = -n;
    }

    *ss = s;
    if (num)
        *num = n;
    return 1;
}

// Look for a number and report it back as a constant-valued register.
// The lexing portion is factored out for reuse by reg_to_sym().
static int number(char **ss, struct reg_name *r)
{
    char *s = *ss;

    eatBlanks(&s);
    r->vi = s;
    if (!scan_number(&s, NULL))
        return 0;
    r->de = s;
    r->cast = 0;
    *ss = s;
    return 1;
}

// Look for "label :" at the beginning of a line.
static int label(char **ss, int pass)
{
    char *s = *ss, *id, *end;
    struct sym *l;

    ifnot (id = identifier(&s))
        return 0;
    end = s;
    ifnot (symbol(&s, ":"))
        return 0;

    // Syntax is good. Note line parsing may continue after the colon.
    *ss = s;
    if (pass != 1) return 1;

    // In pass 1, label goes in symbol table.
    l = find_symbol(id, end, 1, 0);
    if (l->what)
        note_parse_error(s, "label defined more than once", id, end);
    else {
        l->what = 'l';
        l->number = X->n_code;
    }
    return 1;
}

// A cast may be either "<unsigned>" or "<signed>",
// or abbreviated <u.> or <s.>. Returns 'u', 's', or 0 if no cast.
static int cast(char **ss)
{
    int c = 0;
    char *s = *ss;

    do {
        ifnot (symbol(&s, "<"))
            break;

        if (keyword(&s, "signed") || keyword(&s, "s."))
            c = 's';
        else if (keyword(&s, "unsigned") || keyword(&s, "u."))
            c = 'u';
        else
            break;

        ifnot (symbol(&s, ">"))
            break;
        *ss = s;

    } while (0);

    return c;
}

// Lvalues are identifiers, preceded optionally by a cast.
static int lvalue(char **ss, struct reg_name *r)
{
    char *s = *ss;

    r->cast = cast(&s);
    ifnot (r->vi = identifier(&s))
        return 0;

    r->de = *ss = s;
    return 1;
}

// Registers can be numbers or lvalues.
// A number is a constant register containing that number.
// An lvalue is something that is allowed left of an '=' sign.
static int parse_register(char **ss, struct reg_name *r)
{
    return lvalue(ss, r) || number(ss, r);
}
