/*
    syntax.c
    Assembler high-level parsing and instruction encoding
*/

/*
    This function channels parser output into a 36-bit instruction.
    This involves looking up register numbers, and stratifying
    opcodes based on each register's declared or cast signedness.
*/
static void encode_instruction(
    char *place,             // point in source to report if error occurs
    int opcode,              // what we're compiling
    struct reg_name *lreg,   // left source register
    struct reg_name *rreg,   // right source register
    struct reg_name *yreg,   // destination register
    long *code               // assembled code
) {
    int signage = 0,                    // bitmask (enum signage)
        ys = 'u', ls = 'u', rs = 'u',   // y, l, r signedness
        yn = 0, ln = 0, rn = 0;         // y, l, r register numbers

    struct sym *y, *l, *r;              // y, l, r symbols

    // Convert register names to register numbers and sigednesses.
    if (yreg) {
        y = reg_to_sym(yreg, place, &ys);
        if (ys == 's') signage |= dest_signed;
        yn = y->number;
    }

    if (lreg) {
        l = reg_to_sym(lreg, place, &ls);
        if (ls == 's') signage |= left_signed;
        ln = l->number;
    }

    if (rreg) {
        r = reg_to_sym(rreg, place, &rs);
        if (rs == 's') signage |= right_signed;
        rn = r->number;
    }

    opcode = stratify_opcode(opcode, signage);

    // pack the CPU instruction
    *code = (long) opcode << 27 | ln << 18 | rn << 9 | yn;
}

/*
    This parses unary operations on the right side of an assignment
    statement, and generates a 36-bit instruction if successful.
    Otherwise, it returns zero.

    The two basic formats (after the = sign) are

        register
        OP register

    where OP is one of { !, not, +, ++, -, -- }.
*/
static int rhs_unary(char **ss, int pass, struct reg_name *lhs, long *code)
{
    static struct opcodes {
        int opcode;
        char *sym;
        int zero_left;
    } opcodes[] = {
        oc_NL , "not", 0, oc_NL, "!" , 0,
        ocat_AC , "++" , 1, ocat_A , "+" , 1,
        ocat_SC , "--" , 1, ocat_S , "-" , 1, -1, NULL, -1
    };

    static struct reg_name zeroReg = { "0" };

    int opcode = ocat_A, zl = 1;        // if none given, identity
    char *s = *ss, *where;
    struct reg_name r;
    struct opcodes *o;

    for (o = opcodes; o->sym; ++o)      // look for symbol
        if ((isalpha(*o->sym) ? keyword : symbol)(&s, o->sym)) {
            opcode = o->opcode;
            zl = o->zero_left;          // may need to subst left arg
            break;
        }

    ifnot (parse_register(&s, &r))      // register follows symbol
        return 0;
    where = s;
    ifnot (newline(&s))
        return 0;

    *ss = s;                            // syntax checks out
    encode_instruction(where, opcode, zl ? &zeroReg : &r, &r, lhs, code);
    return 1;
}

/*
    This parses binary operations on the right side of an assignment
    statement, and generates a 36-bit instruction if successful.
    Otherwise, it returns zero.

    The basic format (after the = sign) is

        register OP register

    where OP is one of { +, ++, -, --, ~-, ~-- } or one of many
    keywords in 'opwords' below.
*/
static int rhs_binary(char **ss, int pass, struct reg_name *lhs, long *code)
{
    static char *opwords[] = {      // sync with below
        "igf", "igt", "xl", "xr", "nl", "nr", "and", "nand",
        "or", "nor", "xor", "xnor", "lanr", "ranl", "lonr", "ronl",
        "a", "ac", "s", "sc", "rs", "rsc",
        "min", "max", "asl",
        "asr", "lsl", "lsr", "rol", "ml", "mh", "mhns", "dsl",
        "copy", "nudge", "pit", "pat", "pait", "mix", "xim", "ham.3",
        "stun.a", "stun.b", "stun.c", "stun.d", "stun.e", "stun.f",
        "stun.g", "stun.h", "stun.i", "stun.j",
        "++", "+", "--", "-", "~--", "~-", NULL
    };

    static int opcodes[] = {            // sync with above
        oc_IGF, oc_IGT, oc_XL, oc_XR, oc_NL, oc_NR, oc_AND, oc_NAND,
        oc_OR, oc_NOR, oc_XOR, oc_XNOR, oc_LANR, oc_RANL, oc_LONR, oc_RONL,
        ocat_A, ocat_AC, ocat_S, ocat_SC, ocat_RS, ocat_RSC,
        ocat_MIN, ocat_MAX, ocat_ASL,
        ocat_ASR, oc_LSL, oc_LSR, oc_ROL, oc_ML, oc_MH, oc_MHNS, oc_DSL,
        oc_SWIZ, oc_NUDGE, oc_PIT, oc_PAT, oc_PAIT, oc_MIX, oc_XIM, oc_HAM3,
        oc_STUN_A, oc_STUN_B, oc_STUN_C, oc_STUN_D, oc_STUN_E, oc_STUN_F,
        oc_STUN_G, oc_STUN_H, oc_STUN_I, oc_STUN_J,
        ocat_AC, ocat_A, ocat_SC, ocat_S, ocat_RSC, ocat_RS
    };  

    static struct reg_name neg1Reg = { "-1" };

    char *s = *ss, *where, *ow;
    struct reg_name l, r;
    int i, opcode;

    ifnot (parse_register(&s, &l))
        return 0;

    for (i=0; opwords[i]; i++) {
        ow = opwords[i];
        if (isalpha(*ow)) {
            if (keyword(&s, ow))
                goto yes;
        }
        else if (symbol(&s, ow))
            goto yes;
    }
    return 0;

yes:
    opcode = opcodes[i];
    where = s;
    ifnot (parse_register(&s, &r) && newline(&s))
        return 0;
    *ss = s;                            // syntax checks out

    // IGT and NR each override one operand with all ones.
    // This makes IGT kind of useless, as its output is also all ones.
    // NR isn't much better, as NL can achieve the same result on most
    // implementations. But I don't feel right implementing just 14
    // out of the 16 bitwise binary operations, and leaving two out,
    // even though they are the exact two that (a) don't add value, and
    // (b) require the assembler to supply a special constant.
    if (opcode == oc_IGT)
        r = neg1Reg;
    else if (opcode == oc_NR)
        l = neg1Reg;

    encode_instruction(where, opcode, &l, &r, lhs, code);
    return 1;
}

/*
    This parses two-argument bitwise boolean operations on the
    right side of an assignment statement, and generates a 36-bit
    instruction if successful. Otherwise, it returns zero.

    There are 4 basic formats (after the = sign):

          register LOGIC   register
          register LOGIC ! register
        ! register LOGIC   register
        ! register LOGIC ! register

    where LOGIC is one of: { &, |, ^, !&, !|, !^ }.
*/
static int rhs_boolean(char **ss, int pass, struct reg_name *lhs, long *code)
{
    // In this table, a leading and/or trailing ~ means that a ! was found
    // before the left and/or right register operand, respectively.
    static char *logic[] = {
        "&"  , "|"  , "^"  , "!&"  , "!|"  , "!^"  ,
        "~&" , "~|" , "~^" , "~!&" , "~!|" , "~!^" ,
        "&~" , "|~" , "^~" , "!&~" , "!|~" , "!^~" ,
        "~&~", "~|~", "~^~", "~!&~", "~!|~", "~!^~", NULL
    };

    static int opcodes[] = {
        oc_AND , oc_OR  , oc_XOR , oc_NAND, oc_NOR, oc_XNOR,
        oc_RANL, oc_RONL, oc_XNOR, oc_LONR, oc_LANR, oc_XOR,
        oc_LANR, oc_LONR, oc_XNOR, oc_RONL, oc_RANL, oc_XOR,
        oc_NOR , oc_NAND, oc_XOR , oc_OR  , oc_AND, oc_XNOR,
    };

    char *s = *ss, *where;
    char found[5], *f = found;
    int i;
    struct reg_name l, r;

    if (symbol(&s, "!"))
        *f++ = '~';

    ifnot (parse_register(&s, &l))
        return 0;
    where = s;

    eatBlanks(&s);
    if (*s == '!')
        *f++ = *s++;
    if (strchr("&|^", *s))
        *f++ = *s++;
    else
        return 0;

    if (symbol(&s, "!"))
        *f++ = '~';
    *f = '\0';

    ifnot (parse_register(&s, &r) && newline(&s))
        return 0;

    // syntax checks out
    *ss = s;

    for (i = 0; logic[i]; ++i)
        if (!strcmp(found, logic[i]))
            goto yes;
    ng("!rhs_boolean internal error %s", found);

yes:
    encode_instruction(where, opcodes[i], &l, &r, lhs, code);
    return 1;
}

/*
    This parses an assignment statement, basically anything where '='
    is used to store something to a register. In most cases, the right
    hand side involves a non-identity ALU operation.
*/
static int assignment_statement(char **ss, int pass, long *code)
{
    char *s = *ss;
    struct reg_name lhs;

    ifnot (lvalue(&s, &lhs) && symbol(&s, "="))
        return 0;

    if (rhs_unary(&s, pass, &lhs, code))
        ;

    else if (rhs_binary(&s, pass, &lhs, code))
        ;

    else if (rhs_boolean(&s, pass, &lhs, code))
        ;

    else
        return 0;

    *ss = s;
    return 1;
}

/*
    This parses the conditional specification from a JUMP instruction
    and returns its stratified opcode.

        -t  +t  -r  +r  <  <=  ==  !=  >=  >  (none)
*/
static int jump_condition(char **ss)
{
    enum {
        plus = -2, minus = -1,
    };

    static struct {
        char *sym;
        int opcode;
    } jtable[] = {
        "<=", oc_J_le,
        "<", oc_J_lt,
        "==", oc_J_eq,
        "!=", oc_J_ne,
        ">=", oc_J_ge,
        ">", oc_J_gt,
        "+", plus,
        "-", minus,
    };

    int i, oc;
    char *s = *ss;                          // parser has not advanced (yet)

    for (i = 0; i < CARD(jtable); i++)
        if (symbol(&s, jtable[i].sym))
            goto yes;

    return oc_J;                            // unconditional jump

yes:
    *ss = s;                                // parser did advance
    oc = jtable[i].opcode;
    if (oc >= 0)
        return oc;                          // jump after magnitude compare

    if (keyword(ss, "r"))                   // jump if R set or clear
        return oc == plus ? oc_J_r : oc_J_nr;

    if (keyword(ss, "t"))                   // jump if T set or clear
        return oc == plus ? oc_J_t : oc_J_nt;

    note_parse_error(s, "flag \"r\" or \"t\" must follow + or -",
                    NULL, NULL);
    return oc_J;
}

/*
    This parses keyword statements, which differ from assignment statements
    in that they begin with a keyword instead of a register and equal sign.
    Some of these have abbreviations; e.g., "u." == "unsigned".
    Presently supported keywords include:

    bound  call  cmp  crf  eject  halt  jump  nop  return  signed  unsigned
*/
static int keyword_statement(char **ss, int pass, long *code)
{
    static char *encode_kw[] = { "s signed", "s s.", "u unsigned", "u u.",
        "b bound", "b b.", "a call", "a c.", "j jump", "j j.",
        "r return", "r r.", "e eject'", "h halt",
        "n nop", "m cmp", "c crf", "i imm", "w wcm", 
        "1 rcm1", "2 rcm2", NULL };

    char *s = *ss, *id, *where;
    int kw, opcode = 0;
    long addr = 0, imm_val;
    struct sym *sym;
    struct reg_name l, r;

    for (kw = 0; encode_kw[kw]; ++kw)       // grab the keyword
        if (keyword(&s, encode_kw[kw] + 2)) {
            kw = *encode_kw[kw];
            goto yes;
        }
    return 0;

yes:
    where = s;
    switch (kw) {
        case 's': case 'u':                 // signed, unsigned
            while (id = identifier(&s)) {
                if (pass != 1)
                    continue;
                sym = find_symbol(id, s, 1, 0);
                if (sym->what)
                    note_parse_error
                            (s, "register defined more than once", id, s);
                else
                    sym->what = kw;
                if (symbol(&s, ","))
                    note_parse_error(s, "lose the comma", sym->name, NULL);
            }
            break;

        // The ALU's BOUND has the index as the right operand, but the
        // assembler syntax (for the programmer's ease of learning) has the
        // index as the left operand. So left and right are reversed here.
        // The required "<" in the syntax is to help the programmer
        // remember what the instruction does.
        case 'b':                           // bound
            ifnot (parse_register(&s, &r))
                return 0;
            ifnot (symbol(&s, "<"))
                return 0;
            ifnot (parse_register(&s, &l))
                return 0;
            break;

        // The required "-" in CMP's syntax is to help the programmer
        // remember what the instruction does.
        case 'm':                           // cmp
            ifnot (parse_register(&s, &l))
                return 0;
            ifnot (symbol(&s, "-"))
                return 0;
            ifnot (parse_register(&s, &r))
                return 0;
            break;

        case 'r':                           // return
            opcode = oc_R;
            break;

        case 'a':                           // call
            opcode = oc_C;
            goto get_label;

        case 'j':                           // jump
            opcode = jump_condition(&s);

get_label:
            ifnot (id = identifier(&s)) {   // get call/jump label
                note_parse_error
                        (s, "need destination for jump or call", NULL, NULL);
                break;
            }

            if (pass == 1)
                break;

            sym = find_symbol(id, s, 1, 0);
            if (sym->what != 'l')         // ensure dest is a label
                note_parse_error
                        (s, "destination must be a label", r.vi, r.de);
            addr = sym->number;
            break;

        case 'e':                           // eject
            if (pass == 2 && X->list)
                putchar('\n');
            break;

        case 'i':                           // imm
            if (!lvalue(&s, &l))
                return 0;
            eatBlanks(&s);
            if (!scan_number(&s, &imm_val)) {
                note_parse_error
                        (s, "need immediate value after register", NULL, NULL);
                return 0;
            }

            // Only certain immediate values can be represented.
            // case 1: [0, 262143] (that's 777777`o)
            ifnot (imm_val & ~0777777L)
                opcode = oc_IMP;

            // case 2: positive multiples of 262144
            else ifnot (imm_val & 0777777L) {
                opcode = oc_IMH;
                imm_val >>= 18;
            }

            // case 3: positive multiples of 262145,
            // meaning the high and low 18 bits are identical
            else ifnot (0777777L & (imm_val >> 18 ^ imm_val)) {
                opcode = oc_IMB;
                imm_val &= 0777777L;
            }

            // case 4: negative and within [-262143, -1]
            else if (imm_val >= -0777777L && imm_val < 0) {
                opcode = oc_IMN;
                imm_val &= 0777777L;
            }

            // There are no more cases.
            else {
                note_parse_error
                        (s, "immediate value not representable", NULL, NULL);
                return 0;
            }
            break;

        case 'c':                           // crf
        case 'h':                           // halt
        case 'n':                           // nop
            break;

        case 'w':                           // wcm (write code memory)
            opcode = oc_WCM;
            if (!parse_register(&s, &l)) {
                note_parse_error
                        (s, "need code memory address as register", NULL, NULL);
                return 0;
            }

            if (!symbol(&s, "=")) {
                note_parse_error(s, "wcm has a mandatory =", NULL, NULL);
                return 0;
            }

            if (!parse_register(&s, &r)) {
                note_parse_error
                        (s, "need code memory value as register", NULL, NULL);
                return 0;
            }
            break;

        case '1':                           // rcm1 (read code memory part 1)
            opcode = oc_RCM1;
            if (!parse_register(&s, &l)) {
                note_parse_error
                        (s, "need code memory address as register", NULL, NULL);
                return 0;
            }
            break;

        case '2':                           // rcm1 (read code memory part 2)
            opcode = oc_RCM2;
            if (!parse_register(&s, &l)) {
                note_parse_error
                        (s, "need destination register", NULL, NULL);
                return 0;
            }
            break;

        default:
            ng("!keyword_statement %c", kw);
    }

    // grab the newline
    if (!newline(&s)) {
        note_parse_error(where, "bad syntax between keyword and end of line",
                NULL, NULL);
        return 0;
    }

    switch (kw) {
        case 'c':                           // crf
            *code = (long) oc_CRF << 27;
            break;

        case 'h':                           // halt
            *code = (long) oc_HALT << 27;
            break;

        case 'n':                           // nop
            *code = (long) oc_NOP << 27;
            break;

        case 'b': case 'm':                 // bound, cmp
            encode_instruction(where, kw == 'b' ? ocat_BOUND : ocat_CMP,
                    &l, &r, NULL, code);
            break;

        case 'i':                           // imm
            // Get the opcode and destination register right.
            encode_instruction(where, opcode, &l, &l, &l, code);

            // Overwrite 12 middle bits with the given value.
            *code = *code & 0777000000777L | imm_val << 9;
            break;

        case 'a': case 'j': case 'r':       // call, jump, return
            *code = (long) opcode << 27 | addr;
            break;

        case 'w':                           // wcm
            // Notation is C[r] = l, so operands are reversed here.
            encode_instruction(where, opcode, &r, &l, NULL, code);
            break;

        case '1':                           // rcm1
            encode_instruction(where, opcode, &l, NULL, NULL, code);
            break;

        case '2':                           // rcm2
            encode_instruction(where, opcode, NULL, NULL, &l, code);
            break;

    }

    *ss = s;
    return 1;
}
