// slots.h
// no dependencies

// ALU alpha layer operations
enum {
    // For these 16 enums, bits 1 and 0 specify signedness of the
    // left and right operands. So order and alignment matter here.
    alpha_cmp_uu,       // magnitude compare
    alpha_cmp_us,
    alpha_cmp_su,
    alpha_cmp_ss,
    alpha_add_uu,       // add
    alpha_add_us,
    alpha_add_su,
    alpha_add_ss,
    alpha_sub_uu,       // subtract
    alpha_sub_us,
    alpha_sub_su,
    alpha_sub_ss,
    alpha_rev_uu,       // reverse subtract
    alpha_rev_us,
    alpha_rev_su,
    alpha_rev_ss,
    alpha_end_signage,  // (placeholder)

    // The remaining enums don't have any particular meaning for
    // any of the bits. Their order is not important.
    alpha_and           // bitwise and
        = alpha_end_signage,
    alpha_or,           // bitwise or
    alpha_ml,           // multiply low
    alpha_mh,           // multiply high
    alpha_dsl,          // double shift left
    alpha_e36,          // left shift bit 36 encroached
    alpha_e35,          // left shift bit 35 or 36 encroached
    alpha_mix,          // forward S-box
    alpha_xim,          // inverse S-box
    alpha_uny,          // simple unary
    alpha_stu,          // stacked unary
    alpha_nud,          // nudge
    alpha_asr,          // arithmetic shift right

    // Magnitude comparison and certain boolean functions manage to 
    // use the same alpha slot (with theta ignoring propagate and
    // carry in the latter case). A second name is given here to
    // show the alternate intent in the instruction decoder table.
    alpha_boo           // boolean logic
        = alpha_cmp_uu,
};

// ALU beta layer operations
enum {
    beta_id = 0,        // identity
    beta_shl,           // shift left
    beta_shr,           // shift right
    beta_rol,           // rotate left
    beta_mix,           // forward S-box
    beta_xim,           // inverse S-box
    beta_uny,           // simple unary
    beta_stu,           // stacked unary
    beta_not,           // bitwise not
    beta_txo,           // transposed xor
    beta_rtl,           // rotate one tribble left (for mult)
    beta_pat,           // permute across tribbles
    beta_swz,           // freq. used swizzle operations
    beta_imm,           // swizzle for all load immediates
};

// ALU gamma layer operations
enum {
    // double-wides and symmetric pairs (T = 0, T = 1)
    gamma_mux = 0,      // mux (2 slots)
    gamma_mux_more,
    gamma_add,          // carry decision from theta (2 slots)
    gamma_add_more,
    gamma_kil,          // kill (2 slots)
    gamma_kil_more,
    gamma_stu,          // stacked unary (2 slots)
    gamma_stu_more,
    gamma_rol,          // rotate left (2 slots)
    gamma_rol_more,
    gamma_mix,          // forward S-box
    gamma_xim,          // inverse S-box
    gamma_imp,          // load immediate positive
    gamma_imn,          // load immediate negative 
    gamma_imh,          // load immediate high 18 bits
    gamma_imb,          // load immediate with high duplicating low

    // singletons
    gamma_uny,          // simple unary
    gamma_xor,          // bitwise xor
    gamma_pit,          // permute inside tribbles
};

// ALU theta (carry decision) operations
enum {
    // double-wides and symmetric pairs (T = 0, T = 1)
    theta_add = 0,      // add carry decisions
    theta_add_more,
    theta_rev,          // reverse add
    theta_rev_more,
    theta_dsl,          // DSL's rotated add-with-carry
    theta_dsl_more,
    theta_lt,           // 111111 if L < R else 000000
    theta_gt,           // 111111 if L > r else 000000
    theta_2l,           // set all d to left of any set c
    theta_2r,           // set all d to right of any set c
    theta_0,            // set all d 0
    theta_1,            // set all d 1

    // singletons
    theta_c5,           // copy c5 to all d
};

// ALU zeta (flag) operations
enum {
    zeta_logic,         // non-numeric: Z and N come from gamma
    zeta_cmp,           // N comes from alpha, except for overflow
    zeta_add_s,         // addition with signed result
    zeta_add_u,         // addition with unsigned result
    zeta_bound,         // interrupt if array index out of bounds
    zeta_asl_uu,        // arithmetic shift left flag cases
    zeta_asl_us,
    zeta_asl_su,
    zeta_asl_ss,
    zeta_asr_uu,        // arithmetic shift right flag cases
    zeta_asr_us,
    zeta_asr_su,
    zeta_asr_ss,
    zeta_clear_r,       // clear R flag after copying it to T
    zeta_wrap,          // addition with wrapping allowed
    zeta_t_adj,         // alter T based on p_i and c_i
};

// ALU stacked unary operation names
enum {
    su_abs2s,           // abs value: negation step to signed
    su_abs2i,           // abs value: identity step (no bit 35)
    su_abs2u,           // abs value: negation step to unsigned
    su_abs1s,           // {0} if negative else {1} for abs value
    su_abs1u,           // {2} if negative else {1} for abs value
    su_fabs,            // |x|, only works when |x| < 2**30, checked
    su_signum,          // full signum
    su_lsprep,          // uint to left shift control word
    su_rsprep,          // uint to right shift control word
    su_lrprep,          // uint to left rotate control word
    su_rrprep,          // uint to right rotate control word
    su_ham0,            // 3-cycle popcount, count 0s, cycle 1
    su_ham1,            // 3-cycle popcount, count 1s, cycle 1
    su_ham2,            // 3-cycle popcount, count either, cycle 2
    su_llz,             // light leading 0s for CLZ
    su_llo,             // light leading 1s for CLO
    su_ltz,             // light trailing 0s for CTZ
    su_lto,             // light trailing 1s for CTO
    su_parity,          // parity, result in 2**0 position only
    su_increv,          // inc bit-reversed word mod 2**36. no T
    su_decrev,          // dec bit-reversed word mod 2**36. no T
    su_lfsr,            // 36-bit LFSR
    su_tpoly,           // xor XFSR poly if T is set for 72-bit LFSR
    su_rcl,             // shift/rotate one place left into/thru T
    su_rcr,             // shift/rotate one place right into/thru T
};

// ALU simple unary operation names for 6-bit tribbles
enum {
    un_zt,              // 000000
    un_ot,              // 000001
    un_oxt,             // 111111
    un_idt,             // identity
    un_nott,            // bitwise NOT
    un_iszt,            // 000001 if tribble = 0 else 000000
    un_isnzt,           // 000001 if tribble <> 0 else 000000
    un_iszxt,           // 111111 if tribble = 0 else 000000
    un_isnzxt,          // 111111 if tribble <> 0 else 000000
    un_czt,             // count zeros in tribble
    un_clzt,            // count leading zeros in tribble
    un_ctzt,            // count trailing zeros in tribble
    un_cot,             // count ones in tribble
    un_clot,            // count leading ones in tribble
    un_ctot,            // count trailing ones in tribble
    un_inct,            // increment tribble with wrap
    un_dect,            // decrement tribble with wrap
    un_recipt,          // floor( (2**36 - 1) / max(tribble, 1) )
    un_revt,            // reverse bits
};
