/* alpha SRAM definition

INPUT BITS

    0-5     left input
    6-11    right input
    12-17   operation

OUTPUT BITS

    0-5     result
    6       carry
    7       propagate
    8-12    a5 only: encoded ranges / ASL / ASR / MH

RANGE ENCODING

    Sum of two alternatives based on carry decision 5 clear or set:

    clr set     result
    ------------------------------
     0   0      too negative
     1   5      fits signed only
     2  10      fits always
     3  15      fits unsigned only
     4  20      too positive

    Above applies to fs_add_u, fs_add_s, fs_cmp, fs_bound.

    Although this 5-bit encoding collapses to 9 cases and would fit
    in 4 bits for addition and subtraction, the same field is used
    for MAX and MIN, where there remain 25 scenarios. So we need
    all five bits.

ASL AND ASR ENCODING

    Sum of:

    1 (bit 8)   L35 is set
    2 (bit 9)   R = 0

    C   for ASR, 0 encroaches bit 35 (e) or 36 (E)
    P   for ASR, 1 encroaches bit 35 (e) or 36 (E)

MH ENCODING

    0               no overflow
    3 (bits 8 & 9)  short mul exceeds 36 bits

*/

// Consider high tribble of a word for range encoding.
// If d5 is set, result must be multiplied by 5.
static int encode_range(int q)
{
    if (q < -32) return 0;      // too  negative
    if (q <   0) return 1;      // fits signed only
    if (q <  32) return 2;      // fits always
    if (q <  64) return 3;      // fits unsigned only
    return 4;                   // too  positive
}

/* compute a single value in the "regular" (slots 0-15) alpha table */
static int alpha_operation(int slot, int subword, int l, int r)
{
    int al = 0, t, lim = 35, bit, ign = 0;

    switch (slot) {
        case alpha_cmp_uu:                  // a0-a4 with a5_operation()
        case alpha_asr:                     // a0-a5

            // For a0-a4, do a magnitude comparison.
            // a5's magnitude comparison is in its high-memory section.
            al = l;
            if (subword < 5) {
                if (l < r)
                    al |= a_carry_bit;
                else if (l > r)
                    al |= a_prop_bit;
            }

A5:         if (subword == 5) {
                if (l & bit5) {
                    al |= a_carry_bit;      // theta: ASR.su/ss sign extension
N:                  al |= bit8;             // flags: L35 was set for ASR
                }
                if (!r)
                    al |= bit9;             // flags: R == 0 for ASR
            }
            break;

        case alpha_rev_uu:                  // reverse subtract
            l ^= r, r ^= l, l ^= r;
            nobreak;

        case alpha_sub_uu:
            r = 63 & ~r;
            nobreak;

        case alpha_add_uu:
            al = l + r;
            if (al == 63)
                al |= a_prop_bit;
            break;

        case alpha_and:
            al = l & r;
            break;

        case alpha_or:
            al = l | r;
            break;

        case alpha_ml:
            al = 63 & l * r;
            break;

        case alpha_mh:
            al = 63 & l * r >> 6;
            if (al && subword == 5)
                al |= bit8 | bit9;      // fs_fsr_us works for this
            break;

        case alpha_dsl:                 // DSL only
            al = subword == 5 ? r : l;
            if (al == 63)
                al |= a_prop_bit;
            break;

        case alpha_e36:                 // left shift bit 36 encroached?
            lim = 36;
            nobreak;

        case alpha_e35:                 // left shift bit 35/36 encroached?
            al = l;
            for (bit=0; bit<6; ++bit) {
                if (r < lim - 6 * subword - bit)
                    continue;           // no encroachment of bit 35/36
                if (al & 1 << bit)
                    al |= a_prop_bit;   // encroached by 1
                else
                    al |= a_carry_bit;  // encroached by 0
            }
            if (r > lim)
                al |= a_carry_bit;      // encroached by 0 right of radix point
            if (subword == 5 && l & bit5)
                al |= bit8;             // flags: L35 was set for ASL
            break;

        case alpha_mix:
            al = sbox_unquote(alpha_sbox[subword][r][l]);
            break;

        case alpha_xim:
            al = sbox_invert(gamma_sbox[subword][r], l);
            break;

        case alpha_uny:
            al = simple_unary(subword, l, r);
            break;

        case alpha_stu:
            stacked_unary(0, subword, l, r, &al, &ign, &ign);
            break;

        case alpha_nud:                 // nudge ... gonna use it?
            for (t = 32; t; t >>= 1)
                if (r & t) {
                    l |= a_carry_bit;
                    --t;
                    break;
                }
            al = l & ~t | r & t;
            break;

        default:
            break;                      // all zero in unallocated slots
    }

    return al;
}

/* single value in sign-handling alpha5 RAM */
static int alpha5_operation(int slot, int l, int r, int l_signed, int r_signed)
{
    int al = 0, range = 0;

    // Extend signs of any negative inputs.
    if (l_signed && l & 32)
        l |= ~63;
    if (r_signed && r & 32)
        r |= ~63;

    switch (slot) {

        case alpha_cmp_uu:                  /* signed comparison */
            al = l & 63;                    /* un-extend sign in result */
            if (l < r)                      /* compute comparison result */
                al |= a_carry_bit;
            else if (l > r)
                al |= a_prop_bit;

            /* overflow and negative tests (used by MIN and MAX) */
            range = 5 * encode_range(r) + encode_range(l);
            break;

        case alpha_rev_uu:                  /* reverse subtract */
            l ^= r, r ^= l, l ^= r;
            nobreak;

        case alpha_sub_uu:                  /* subtract */
            r = ~r;                         /* method of complements */
            nobreak;

        case alpha_add_uu:
            al = l + r;

            /* overflow and negative tests */
            range = 5 * encode_range(al + 1) + encode_range(al);

            al &= 63;
            if (al == 63)                   /* is carry uncertain? */
                al |= a_prop_bit;           /* yes, carries propagate */
            break;

        default:
            ng("alpha5_operation: what is %i?\n", slot);
    }

    return al | range << 8;
}

static int alpha_calc(int subword, int i)
{
    int l, r, slot, l_signed, r_signed;

    if (subword < 0 || subword > 5)
        ng("alpha_calc");

    l = i & 63;
    r = i >> 6 & 63;
    slot = i >> 12 & 63;

    if (slot < alpha_end_signage) {
        l_signed = slot & 2;
        r_signed = slot & 1;
        slot &= ~3;
    
        if (subword == 5)
            return alpha5_operation(slot, l, r, l_signed, r_signed);
    }

    return alpha_operation(slot, subword, l, r);
}

// For the ALU console, convert an alpha slot enum into the old
// character code.
int alpha_emoji(int slot)
{
    static struct {
        int i;
        int c;
    } moji[] = {
        alpha_cmp_uu, '=',
        alpha_cmp_us, '=',
        alpha_cmp_su, '=',
        alpha_cmp_ss, '=',
        alpha_add_uu, '+',
        alpha_add_us, '+',
        alpha_add_su, '+',
        alpha_add_ss, '+',
        alpha_sub_uu, '-',
        alpha_sub_us, '-',
        alpha_sub_su, '-',
        alpha_sub_ss, '-',
        alpha_rev_uu, 'm',
        alpha_rev_us, 'm',
        alpha_rev_su, 'm',
        alpha_rev_ss, 'm',
        alpha_and, '&',
        alpha_or, '|',
        alpha_ml, 'x',
        alpha_mh, 'X',
        alpha_dsl, '*',
        alpha_e36, 'E',
        alpha_e35, 'e',
        alpha_mix, 's',
        alpha_xim, 'S',
        alpha_uny, 'u',
        alpha_stu, 'w',
        alpha_nud, 'n',
        -1, -1,
    };

    int i;

    if (slot < alpha_end_signage)
        slot &= ~3;             // remove a5uu / a5us / a5su / a5ss

    for (i = 0; moji[i].i >= 0; i++)
        if (slot == moji[i].i)
            return moji[i].c;

    return '$';                 // This is an error.
}

/* FOOTNOTES

A5: a5.= in low memory is used for bitwise boolean, LSL, LSR, ROL, PIT, PAT,
    COPY, LEAD, TRAIL, and ASR. Of these, generated propagate and carries are
    ignored except for LEAD, TRAIL, and ASR. LEAD and TRAIL present a conflict,
    so these opcodes are being reimplemented as stacked unary, which won't use
    a5.=. That leaves the propagate and carry signals from a5.= available for
    exclusive use by ASR.

N:  L is negative if we're doing ASR.ss or ASR.su.
    Setting carry only affects theta.n, used only by ASR.ss and ASR.su.
    This carry is the sign bit (l & 32), copied by theta to all gamma.l
    so the sign can be extended the correct number of bits in gamma.

*/
