/*
    audited.c

    These routines are not regression tested, and may be used both as
    implementation and verification of the same code. Their correctness
    was certified by Marc. Code in this module should not be revised
    lightly.
*/

// Mirror a 6-bit subword.
// This table was not generated by hand. It is free of errors.
// MWA ok 29Sep2020
static const char rev_bits[64] = { 0, 32, 16, 48, 8, 40, 24, 56, 4, 36,
    20, 52, 12, 44, 28, 60, 2, 34, 18, 50, 10, 42, 26, 58, 6, 38, 22, 54,
    14, 46, 30, 62, 1, 33, 17, 49, 9, 41, 25, 57, 5, 37, 21, 53, 13, 45,
    29, 61, 3, 35, 19, 51, 11, 43, 27, 59, 7, 39, 23, 55, 15, 47, 31, 63 };

// Count ones in a 6-bit subword.
// This table was not generated by hand. It is free of errors.
// MWA ok 29Sep2020
static const char count_ones[64] = { 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3,
    2, 3, 3, 4, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 1, 2, 2,
    3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4,
    4, 5, 4, 5, 5, 6 };

// Generalized function to count 0s or 1s in a tribble
// MWA ok 29Sep2020
static int count_bits(int n, int use_zeros, int trailing, int reverse_order)
{
    if (use_zeros)
        n = ~n;             // count zeros instead of ones
    n &= 63;                // array sanity
    if (reverse_order)
        n = rev_bits[n];    // leading rather than trailing
    if (trailing)
        n = n & ~(n+1);     // set zero, except for any trailing ones
    return count_ones[n];
}

// Simple unary tables common to alpha, beta, and gamma
// XXX This is, in a sense, a solution looking for a problem.
// XXX Why beta? Why both alpha and gamma? Subject to reallocation.
// MWA ok 29Sep2020
static int simple_unary(int subword, int l, int r)
{
    int u = 0;

    switch (r) {
        case un_zt  :       // 000000
            u = 0;
            break;

        case un_ot:         // 000001
            u = 1;
            break;

        case un_oxt:        // 111111
            u = 63;
            break;

        case un_idt:        // identity
            u = l;
            break;

        case un_nott:       // bitwise NOT
            u = ~l & 63;
            break;

        case un_iszt:       // 000001 if tribble = 0 else 000000
            u = !l;
            break;

        case un_isnzt:      // 000001 if tribble <> 0 else 000000
            u = !!l;
            break;

        case un_iszxt:      // 111111 if tribble = 0 else 000000
            u = l ? 0 : 63;
            break;

        case un_isnzxt:     // 111111 if tribble <> 0 else 000000
            u = l ? 63 : 0;
            break;

        case un_czt:        // count zeros in tribble
            u = count_bits(l, 1, 0, 0);
            break;

        case un_clzt:       // count leading zeros in tribble
            u = count_bits(l, 1, 1, 1);
            break;

        case un_ctzt:       // count trailing zeros in tribble
            u = count_bits(l, 1, 1, 0);
            break;

        case un_cot:        // count ones in tribble
            u = count_bits(l, 0, 0, 0);
            break;

        case un_clot:       // count leading ones in tribble
            u = count_bits(l, 0, 1, 1);
            break;

        case un_ctot:       // count trailing ones in tribble
            u = count_bits(l, 0, 1, 0);
            break;

        case un_inct:       // increment tribble with wrap
            u = (l + 1) & 63;
            break;

        case un_dect:       // decrement tribble with wrap
            u = (l - 1) & 63;
            break;

        case un_recipt:     // floor( (2**36 - 1) / max(tribble, 1) )
            u = mask36 / (l ? l : 1) >> 6 * subword & 63;
            break;

        case un_revt:       // reverse bits
            u = rev_bits[l & 63];
            break;
    }

    return u;
}

// Reverse bits in a subword.
// MWA ok 29Sep2020
static int reverse_six_bits(int x)
{
    return rev_bits[x & 63];
}
