/* beta SRAM definition

INPUT BITS

    0-5     transposed input from a0..a5
    6-11    right input
    12-17   operation

OUTPUT BITS

    0-5     transposed output to g0..g5

*/

static int beta_operation(int slot, int subword, int l, int r)
{
    int b = 0, bit, mask, ign = 0;
    int sign_ext = 0;

    if (subword < 0 || subword > 5)
        ng("beta_operation");

    /* see "beta operations" table from PDF */
    switch (slot) {
        case beta_id:
            b = l;
            break;

        case beta_shl:
            for (mask = bit = 0; bit < 6; ++bit)
                if (r <= 6 * (5 - bit) + 5 - subword)
                    mask |= 1 << bit;

            l &= mask;
            nobreak;

        case beta_rol:                          /* rotate word left */
l:          r %= 36;                            /* word rotation, 0-35 */
            r = (36 + r + subword) / 6 % 6;     /* subword's rotation, 0-5 */
            b = rotate_tribble_left(l, r);      /* rotated subword */
            break;

        case beta_shr:
            if (!r)                             /* no shift? */
                goto l;
            for (mask = bit = 0; bit < 6; ++bit)
                if (r < 36 && r >= 36 - 6 * bit - subword)
                    mask |= 1 << bit;
            l = l & mask | sign_ext & ~mask;
            goto l;

        case beta_mix:
            b = sbox_unquote(beta_sbox[subword][r][l]);
            break;

        case beta_xim:
            b = sbox_invert(beta_sbox[subword][r], l);
            break;

        case beta_stu:
            stacked_unary(0, subword, l, r, &ign, &b, &ign);
            break;

        case beta_not:
            b = 63 & ~l;
            break;

        case beta_txo:
            b = 63 & (l ^ r);
            break;

        case beta_rtl:
            b = 63 & l << 1 | l >> 5;
            break;

        case beta_pat:
            b = permute(r, l);
            break;

        case beta_swz:
            /* XXX only the first 6 swizzles are defined */
            if (r < 6)
                b = l & 1 << r ? 63 : 0;
            break;

        case beta_uny:      /* transposed simple unary (WHAT FOR?!?) */
            b = simple_unary(subword, l, r);
            break;

        case beta_imm:      // load immediate swizzle
            if (subword < 3)
                // -ABC-- becomes ABCABC
                b = 9 * (7 & l >> 2);
            else
                // --ABC- becomes ABCABC
                b = 9 * (7 & l >> 1);
            break;

        default:            /* unallocated slots all zero */
            break;
    }

    return b;
}

static int beta_calc(int subword, int i)
{
    int l, r, slot;

    l = i & 63;
    r = i >> 6 & 63;
    slot = i >> 12 & 63;

    return beta_operation(slot, subword, l, r);
}

// For the ALU console, convert an beta slot enum into the old
// character code.
int beta_emoji(int slot)
{
    static struct {
        int i;
        int c;
    } moji[] = {
        beta_id , '=',
        beta_shl, 'L',
        beta_shr, 'R',
        beta_rol, 'l',
        beta_mix, 's',
        beta_xim, 'S',
        beta_uny, 'u',
        beta_stu, 'w',
        beta_not, '!',
        beta_txo, '^',
        beta_rtl, '*',
        beta_pat, 'p',
        beta_swz, 'z',
        beta_imm, 'i',
        -1, -1,
    };

    int i;

    for (i = 0; moji[i].i >= 0; i++)
        if (slot == moji[i].i)
            return moji[i].c;

    return '$';                 // This is an error.
}
