/*
    decoder.c
    Using the instruct.c database, fill in the instruction decoder SRAMs.
    The various CPU cycle types are implemented below.
*/

/*
    The decoder RAMs are 512 instructions * 4 steps, and are 72 bits
    (2 36-bit RAMs) wide. The following constants are too large for
    some compilers' enum keyword, so they are macros here. By or-ing
    these together appropriately, we form instruction decoder words.
    For what these all mean, refer to the architecture documentation.
*/
#define V(bit, val) (val ## L << bit)

#define lo_a_A  V(11, 0)
#define lo_a_m  V(11, 3)
#define lo_a_P  V(11, 1)
#define lo_a_Pw V(11, 0)
#define lo_a_Aw V(11, 0)
#define lo_a_Ah V(11, 2)

#define lo_b_init V(13, 0)
#define lo_b_B    V(13, 0)          // XXX see question on spreadsheet here
#define lo_b_Mw   V(13, 2)
#define lo_b_M    V(13, 1)
#define lo_b_io   V(13, 3)

#define lo_c_C    V(15, 0)
#define lo_c_Cw   V(15, 2)
#define lo_c_dest V(15, 1)

#define lo_e_E   V(0, 1)            // BEWARE: lo_e_* must go to D1, never D0.
#define lo_e_Eh  V(0, 3)
#define lo_e_Ews V(0, 0)
#define lo_e_Ewe V(0, 2)

#define lo_g_G    V(17, 0)
#define lo_g_Gn   V(17, 2)
#define lo_g_LR   V(17, 1)
#define lo_g_LRh  V(17, 3)

#define lo_ip_priv   V(19, 0)
#define lo_ip_return V(19, 2)
#define lo_ip_next   V(19, 1)
#define lo_ip_jump   V(19, 1)

#define lo_s_S  V(34, 1)
#define lo_s_Sh V(34, 3)
#define lo_s_Sw V(34, 0)

#define lo_z_Z    V(33, 0)
#define lo_z_Zh   V(33, 1)
#define lo_z_init V(33, 0)

#define save_inst V(31, 1)

#define branch_T  V(22, 1)
#define branch_R  V(22, 2)
#define branch_N  V(22, 4)
#define branch_Z  V(22, 8)
#define branch_P  V(22, 16)
#define branch_RT_flip V(22, 32)

#define latch_jump_conds V(30, 1)

#define alpha_beta_gamma V(0, 1)

#define theta_zeta V(6, 1)

#define t_flip_allow V(28, 1)

#define p_no_write V(21, 1)

#define preset_call_depth V(2, 1)   // BEWARE: must go to D1, never D0.

#define store_n_z_flag V(3, 1)      // BEWARE: must go to D1, never D0.

/*
    Given a pointer to a 4-word instruction slot within decoder SRAM D0,
    fill in the firmware for that opcode. This is the more-full of the
    two decoder RAMs.

    Many instructions have lo_ip_jump at click 3 in order to have
    ff a's clock low. This lets us snatch the desired code addresss in
    ff a via a rising clock in click 0 for instructions like wcm (write
    code memory).
*/
static void fill_inst_0(long *m, struct instruction *in)
{
    long conds = 0;                     // branch conditions

    // Here we fill in the decoder bits for control and sequencing.
    switch (in->cycle) {
        default:
            ng("!the_insts insane");

        case cy_alu:    // ALU operation
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | p_no_write | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | p_no_write | lo_s_S;
            m[2] = lo_z_Z | lo_g_G | lo_c_C | p_no_write | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | save_inst | p_no_write | lo_s_S;
            break;

        case cy_now:    // no-write ALU operation
            m[0] = lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | p_no_write | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | p_no_write | lo_s_S;
            m[2] = lo_z_Z | lo_g_Gn | lo_c_C | p_no_write | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | save_inst | p_no_write | lo_s_S;
            break;

        case cy_j:      // jump
            switch (in->opcode) {
                case oc_J:
                    conds = branch_N | branch_Z | branch_P;
                    break;

                case oc_J_lt:
                    conds = branch_N;
                    break;

                case oc_J_le:
                    conds = branch_N | branch_Z;
                    break;

                case oc_J_eq:
                    conds = branch_Z;
                    break;

                case oc_J_ne:
                    conds = branch_N | branch_P;
                    break;

                case oc_J_ge:
                    conds = branch_Z | branch_P;
                    break;

                case oc_J_gt:
                    conds = branch_P;
                    break;

                case oc_J_r:
                    conds = branch_R;
                    break;

                case oc_J_nr:
                    conds = branch_R | branch_RT_flip;
                    break;

                case oc_J_t:
                    conds = branch_T;
                    break;

                case oc_J_nt:
                    conds = branch_T | branch_RT_flip;
                    break;

                default:
                    ng("!bad decoder jump %i %s",
                            in->opcode, opcode_name(in->opcode));
            }

            m[0] = conds | lo_z_Zh | lo_g_LRh 
                 | lo_c_dest | p_no_write | lo_s_S;
            m[1] = lo_z_Zh | lo_g_LRh | lo_c_dest 
                 | lo_ip_jump | p_no_write | lo_s_S;
            m[2] = lo_z_Zh | lo_g_LRh | lo_c_C | p_no_write | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | save_inst | p_no_write | lo_s_S;
            break;

        case cy_c:      // call
            break;

        case cy_r:      // return
            break;

        case cy_imm:    // load immediate / privileged read code memory 2
            m[0] = lo_a_m | lo_z_Zh | lo_g_LRh | lo_c_dest | p_no_write | lo_s_S;
            m[1] = lo_a_A | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_dest | lo_ip_next | p_no_write | lo_s_S;
            m[2] = lo_z_Z | lo_g_G | lo_c_C | p_no_write | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | save_inst | p_no_write | lo_s_S;
            break;

        case cy_ld:     // load memory (honors page table)
            break;

        case cy_sto:    // store memory (honors page table)
            break;

        case cy_hlt:    // privileged halt
            break;

        case cy_wcm:    // privileged write code memory
            // This derives from cy_now.
            m[0] = lo_ip_priv | lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | p_no_write | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_Cw | lo_ip_next | p_no_write | lo_s_S;
            m[2] = lo_z_Z | lo_g_Gn | lo_c_C | p_no_write | lo_s_S;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | save_inst | p_no_write | lo_s_S;
            break;

        case cy_rcm:    // privileged read code memory 1
            // This derives from cy_wcm.
            m[0] = lo_ip_priv | lo_a_A | lo_z_Zh | lo_g_LRh | lo_c_dest | p_no_write | lo_s_S;
            m[1] = lo_a_Ah | lo_b_B | lo_z_Zh 
                 | lo_g_LRh | lo_c_C | lo_ip_next | p_no_write | lo_s_S;
            m[2] = lo_z_Z | lo_g_Gn | lo_c_C | p_no_write | lo_s_S | save_inst;
            m[3] = lo_ip_jump | lo_z_Zh | lo_g_LR | lo_c_dest | p_no_write | lo_s_S;
            break;

        case cy_wdm:    // privileged write data memory (bypasses page table)
            break;

        case cy_rdm:    // privileged read data memory (bypasses page table)
            break;

        case cy_wpt:    // privileged write page table
            break;

        case cy_rpt:    // privileged read page table
            break;
    }

    // Now we fill in operations for the ALU SRAMs.
    // We also indicate how the incoming T flag is seen by the ALU.
    // The cycle when t_flip_allow is inserted affects how alu.c is written,
    // so watch for that if you make changes.
    // XXX There is probably a simpler way than this / 2 scheme for
    //     understanding how theta works and its slots. Same for gamma.
    //     An issue is that in gamma's case, the hardware overwrites
    //     bit 0 of the operation with the carry decision, but with
    //     theta, there are 5 bits in the control decoder (not 6), so
    //     we shift them left to insert the adjusted T flag.
    m[0] |= alpha_beta_gamma * in->alpha;
    m[1] |= alpha_beta_gamma * in->beta  | theta_zeta * (in->theta / 2) | t_flip_allow * in->t_vision;
    m[2] |= alpha_beta_gamma * in->gamma | theta_zeta * in->zeta;

    // In order to NOT jump, ff l needs five zeros clocked in.
    // So we have to hit that clock on every instruction.
    // We also need to clock in non-zeros when we're jumping.
    m[0] |= latch_jump_conds;
}

/*
    Given a pointer to a 4-word instruction slot within decoder SRAM D1,
    fill in the firmware for that opcode. This is the less-full of the
    two decoder RAMs.

    XXX Little is here yet, but when I return to testing (especially
    flag handling and later call stack), this will find use.
*/
static void fill_inst_1(
long *m, struct instruction *in)
{
    switch (in->cycle) {
        default:
            ng("!the_insts insane");

        case cy_alu:    // ALU operation
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E | store_n_z_flag;
            break;

        case cy_now:    // no-write ALU operation
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E | store_n_z_flag;
            break;

        case cy_j:      // jump
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_c:      // call
            break;

        case cy_r:      // return
            break;

        case cy_imm:    // load immediate / privileged read code memory 2
            // XXX We have open questions on N and Z immediate semantics.
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E | store_n_z_flag;
            break;

        case cy_ld:     // load memory (honors page table)
            break;

        case cy_sto:    // store memory (honors page table)
            break;

        case cy_hlt:    // privileged halt
            break;

        case cy_wcm:    // privileged write code memory
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_rcm:    // privileged read code memory 1
            m[0] = lo_e_E;
            m[1] = lo_e_E;
            m[2] = lo_e_E;
            m[3] = lo_e_E;
            break;

        case cy_wdm:    // privileged write data memory (bypasses page table)
            break;

        case cy_rdm:    // privileged read data memory (bypasses page table)
            break;

        case cy_wpt:    // privileged write page table
            break;

        case cy_rpt:    // privileged read page table
            break;
    }
}

/*
    This loads instruction decoder SRAM D0 or D1.
    The caller already allocated memory.
*/
static void load_decoder(long *mem, int rows, int chip)
{
    void (*filler)(long *mem, struct instruction *in) =
            chip ? fill_inst_1 : fill_inst_0;
    static struct instruction *nop;
    int i;

    if (rows != 4 * 512)
        ng("!decoder.c and firmware.c conflict");

    // To preclude uninitialized parts of the decoder, pre-fill
    // the entire table with NOP instructions.
    if (!nop) {
        for (i = 0; i < CARD(the_insts); i++)
            if (the_insts[i].opcode == oc_NOP) {
                nop = the_insts + i;
                goto ok;
            }
        ng("no NOP instruction?");
ok:
        for (i = 0; i < 512; i++)
            filler(mem + 4 * i, nop);
    }

    // Loop through the instructions that are in fact defined, and fill those.
    for (i = 0; i < CARD(the_insts); i++)
        filler(mem + 4 * the_insts[i].opcode, the_insts + i);
}

#undef V
