/*  firmware.c

    This module provides a single means of allocating firmware memory
    that can be used across both simulations. Names for the RAMs are
    standardized in order to find the firmware.

    To keep the namespace small and units encapsulated, this translation
    unit only exposes the get_firmware() function, but #includes quite
    a few C source files with static linkage. These modules compute
    the firmware contents.
*/

// S-boxes are a separate .o because they contain a lot of data
// that on the surface looks like a lot of code, and I don't want
// my .o files for code to look bloated. The downside is that
// the S-boxes now have otherwise unnecessary external linkage.
extern char *alpha_sbox[6][64], *beta_sbox[6][64], *gamma_sbox[6][64];

#include <string.h>     // strchr()
#include <stdlib.h>     // calloc()

#include "../misc/misc.h"
#include "../consts/opcode.h"
#include "../consts/slots.h"

#include "audited.c"    // simple unary ops (nothing here is regression tested)
#include "unary.c"      // stacked unary ops
#include "pieces.c"     // misc. permutation, S-box, and rotation aids
#include "alpha.c"      // ALU alpha layer firmware
#include "beta.c"       // ALU beta layer firmware
#include "gamma.c"      // ALU gamma layer firmware
#include "theta.c"      // ALU theta (carry decision) firmware
#include "zeta.c"       // ALU zeta (flag handling) firmware
#include "instruct.c"   // table of CPU and ALU instructions
#include "decoder.c"    // instruction decoder firmware
#include "uninit.c"     // uninitialized chips (non-firmware)

/*
    By sheer chance, the 18-wide SRAMs are easily computed a single
    address at a time. Here's how we find the code for that.
*/
static struct narrow {
    char *name;         // name of SRAM chip
    int rows;           // number of words on chip
    int subword;        // passed to calc() to identify chip within a family
    int *mem;           // caches firmware for reuse
    int (*calc)         // computes contents at any address
        (int subword, int addr);
} narrows[] = {
    "alpha0", 1 << 18, 0, NULL, alpha_calc,
    "alpha1", 1 << 18, 1, NULL, alpha_calc,
    "alpha2", 1 << 18, 2, NULL, alpha_calc,
    "alpha3", 1 << 18, 3, NULL, alpha_calc,
    "alpha4", 1 << 18, 4, NULL, alpha_calc,
    "alpha5", 1 << 18, 5, NULL, alpha_calc,
    "beta0" , 1 << 18, 0, NULL, beta_calc,
    "beta1" , 1 << 18, 1, NULL, beta_calc,
    "beta2" , 1 << 18, 2, NULL, beta_calc,
    "beta3" , 1 << 18, 3, NULL, beta_calc,
    "beta4" , 1 << 18, 4, NULL, beta_calc,
    "beta5" , 1 << 18, 5, NULL, beta_calc,
    "gamma0", 1 << 18, 0, NULL, gamma_calc,
    "gamma1", 1 << 18, 1, NULL, gamma_calc,
    "gamma2", 1 << 18, 2, NULL, gamma_calc,
    "gamma3", 1 << 18, 3, NULL, gamma_calc,
    "gamma4", 1 << 18, 4, NULL, gamma_calc,
    "gamma5", 1 << 18, 5, NULL, gamma_calc,
    "theta" , 1 << 18, 0, NULL, theta_calc,
    "zeta"  , 1 << 18, 0, NULL, zeta_calc,
    "E"     , 1 << 18, 0, NULL, uninit_stack_depths,
    "P"     , 1 << 18, 0, NULL, uninit_page_table,
    NULL    ,       0, 0, NULL,
};

/*
    Also by sheer chance, the 36-wide SRAMs are easiest to compute
    in their entirety. Here's how we find the code for that.
*/
static struct wide {
    char *name;         // name of SRAM chip
    int rows;           // number of words on chip
    int chip;           // passed to calc() to identify chip within a family
    long *mem;          // caches firmware for reuse
    void (*calc)        // computes firmware for entire chip in one call
        (long *mem, int rows, int chip);
} wides[] = {
    "D0", 4 * 512, 0, NULL, load_decoder,
    "D1", 4 * 512, 1, NULL, load_decoder,
    "M0", 65536  , 0, NULL, uninit_data_ram,
    "M1", 65536  , 1, NULL, uninit_data_ram,    // TODO support uninstalled chip
    "C",  65536  , 0, NULL, uninit_code_ram,
    "L",  1 << 17, 0, NULL, uninit_register_file,
    "R",  1 << 17, 1, NULL, uninit_register_file,
    "S",  1 << 17, 0, NULL, uninit_return_stack,
    NULL, 0, 0, NULL,
};

/*
    Allocate and populate firmware memory.
    This serves either simulator (functional VM or netlist emulation).
    At present, it's permissible to call free() on the returned memory.
*/
void get_firmware(char *name, int *width, int *rows, void **mem)
{
    struct narrow *n;
    struct wide *w;
    int i;

    for (n = narrows; n->name; ++n)
        if (!strcmp(n->name, name))
            goto narrow;

    for (w = wides; w->name; ++w)
        if (!strcmp(w->name, name))
            goto wide;

    ng("!no firmware for '%s'", name);

narrow:
    if (n->mem) {
        *mem = n->mem;
        return;
    }
    n->mem = trackmem(NULL, n->rows * sizeof(uint));
    for (i = 0; i < n->rows; i++)
        n->mem[i] = n->calc(n->subword, i);
    if (width)
        *width = 18;
    if (rows)
        *rows = n->rows;
    if (mem)
        *mem = (void *) n->mem;
    return;

wide:
    if (w->mem) {
        *mem = w->mem;
        return;
    }
    w->mem = trackmem(NULL, w->rows * sizeof(ulong));
    if (width)
        *width = 36;
    if (rows)
        *rows = w->rows;
    if (mem)
        *mem = (void *) w->mem;
    w->calc(w->mem, w->rows, w->chip);
}

// This will release all firmware memory. Be sure not to reference afterward.
void discard_all_firmware(void)
{
    struct narrow *n;
    struct wide *w;

    for (n = narrows; n->name; ++n)
        if (n->mem)
            trackmem(n->mem, 0);

    for (w = wides; w->name; ++w)
        if (w->mem)
            trackmem(w->mem, 0);
}
