/* gamma SRAM definition

INPUT BITS

    0-5     left input
    6-11    right input
    12      carry decision from theta
    13-17   operation

OUTPUT BITS

    0-5     result on node 2 (save to left register file)
    6-11    result on node 5 (save to right register file)
    12      set iff result is zero (for zero-detect NAND)
    13      copy of gamma35 in subword 5 (reduces capacitance)

*/

static int gamma_operation(int slot, int subword, int l, int r)
{
    int g = 0, ign = 0;
    int carry = slot & 1;

    if (subword < 0 || subword > 5)
        ng("gamma_operation");

    /* see "gamma operations" table from PDF */
    switch (slot) {
        case gamma_mux:
        case gamma_mux_more:
            g = carry ? r : l;
            break;

        case gamma_add:
        case gamma_add_more:
            g = carry ? 63 & l + 1 : l;
            break;

        case gamma_kil:
        case gamma_kil_more:
            g = carry ? 0 : l;
            break;

        case gamma_stu:
        case gamma_stu_more:
            stacked_unary(carry, subword, l, r, &ign, &ign, &g);
            break;

        case gamma_rol:
        case gamma_rol_more:
            g = rotate_tribble_left(l, r % 6);
            if (carry) {
                /* extend - sign from right. */
                if (!r) 
                    break;          /* left shift of 0 bits: identity */
                if (r > 35)
                    r = 0;          /* left shift more than 35 bits: all 1s */
                g |= 63 & ((~((1L << r) - 1) & mask36) >> 6 * subword);
            }
            break;

        /* ABOVE use two slots each, and require T input. */
        /* BELOW use one slot each, using T to select odd or even slot. */

        case gamma_mix:
            g = sbox_unquote(gamma_sbox[subword][r][l]);
            break;

        case gamma_xim:
            g = sbox_invert(alpha_sbox[subword][r], l);
            break;

        case gamma_uny:
            g = simple_unary(subword, l, r);
            break;

        case gamma_xor:
            g = 63 & (l ^ r);
            break;

        case gamma_pit:
            g = permute(r, l);
            break;

        case gamma_imp:
            g = subword < 3 ? 63 & (l << 3 | l >> 3) : 0;
            break;

        case gamma_imn:
            g = subword < 3 ? 63 & (l << 3 | l >> 3) : 63;
            break;

        case gamma_imh:
            g = subword > 2 ? 63 & (l << 3 | l >> 3) : 0;
            break;

        case gamma_imb:
            g = 63 & (l << 3 | l >> 3);
            break;

        default:                /* unallocated */
            break;
    }

    // Second copy of 6-bit result
    g |= (g & 63) << 6;

    if (!g)
        /* zero flag */
        g |= g_zero_bit;

    if (subword == 5 && g & bit5)   // copy to reduce capacitance at N mux
        g |= bit13;

    return g;
}

static int gamma_calc(int subword, int i)
{
    int l, r, slot;

    l = i & 63;
    r = i >> 6 & 63;
    slot = i >> 12 & 63;

    return gamma_operation(slot, subword, l, r);
}

// For the ALU console, convert an gamma slot enum into the old
// character code.
int gamma_emoji(int slot)
{
    static struct {
        int i;
        int c;
    } moji[] = {
        gamma_mux, '?',
        gamma_mux + 1, '?',
        gamma_add, '+',
        gamma_add + 1, '+',
        gamma_kil, 'k',
        gamma_kil + 1, 'k',
        gamma_stu, 'w',
        gamma_stu + 1, 'w',
        gamma_rol, 'l',
        gamma_rol + 1, 'L',
        gamma_mix, 's',
        gamma_xim, 'S',
        gamma_uny, 'u',
        gamma_xor, '^',
        gamma_pit, 'p',
        -1, -1,
    };

    int i;

    for (i = 0; moji[i].i >= 0; i++)
        if (slot == moji[i].i)
            return moji[i].c;

    return '$';                 // This is an error.
}
