/*
    instruct.c
    Opcode database from which the instruction decoder is fed
*/

/*
    These are the architecture's CPU cycle types. Each opcode is exactly
    one of these types, but some types are used by more than one opcode.
    The cycle type is used to compute what the instruction decoder needs
    to do in terms of handling the CPU's SRAMs, flip-flops, and some of
    the glue logic. Most of this decoding leads to output enables,
    clock enables, etc.

    The implementation of each cycle type is in decoder.c.
*/
enum {
    cy_alu,             // ALU operation
    cy_now,             // no-write ALU operation
    cy_j,               // jump
    cy_c,               // call
    cy_r,               // return
    cy_imm,             // load immediate
    cy_ld,              // load memory (honors page table)
    cy_sto,             // store memory (honors page table)
    cy_hlt,             // privileged halt
    cy_wcm,             // privileged write code memory
    cy_rcm,             // privileged read code memory
    cy_wdm,             // privileged write data memory (bypasses page table)
    cy_rdm,             // privileged read data memory (bypasses page table)
    cy_wpt,             // privileged write page table
    cy_rpt,             // privileged read page table
};

/*
    The ALU can do four things with the incoming T flag.
    This is not the new value of T, but how the existing T is seen.
*/
enum {
    t_allow = 2,                // This bit is ANDed with the incoming T.
    t_flip = 1,                 // The AND result is XORed with this bit.

    t_0 = 0,                    // replace T with 0
    t_1 = t_flip,               // replace T with 1
    t_id = t_allow,             // leave T as-is
    t_inv = t_flip | t_allow,   // invert T
};

/*
    These are the architecture's opcodes. Most columns here represent
    address inputs to the ALU's many SRAMs. The instruction decoder SRAM
    is computed from this table.
*/
static struct instruction {
    // These apply to all instructions.
    char *name;         // assembler name (footnote N)
    int opcode;         // opcode (oc_*)
    int cycle;          // cycle sequence (cy_*)

    // These apply only to ALU instructions.
    int alpha;          // ALU first operation (alpha_*)
    int beta;           // ALU second operation (beta_*)
    int gamma;          // ALU third operation (gamma_*)
    int theta;          // ALU theta operation (theta_*)
    int zeta;           // ALU zeta operation (flag scheme; zeta_*)
    int t_vision;       // ALU's vision of incoming T flag
} the_insts[] = {
#define X(name, op, cy, a, b, g, th, z, t) name, oc_ ## op, cy_ ## cy, \
        alpha_ ## a, beta_ ## b, gamma_ ## g, theta_ ## th, zeta_ ## z, t_ ## t,
//    name       opcode  cycle  alpha beta gamma theta zeta    T
    X("a.uuu"  , A_uuu  , alu, add_uu, id , add, add, add_u  , 0  )
    X("a.uus"  , A_uus  , alu, add_uu, id , add, add, add_s  , 0  )
    X("a.usu"  , A_usu  , alu, add_us, id , add, add, add_u  , 0  )
    X("a.uss"  , A_uss  , alu, add_us, id , add, add, add_s  , 0  )
    X("a.suu"  , A_suu  , alu, add_su, id , add, add, add_u  , 0  )
    X("a.sus"  , A_sus  , alu, add_su, id , add, add, add_s  , 0  )
    X("a.ssu"  , A_ssu  , alu, add_ss, id , add, add, add_u  , 0  )
    X("a.sss"  , A_sss  , alu, add_ss, id , add, add, add_s  , 0  )

    X("s.uuu"  , S_uuu  , alu, sub_uu, id , add, add, add_u  , 1  )
    X("s.uus"  , S_uus  , alu, sub_uu, id , add, add, add_s  , 1  )
    X("s.usu"  , S_usu  , alu, sub_us, id , add, add, add_u  , 1  )
    X("s.uss"  , S_uss  , alu, sub_us, id , add, add, add_s  , 1  )
    X("s.suu"  , S_suu  , alu, sub_su, id , add, add, add_u  , 1  )
    X("s.sus"  , S_sus  , alu, sub_su, id , add, add, add_s  , 1  )
    X("s.ssu"  , S_ssu  , alu, sub_ss, id , add, add, add_u  , 1  )
    X("s.sss"  , S_sss  , alu, sub_ss, id , add, add, add_s  , 1  )

    X("rs.uuu" , RS_uuu , alu, rev_uu, id , add, add, add_u  , 1  )
    X("rs.uus" , RS_uus , alu, rev_uu, id , add, add, add_s  , 1  )
    X("rs.usu" , RS_usu , alu, rev_us, id , add, add, add_u  , 1  )
    X("rs.uss" , RS_uss , alu, rev_us, id , add, add, add_s  , 1  )
    X("rs.suu" , RS_suu , alu, rev_su, id , add, add, add_u  , 1  )
    X("rs.sus" , RS_sus , alu, rev_su, id , add, add, add_s  , 1  )
    X("rs.ssu" , RS_ssu , alu, rev_ss, id , add, add, add_u  , 1  )
    X("rs.sss" , RS_sss , alu, rev_ss, id , add, add, add_s  , 1  )

    X("ac.uuu" , AC_uuu , alu, add_uu, id , add, add, add_u  , id )
    X("ac.uus" , AC_uus , alu, add_uu, id , add, add, add_s  , id )
    X("ac.usu" , AC_usu , alu, add_us, id , add, add, add_u  , id )
    X("ac.uss" , AC_uss , alu, add_us, id , add, add, add_s  , id )
    X("ac.suu" , AC_suu , alu, add_su, id , add, add, add_u  , id )
    X("ac.sus" , AC_sus , alu, add_su, id , add, add, add_s  , id )
    X("ac.ssu" , AC_ssu , alu, add_ss, id , add, add, add_u  , id )
    X("ac.sss" , AC_sss , alu, add_ss, id , add, add, add_s  , id )

    X("sc.uuu" , SC_uuu , alu, sub_uu, id , add, add, add_u  , inv)
    X("sc.uus" , SC_uus , alu, sub_uu, id , add, add, add_s  , inv)
    X("sc.usu" , SC_usu , alu, sub_us, id , add, add, add_u  , inv)
    X("sc.uss" , SC_uss , alu, sub_us, id , add, add, add_s  , inv)
    X("sc.suu" , SC_suu , alu, sub_su, id , add, add, add_u  , inv)
    X("sc.sus" , SC_sus , alu, sub_su, id , add, add, add_s  , inv)
    X("sc.ssu" , SC_ssu , alu, sub_ss, id , add, add, add_u  , inv)
    X("sc.sss" , SC_sss , alu, sub_ss, id , add, add, add_s  , inv)

    X("rsc.uuu", RSC_uuu, alu, rev_uu, id , add, add, add_u  , inv)
    X("rsc.uus", RSC_uus, alu, rev_uu, id , add, add, add_s  , inv)
    X("rsc.usu", RSC_usu, alu, rev_us, id , add, add, add_u  , inv)
    X("rsc.uss", RSC_uss, alu, rev_us, id , add, add, add_s  , inv)
    X("rsc.suu", RSC_suu, alu, rev_su, id , add, add, add_u  , inv)
    X("rsc.sus", RSC_sus, alu, rev_su, id , add, add, add_s  , inv)
    X("rsc.ssu", RSC_ssu, alu, rev_ss, id , add, add, add_u  , inv)
    X("rsc.sss", RSC_sss, alu, rev_ss, id , add, add, add_s  , inv)

    X("a.w"    , A_w    , alu, add_uu, id , add, add, wrap   , 0  )
    X("ac.w"   , AC_w   , alu, add_uu, id , add, add, wrap   , inv)
    X("s.w"    , S_w    , alu, sub_uu, id , add, add, wrap   , 1  )
    X("sc.w"   , SC_w   , alu, sub_uu, id , add, add, wrap   , id )
    X("rs.w"   , RS_w   , alu, rev_uu, id , add, add, wrap   , 1  )
    X("rsc.w"  , RSC_w  , alu, rev_uu, id , add, add, wrap   , id )

    X("igf"    , IGF    , alu, boo   , id , kil, 1  , logic  , 1  )
    X("igt"    , IGT    , alu, boo   , id , mux, 1  , logic  , 1  )
    X("xl"     , XL     , alu, boo   , id , mux, 0  , logic  , 0  )
    X("xr"     , XR     , alu, boo   , id , mux, 1  , logic  , 1  )
    X("nl"     , NL     , alu, boo   , not, add, 0  , logic  , 0  )
    X("nr"     , NR     , alu, boo   , id , xor, 1  , logic  , 1  )
    X("and"    , AND    , alu, and   , id , add, 0  , logic  , 0  )
    X("nand"   , NAND   , alu, and   , not, add, 0  , logic  , 0  )
    X("or"     , OR     , alu, or    , id , add, 0  , logic  , 0  )
    X("nor"    , NOR    , alu, or    , not, add, 0  , logic  , 0  )
    X("xor"    , XOR    , alu, boo   , id , xor, 1  , logic  , 1  )
    X("xnor"   , XNOR   , alu, boo   , not, xor, 1  , logic  , 1  )
    X("lanr"   , LANR   , alu, or    , id , xor, 1  , logic  , 1  )
    X("ranl"   , RANL   , alu, and   , id , xor, 1  , logic  , 1  )
    X("lonr"   , LONR   , alu, and   , not, xor, 1  , logic  , 1  )
    X("ronl"   , RONL   , alu, or    , not, xor, 1  , logic  , 1  )

    X("max.uuu", MAX_uuu, alu, cmp_uu, id , mux, lt , add_u  , 0  )
    X("max.uus", MAX_uus, alu, cmp_uu, id , mux, lt , add_s  , 0  )
    X("max.usu", MAX_usu, alu, cmp_us, id , mux, lt , add_u  , 0  )
    X("max.uss", MAX_uss, alu, cmp_us, id , mux, lt , add_s  , 0  )
    X("max.suu", MAX_suu, alu, cmp_su, id , mux, lt , add_u  , 0  )
    X("max.sus", MAX_sus, alu, cmp_su, id , mux, lt , add_s  , 0  )
    X("max.ssu", MAX_ssu, alu, cmp_ss, id , mux, lt , add_u  , 0  )
    X("max.sss", MAX_sss, alu, cmp_ss, id , mux, lt , add_s  , 0  )

    X("min.uuu", MIN_uuu, alu, cmp_uu, id , mux, gt , add_u  , 1  )
    X("min.uus", MIN_uus, alu, cmp_uu, id , mux, gt , add_s  , 1  )
    X("min.usu", MIN_usu, alu, cmp_us, id , mux, gt , add_u  , 1  )
    X("min.uss", MIN_uss, alu, cmp_us, id , mux, gt , add_s  , 1  )
    X("min.suu", MIN_suu, alu, cmp_su, id , mux, gt , add_u  , 1  )
    X("min.sus", MIN_sus, alu, cmp_su, id , mux, gt , add_s  , 1  )
    X("min.ssu", MIN_ssu, alu, cmp_ss, id , mux, gt , add_u  , 1  )
    X("min.sss", MIN_sss, alu, cmp_ss, id , mux, gt , add_s  , 1  )

    X("bound.u", BOUND_u, now, sub_uu, id , add, add, bound  , 1  )
    X("bound.s", BOUND_s, now, sub_us, id , add, add, bound  , 1  )

    X("cmp.uu" , CMP_uu , now, sub_uu, id , add, add, cmp    , 1  )
    X("cmp.us" , CMP_us , now, sub_us, id , add, add, cmp    , 1  )
    X("cmp.su" , CMP_su , now, sub_su, id , add, add, cmp    , 1  )
    X("cmp.ss" , CMP_ss , now, sub_ss, id , add, add, cmp    , 1  )

    X("asl.uu" , ASL_uu , alu, e36   , shl, rol, 0  , asl_uu , 0  )
    X("asl.us" , ASL_us , alu, e35   , shl, rol, 0  , asl_us , 0  )
    X("asl.su" , ASL_su , alu, e36   , shl, rol, 0  , asl_su , 0  )
    X("asl.ss" , ASL_ss , alu, e35   , shl, rol, 0  , asl_ss , 0  )

    X("asr.uu" , ASR_uu , alu, asr   , shr, rol, 0  , asr_uu , 0  )
    X("asr.us" , ASR_us , alu, asr   , shr, rol, 0  , asr_us , 0  )
    X("asr.su" , ASR_su , alu, asr   , shr, rol, c5 , asr_su , 0  )
    X("asr.ss" , ASR_ss , alu, asr   , shr, rol, c5 , asr_ss , 0  )

//    name       opcode  cycle  alpha beta gamma theta zeta    T
    X("lsl"    , LSL    , alu, boo   , shl, rol, 0  , logic  , 0  )
    X("lsr"    , LSR    , alu, boo   , shr, rol, 0  , logic  , 0  )
    X("rol"    , ROL    , alu, boo   , rol, rol, 0  , logic  , 0  )

    X("ml"     , ML     , alu, ml    , id , add, 0  , asr_uu , 0  )
    X("mh"     , MH     , alu, mh    , rtl, add, 0  , asr_us , 0  )
    X("mhns"   , MHNS   , alu, mh    , id , add, 0  , asr_uu , 0  )
    X("dsl"    , DSL    , alu, dsl   , rtl, add, dsl, logic  , id )

    X("swiz"   , SWIZ   , alu, boo   , swz, add, 0  , logic  , 0  )
    X("nudge"  , NUDGE  , alu, nud   , id , mux, 2r , logic  , 1  )
    X("crf"    , CRF    , now, boo   , id , kil, 1  , clear_r, 1  )
    X("ham3"   , HAM3   , alu, add_uu, id , kil, 2l , logic  , 0  )
    X("txor"   , TXOR   , alu, boo   , txo, add, 0  , logic  , 0  )

    X("pit"    , PIT    , alu, boo   , id , pit, 1  , logic  , 0  )
    X("pat"    , PAT    , alu, boo   , pat, add, 0  , logic  , 0  )
    X("pait"   , PAIT   , alu, boo   , pat, pit, 1  , logic  , 0  )

    X("mix"    , MIX    , alu, mix   , mix, mix, 0  , logic  , 0  )
    X("xim"    , XIM    , alu, xim   , xim, xim, 1  , logic  , 1  )

    X("un.a"   , UNARY_A, alu, uny   , id , add, 0  , logic  , 0  )
    X("un.b"   , UNARY_B, alu, boo   , uny, add, 0  , logic  , 0  )
    X("un.g"   , UNARY_G, alu, boo   , id , uny, 0  , logic  , 0  )
    X("stun.a" , STUN_A , alu, stu   , stu, stu, add, add_u  , 1  )
    X("stun.b" , STUN_B , alu, stu   , stu, stu, gt , logic  , 1  )
    X("stun.c" , STUN_C , alu, stu   , stu, stu, 0  , asl_uu , 0  )
    X("stun.d" , STUN_D , alu, stu   , stu, stu, 2r , logic  , 1  )
    X("stun.e" , STUN_E , alu, stu   , stu, stu, 2l , logic  , 0  )
    X("stun.f" , STUN_F , alu, stu   , stu, stu, rev, t_adj  , 1  )
    X("stun.g" , STUN_G , alu, stu   , stu, stu, 0  , t_adj  , 0  )
    X("stun.h" , STUN_H , alu, stu   , stu, stu, 0  , t_adj  , id )
    X("stun.i" , STUN_I , alu, stu   , stu, stu, add, t_adj  , 1  )
    X("stun.j" , STUN_J , alu, stu   , stu, stu, add, t_adj  , 1  )

    X("c"      , C      , c  , boo   , id , kil, 1  , logic  , 1  )
    X("r"      , R      , r  , boo   , id , kil, 1  , logic  , 1  )
    X("j"      , J      , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.lt"   , J_lt   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.le"   , J_le   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.eq"   , J_eq   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.ne"   , J_ne   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.ge"   , J_ge   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.gt"   , J_gt   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.r"    , J_r    , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.nr"   , J_nr   , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.t"    , J_t    , j  , boo   , id , kil, 1  , logic  , 1  )
    X("j.nt"   , J_nt   , j  , boo   , id , kil, 1  , logic  , 1  )

    X("nop"    , NOP    , now, boo   , id , kil, 1  , logic  , 1  )

    X("imp"    , IMP    , imm, boo   , imm, imp, 0  , logic  , 0  )
    X("imn"    , IMN    , imm, boo   , imm, imn, 1  , logic  , 1  )
    X("imh"    , IMH    , imm, boo   , imm, imh, 0  , logic  , 0  )
    X("imb"    , IMB    , imm, boo   , imm, imb, 1  , logic  , 1  )

    X("ld"     , LD     , ld , boo   , id , add, 0  , logic  , 0  )
    X("sto"    , STO    , sto, boo   , id , add, 0  , logic  , 0  )

    X("test1"  , TEST1  , now, boo   , id , kil, 1  , logic  , 1  )
    X("test2"  , TEST2  , now, boo   , id , kil, 1  , logic  , 1  )
    X("test3"  , TEST2  , now, boo   , id , kil, 1  , logic  , 1  )

    // There are only a few privileged instructions. They are:
    X("halt"   , HALT   , hlt, boo   , id , add, 0  , logic  , 0  )
    X("wcm"    , WCM    , wcm, boo   , id , add, 0  , logic  , 0  )
    X("rcm1"   , RCM1   , rcm, boo   , id , add, 0  , logic  , 0  )
    X("rcm2"   , RCM2   , imm, boo   , id , add, 0  , logic  , 0  )
    X("wdm"    , WDM    , wdm, boo   , id , add, 0  , logic  , 0  )
    X("rdm"    , RDM    , rdm, boo   , id , add, 0  , logic  , 0  )
    X("wpt"    , WPT    , wpt, boo   , id , add, 0  , logic  , 0  )
    X("rpt"    , RPT    , rpt, boo   , id , add, 0  , logic  , 0  )
#undef X
};

// This is a hook to convert an opcode to its textual name.
const char *opcode_name(int o)
{
    int i;

    for (i = 0; i < CARD(the_insts); i++)
        if (the_insts[i].opcode == o)
            return the_insts[i].name;

    return "INVALID";
}

// This converts an opcode name to its numeric form. Returns -1 if not found.
int find_opcode(char *name)
{
    int i;

    for (i = 0; i < CARD(the_insts); i++)
        if (!strcmp(the_insts[i].name, name))
            return the_insts[i].opcode;
    return -1;
}

/* FOOTNOTES

N:  For many instructions, the assembler does not rely solely
    on assembler name to determine the opcode to emit. For
    example, "++" can result in any ac.* instruction. So to
    "rename" instructions, changes to symbols in syntax.c may be
    needed in addition to or instead of changes to this field.

*/
