/*
    pieces.c
    These are firmware-computing functions that don't fit neatly elsewhere.
*/

/*
    The architecture is able to permute 6-bit subwords within (PIT)
    or across (PAT) tribbles. Because the right operand is also 6 bits,
    only 64 permutations of the 6! = 720 we might like to support are
    available. But we can reach any of the 720 with at most two
    permutations from this set.

    Here are the permutations which were selected for the architecture.
    Of these 64, 58 were selected manually as "more likely to need" and
    include stuff like rotations, reversals, and other goodies. The final
    6 permutations were chosen algorithmically to ensure that the set is
    "complete" and can do any permutation of six bits in at most two
    operations from this table.

    The indexing convention of this table differs from the near-universal
    and ordinarily preferable custom of having bit 0 on the right. To
    make this more legible to humans, bit 0 is on the left.

    The interpretation of this table is that a tribble with bit positions
    012345 is altered by the permutation into one of the following orders.
    The list starts with the identity, then the one-position left rotation
    123450, and so on. Reverse bits, or 543210, is pset[11] == pset[013].
*/
const char *pset[64] = {
    "012345", "123450", "234501", "345012", "450123", "501234", "432105",
    "321054", "210543", "105432", "054321", "543210", "102345", "210345",
    "312045", "412305", "512340", "021345", "032145", "042315", "052341",
    "013245", "014325", "015342", "012435", "012543", "012354", "452301",
    "014523", "230145", "024135", "031425", "031524", "142035", "253140",
    "304251", "415302", "520413", "035124", "140253", "251304", "302415",
    "413520", "524031", "102354", "315042", "542310", "513240", "021435",
    "034125", "043215", "103254", "240513", "453201", "521430", "534120",
    "120534", "201453", "034512", "035421", "105243", "130542", "254310",
    "510432",
};

// Apply 'perm' (0-63 from pset) to 6-bit integer 'i'.
static int permute(int perm, int i)
{
    const char *p;
    int dpv,        // 2**dpv is the destination bit's place value.
        spv,        // 2**spv is the source bit's place value.
        out = 0;    // permutation result

    if (perm < 0 || perm > 63)
        ng("permute");
    p = pset[perm];

    // Loop through place values of destination bits.
    // These, unlike the pset[] strings, number with bit 0 rightmost.
    // XXX This 'for' line has no need to be written backwards,
    // XXX but wait for testing before switching it.
    for (dpv = 5; dpv >= 0; --dpv) {
        spv = '0' + 5 - p[5 - dpv];
        if (i & 1 << spv)
            out |= 1 << dpv;
    }

    return out;
}

// Rotate a 6-bit subword left by 0 to 5 bits.
static int rotate_tribble_left(int t, int n)
{
    if (n < 0 || n > 5)                 // changed 28 Apr 2021 from "n %= 6;"
        ng("rotate_tribble_left %i", n);

    t <<= n;
    return 63 & (t | t >> 6);
}

/*
    The MIX S-boxes are computed by expanding sqrt(2)/2 in a
    short Python program (gen-sbox.py). These are 6-bit S-boxes, and
    there are 64 for each of the alpha, beta, and gamma RAMs, for a
    total of 1,152 boxes. These are output in a base64 variant as
    sbox.c, which is used to generate the real S-boxes for the firmware.

    This function converts the base64 characters back to integers 0-63.
    The characters are ASCII ' ' through 'a', excluding quotation mark
    and backslash (" \) so that the sbox.c representation doesn't require
    any characters to be quoted.
*/
static int sbox_unquote(int c)
{
    /* back 'c' down so it may be " or \ */
    if (c > '"') --c;
    if (c > '\\'-1) --c;
    c -= 32;
    if (c < 0 || c > 63)
        ng("sbox_unquote %i", c);
    return c;
}

/*
    This is a counterpart to sbox_unquote, but it inverts the S-box
    to produce firmware for the XIM (inverse MIX) instruction.
    The S-box (in its original C form) is passed as 'q', and the
    integer being un-mixed is passed as 'i'. Better parameter names
    might be nice someday.

    Although sbox_unquote() and sbox_invert() are sister functions,
    their calling conventions differ. See for instance alpha_operation()
    for how they hook together.
*/
static int sbox_invert(char *q, int i)
{
    char *s;

    i += 32;
    /* nudge 'i' up so it may not be " or \ */
    if (i >= '"') ++i;
    if (i >= '\\') ++i;
    ifnot (s = strchr(q, i))
        ng("sbox_invert %s [%i]", q, i);
    i = s - q;
    if (i < 0 || i > 63)
        ng("sbox_invert %i", i);
    return i;
}

/*
    This table describes how opcodes are stratified by signage.
*/
static struct {
    int cat;            // ocat_*, maybe not needed? XXX
    int matters;        // which registers have signage, maybe not needed? XXX
    int opcodes[8];     // stratified: uuu uus usu uss suu sus ssu ssu
} opSigns[] = {
#define X(m, c, o, p, q, r, s, t, u, v) ocat_ ## c, m, \
        Y(c, o), Y(c, p), Y(c, q), Y(c, r), Y(c, s), Y(c, t), Y(c, u), Y(c, v),
#define Y(c, suffix) oc_ ## c ## _ ## suffix
    X(7, A, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, S, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, RS, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, AC, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, SC, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, RSC, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, MAX, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(7, MIN, uuu, uus, usu, uss, suu, sus, ssu, sss)
    X(1, BOUND, u, s, u, s, u, s, u, s)
    X(3, CMP, uu, us, su, ss, uu, us, su, ss)
    X(6, ASL, uu, uu, us, us, su, su, ss, ss)
    X(6, ASR, uu, uu, us, us, su, su, ss, ss)
#undef X
#undef Y
};

/*
    Given an opcode for any signage, along with a desired signage,
    return the stratified opcode for the desired signage.
*/
int stratify_opcode(int oc, int signage)
{
    int row, col;

    for (row = 0; row < CARD(opSigns); row++)
        for (col = 0; col < 8; col++)
            if (opSigns[row].opcodes[col] == oc)
                // Found. Clamp 'signage' in lieu of sanity check.
                return opSigns[row].opcodes[signage & 7];

    // Not found. Opcode is fine as-is, e.g., bitwise boolean.
    return oc;
}

/*
    Separate an opcode from its signage, returning the all-unsigned
    stratification of the opcode, and recording the observed signage.
*/
int unstratify_opcode(int oc, int *signage)
{
    int row, col;

    for (row = 0; row < CARD(opSigns); row++)
        for (col = 0; col < 8; col++)
            if (opSigns[row].opcodes[col] == oc)
                return *signage = col, opSigns[row].opcodes[0];

    // Not in table. Opcode does not stratify by signage.
    return *signage = 0, oc;
}
