/* theta SRAM definition

INPUT BITS

    0-5     carry bits from a0..a5, named c0..c5
    6-11    propagate bits from a0..a5, named p0..p5
    12      T flag (often adjusted)
    13-17   operation (often extended by T flag)

OUTPUT BITS

    0-5     carry decisions to g0..g5, named d0..d5
    6       "summary" of carries c0..c5
    7       "summary" of propagates p0..p5

22 Dec 2020: There are 14 combinations of theta with T flag settings
right now, meaning that expanding theta's operation from 3 to 4 bits
would let us remove the T adjustment glue logic. The selected SRAMs
in fact permit 5 bits for the theta operation flag, so there's oodles
of room. [12 April 2021, not pursuing this and probably won't.]

*/

static int theta_operation(int slot, int props, int carries)
{
    int th = 0, i;
    int t_flag = slot & 1;

    /* see "theta operations" table from PDF */
    switch (slot) {
        case theta_add:         // add carry decisions
        case theta_add_more:
        case theta_dsl:         /* DSL's rotated add-with-carry */
        case theta_dsl_more:
            if (props & carries)
                break;          /* special case for integer absolute value */
            th = t_flag;        /* base case */
            for (i=2; i < 64; i <<= 1) {
                /* Propagate and carry is mutually exclusive here. */
                /* Carry from previous? Set carry decision. */
                th |= i & carries << 1;
                /* Propagate from previous? Set carry decision. */
                th |= i & (props & th) << 1;
            }
            if (slot == theta_dsl || slot == theta_dsl_more)
                th = 63 & th << 1;
            if (th & bit5)
                th |= bit6;     // zeta's view of d5
            break;

        case theta_lt:          // bit 6 muxes the encoded range
            th = props < carries ? bit6 | 63 : 0;
            break;

        case theta_gt:          // bit 6 muxes the encoded range
            th = props > carries ? bit6 | 63 : 0;
            break;

        case theta_2l:
            // This operation sets all decision bits left of any set carry.
            // HAM3 is the only customer today.
            th = 63 & ~(carries ^ carries - 1);
            break;

        case theta_2r:
            // This operation sets all decision bits right of any set carry.
            // NUDGE is only customer today.
            for (i=1; i & 63; i <<= 1)
                if (carries & i)
                    th |= i-1;
            break;

        case theta_0:           // already zero
            if (carries)
                th |= bit6;     // ASR encroachment by 0; fs_t_adj operations
            if (props)
                th |= bit7;     // ASR encroachment by 1; fs_t_adj operations
            break;

        case theta_1:
            th = 63;
            if (carries)
                th |= bit6;     // fs_t_adj operations
            if (props)
                th |= bit7;     // fs_t_adj operations
            break;

        // negative for ASR. We don't use theta.{, because we need the
        // c_5 also copied to d_5, which theta.{ does not do. We can't
        // adjust for this in alpha, because we use plain alpha.=
        case theta_c5:          /* replicate c_5 to all d_i */
            if (carries & 32)
                th = 63;
            break;

        case theta_rev:         // left-to-right (reverse) addition
        case theta_rev_more:
            // XXX We do NOT require two theta slots for this today,
            // although we've allocated them. True addition would
            // require two slots, but for our inc/dec piddling, we
            // could have done that without T and used one slot.
            // But we are not short on theta slots right now.
            if (props & carries)
                break;      // special case for reverse absolute value
            th = t_flag ? 32 :0;   // base case
            for (i = 16; i; i >>= 1) {
                // Propagate and carry are mutually exclusive here.
                // Carry from previous? Set carry decision.
                th |= i & carries >> 1;
                // Propagate from previous? Set carry decision.
                th |= i & (props & th) >> 1;
            }
            if (th & bit0)
                th |= bit6;     // zeta's view of d0
            break;

        default:
            break;
    }

    return th;
}

static int theta_calc(int subword_unused, int i)
{
    int op, props, carries;

    carries = 63 & i;
    props = 63 & i >> 6;
    op = 63 & i >> 12;

    return theta_operation(op, props, carries);
}

// For the ALU console, convert an theta slot enum into the old
// character code.
int theta_emoji(int slot)
{
    static struct {
        int i;
        int c;
    } moji[] = {
        theta_add, '+',
        theta_add_more, '+',
        theta_rev, 'r',
        theta_rev_more, 'r',
        theta_dsl, 'D',
        theta_dsl_more, 'D',
        theta_lt, '<',
        theta_gt, '>',
        theta_2l, '{',
        theta_2r, '}',
        theta_0, '0',
        theta_1, '1',
        theta_c5, 'n',
        -1, -1,
    };

    int i;

    for (i = 0; moji[i].i >= 0; i++)
        if (slot == moji[i].i)
            return moji[i].c;

    return '$';                 // This is an error.
}
