// Stacked unary operations
// The simple unary operations are in audited.c due to testing issues.

static void stacked_unary(
    const int carry,        // Only gamma has a carry input.
    const int subword,      // Inputs are 'const' to avoid inadvertent
    const int l,            // changes during the calculation of one
    const int r,            // output that would invalidate other outputs.
    int *alpha,
    int *beta,
    int *gamma              // Outputs
) {
    int a = 0, b = 0, g = 0;
    int range = 0;          // Only alpha offers encoded range bits.

    switch (r) {

        case su_abs2s:      // abs value: negation step to signed
/*
            Designed and guaranteed only for negative inputs.
            Believed to be, but not guaranteed to be, a working
            signed -> signed negation, except the N flag does NOT
            work. Z and T are believed to work, and the 36-bit
            output is believed to be correct. T is only set if
            the pole is input.
*/
            a = 63 ^ l;
            if (a == 63)
                a |= a_prop_bit;
            if (subword == 5) {
                if (l == 32)
                    range = 2;  // fits ONLY if d5 clear (and won't be zero)
                else
                    range = 12; // always fits (and can be zero)
            }

            b = l;

            g = carry ? 63 & l + 1 : l;
            break;

        case su_abs2i:      // abs value: identity step (no bit 35)
/*
            Input may not have bit 35 set.
*/
            a = subword ? l : 63 & l - 1;   // T will roll subword 0
            if (subword == 5)
                range = 12;                 // always fits (and can be zero)

            b = l;

            g = carry ? 63 & l + 1 : l;
            break;

        case su_abs2u:      // abs value: negation step to unsigned
/*
            This only works for non-positive inputs.
*/
            a = 63 ^ l;
            if (a == 63)
                a |= a_prop_bit;
            if (subword == 5)
                range = 12;                 // always fits; can be zero

            b = l;

            g = carry ? 63 & l + 1 : l;
            break;

        case su_abs1s:      // {0} if negative else {1} for abs value

            a = subword == 5 && l & 32 ? 63 : 0;

            b = l ? 63 : 0;

            g = l ? 0 : 1;
            break;

        case su_abs1u:      // {2} if negative else {1} for abs value

            a = subword == 5 && l & 32 ? 63 : 0;

            b = l ? 63 : 0;

            g = l ? 2 : 1;          // {2} if negative else {1}
            break;

        case su_fabs:       // |x|, only works when |x| < 2**30, checked

            if (subword < 5) {
                // assume beta will invert and gamma will increment from T
                a = l;
                if (!a)
                    a |= a_prop_bit;
            }
            else {
                // if negative, send 1 to all betas
                // otherwise, hit CANCEL (prop + carry special case)
                a = l & 32 ? 63 : a_prop_bit | a_carry_bit;

                // This correctly sets T for signed output.
                // For unsigned output, could use 12 instead of 2,
                // in which case pole in -> pole out is marked valid.
                // But rather than support this, we conseratively
                // document that pole is not a supported input.
                range = l == 0 || l == 63 ? 2 : 0;
            }

            b = l & 32 ? 31 & ~l : l;

            g = carry ? 63 & l + 1 : l;
            break;

        case su_signum:     // full signum
/*
            All inputs are valid; never sets T.
*/
            a = l ? 63 : 0;                 // show if nonzero
            if (subword == 5 && l & 32)
                a |= a_prop_bit;            // show if negative

            b = l ? 63 : 0;         // cascade non-zero condition

            if (!l)
                g = 0;
            else if (carry)
                g = 63;
            else
                g = !subword;
            break;

        case su_lsprep:     // uint to left shift control word
/*
            All inputs are valid; never sets T.
*/
            if (subword && l)
                a |= a_prop_bit;            // maximum shift (via d[i])
            else if (l > 35)
                a = 36;                     // maximum shift (via beta)
            else
                a = l;                      // ordinary shift

            b = l & 1 ? 63 : 0;     // copy tribble 0 to all others

            g = carry ? 36 : l;
            break;

        case su_rsprep:     // uint to right shift control word
/*
            All inputs are valid; never sets T.
*/
            if (subword && l)
                a |= a_prop_bit;            // maximum shift (via d[i])
            else if (l > 35)
                a = 36;                     // maximum shift (via beta)
            else if (!l)
                a = 0;                      // no shift
            else
                a = 36 - l;                 // ordinary shift, left perspective

            b = l & 1 ? 63 : 0;     // copy tribble 0 to all others

            g = carry ? 36 : l;
            break;

        case su_lrprep:     // uint to left rotate control word
/*
            Inputs > 63 give wrong result and set T.
*/
            if (subword && l)
                a |= a_prop_bit;            // error; set T
            else
                a = l % 36;

            b = l & 1 ? 63 : 0;     // copy tribble 0 to all others

            g = l;
            break;

        case su_rrprep:     // uint to right rotate control word
/*
            Inputs > 63 give wrong result and set T.
*/
            if (subword && l)
                a |= a_prop_bit;            // error; set T
            else {
                a = 36 - l % 36;
                if (a == 36)
                    a = 0;
            }

            b = l & 1 ? 63 : 0;     // copy tribble 0 to all others

            g = l;
            break;

        case su_ham0:       // 3-cycle popcount, count 0s, cycle 1

            a = count_bits(l, 1, 0, 0);

            b = count_bits(l, 0, 0, 0);

            g = 32 | l;
            break;

        case su_ham1:       // 3-cycle popcount, count 1s, cycle 1

            a = count_bits(l, 0, 0, 0);

            b = count_bits(l, 0, 0, 0);

            g = 32 | l;
            break;

        case su_ham2:       // 3-cycle popcount, count either, cycle 2

            a = subword == 2 ? l << 3 : l;

            switch (subword) {
                case 0:
                case 1:
                case 2:
                    if (l & 2) b |= 1;
                    break;

                case 3:
                case 4:
                case 5:
                    if (l & 4) b |= 1;
                    break;
            }

            g = 32 | 4 * (l >> 3) + 2 * (l & 7);
            break;

        case su_llz:        // light leading 0s for CLZ

            a = 63 ^ l;
            a = reverse_six_bits(a);
            a = 63 & a & ~(a + 1);
            a = reverse_six_bits(a);
            if (a != 63)
                a |= a_carry_bit;

            b = l;

            g = carry ? 0 : l;
            break;

        case su_llo:        // light leading 1s for CLO

            a = reverse_six_bits(l);
            a = 63 & a & ~(a + 1);
            a = reverse_six_bits(a);
            if (a != 63)
                a |= a_carry_bit;

            b = l;

            g = carry ? 0 : l;
            break;

        case su_ltz:        // light trailing 0s for CTZ

            a = 63 ^ l;
            a = 63 & a & ~(a + 1);
            if (a != 63)
                a |= a_carry_bit;

            b = l;

            g = carry ? 0 : l;
            break;

        case su_lto:        // light trailing 1s for CTO

            a = 63 & l & ~(l + 1);
            if (a != 63)
                a |= a_carry_bit;

            b = l;

            g = carry ? 0 : l;
            break;

        case su_parity:     // parity, result in 2**0 position only

            a = 1 & count_bits(l, 0, 0, 0);

            b = 1 & count_bits(l, 0, 0, 0);

            g = l;
            break;

        case su_increv:     // inc bit-reversed word mod 2**36. no T

            a = l;
            if (a == 63)
                a |= a_prop_bit;

            b = l;

            g = reverse_six_bits(l);
            g = carry ? 63 & g + 1 : g;
            g = reverse_six_bits(g);
            break;

        case su_decrev:     // dec bit-reversed word mod 2**36. no T

            a = l;
            if (!a)
                a |= a_prop_bit;

            b = l;

            g = reverse_six_bits(l);
            g = carry ? 63 & g - 1 : g;
            g = reverse_six_bits(g);
            break;

        case su_lfsr:       // 36-bit LFSR

            a = l;
            if (!subword && l & 1)
                a |= a_prop_bit;

            b = subword ? l : l >> 1;

            g = 63 & (l << 5 | l >> 1);
            if (carry)
                g ^= 63 & (lfsr_taps >> 6 * subword);
            break;

        case su_tpoly:      // xor XFSR poly if T is set for 72-bit LFSR

            a = l;

            b = l;

            g = l;
            if (carry)
                g ^= 63 & (lfsr_taps >> 6 * subword);
            break;

        case su_rcl:        // shift/rotate one place left into/thru T

            a = l;
            if (subword == 5)
                a |= a & 32 ? a_carry_bit : a_prop_bit;

            b = subword == 5 ? 63 & (l << 1 | l >> 5) : l;

            g = 63 & (l << 1 | l >> 5);
            if (!subword) {
                g &= 62;
                if (carry)
                    g |= 1;
            }
            break;

        case su_rcr:        // shift/rotate one place right into/thru T

            a = l;
            if (!subword)
                a |= a & 1 ? a_carry_bit : a_prop_bit;

            b = subword ? l : 63 & (l << 5 | l >> 1);

            g = 63 & (l << 1 | l >> 5);
            if (!subword) {
                g &= 62;
                if (carry)
                    g |= 1;
            }
            break;

    }

    *alpha = a | range << 8;
    *beta = b;
    *gamma = g;
}
