/* zeta (flags) SRAM definition

INPUT BITS

    0       previous N
    1       previous R
    2       previous T
    3       previous Z
    4-8     alpha5 encoded range
    9       theta carry summary -OR- flags' view of carry decision bit 5
    10      theta propagate summary
    11-15   flag scheme from control decoder
    16-17   - unallocated -

OUTPUT BITS
XXX This order is functionally acceptable, but not aesthetically ideal.

    0       N'egative (per additive functions), or 0 if using gamma's N
    1       R'ange
    2       T'emporal range
    3       Z'ero allowed (active high)
    4       I'nterrupt
    5       if set, use gamma_35 for N flag (bit 0 above must be ZERO)
    6       if set, invert AND8 for Z output

*/

// XXX There may be "safer" ways to write this routine, maybe table driven,
//     to avoid some types of oversights like not preserving Z.
static int zeta_operation(
    int prev_n,
    int prev_r,
    int prev_t,
    int prev_z,
    int encoded_range,
    int c_summ,                     // alternatively: carry decision bit 5
    int p_summ,
    int scheme
) {
    int f = 0, t_changes = 1;
    int range = c_summ ? encoded_range / 5 : encoded_range % 5;
    int invert_zero_and = 0;

    switch (scheme) {
        default:                    // Don't spend NOR boot flash or
            return 0;               // boot time on unused schemes.

        case zeta_logic:              // keep T; N & Z come from gamma
            f = 0x28 | prev_t << 2;
            t_changes = 0;
            break;

        case zeta_cmp:                // Z from gamma, N from subtraction, keep T
            if (range < 2)
                f |= bit0;          // output is negative
            if (range == 2)
                f |= bit3;          // possible for a zero result
            f |= prev_t << 2;       // keep T
            t_changes = 0;
            break;

        case zeta_add_s:
            if (range != 1 && range != 2)
                f = bit2;           // does not fit in signed result
            if (range < 2)
                f |= bit0;          // output is negative
            if (range == 2)
                f |= bit3;          // possible for a zero result
            break;

        case zeta_add_u:
            if (range != 2 && range != 3)
                f = bit2;           // does not fit in unsigned result
            if (range < 2)
                f |= bit0;          // output is negative
            if (range == 2)
                f |= bit3;          // possible for a zero result
            break;

        case zeta_wrap:
            // Derived from zeta_add_u.
            // Document result is never negative, or tell assembler
            // not to permit to a signed destination, or both. XXX
            t_changes = 0;          // T actually DOES change, but R doesn't
            if (range != 2 && range != 3)
                f = bit2;           // does not fit in unsigned result
            if (range == 2)
                f |= bit3;          // possible for a zero result
            break;

        case zeta_t_adj:
            t_changes = 0;          // T DOES change, but R doesn't

            // Four cases: any C clears T, any P sets it
            // If we see any C with any P, invert T
            // Otherwise do not change T
            if (p_summ && c_summ)
                prev_t = !prev_t;
            else if (p_summ)
                prev_t = 0;
            else if (c_summ)
                prev_t = 1;

            f = 0x28 | prev_t << 2; // use gamma36 for N, allow zero
            break;

        case zeta_bound:            // can only change I'nterrupt flag
            f = prev_t << 2 | prev_n;
            t_changes = 0;
            if (prev_z)             // force Z true via AND8 & XOR
                invert_zero_and = bit6;
            if (range > 1)
                f |= bit4;          // array subscript out of bounds
            break;

        case zeta_clear_r:          // copy R to T, then clear R
            f = prev_r << 2 | prev_n;
            if (prev_z)             // force Z true via AND8 & XOR
                invert_zero_and = bit6;
            t_changes = 0;
            prev_r = 0;
            break;

        case zeta_asr_uu:
            f = bit3;               // Z follows gamma, N=0, T=0
            break;

        case zeta_asr_us:
            f = bit3;               // Z follows gamma, N=0
            if ((encoded_range & 3) == 3)
                f |= bit2;          // T if negative input, no shift, uns. out
            break;

        case zeta_asr_su:
            f = bit3;               // Z follows gamma
            if (encoded_range & 1)
                f |= bit0 | bit2;   // N and T follows L35
            break;

        case zeta_asr_ss:
            f = bit3;               // Z follows gamma
            if (encoded_range & 1)
                f |= bit0;          // N follows L35
            break;

        case zeta_asl_uu:           // N=0
            if (p_summ)             // 1 encroaches bit 36?
                f |= bit2;          // set T, force Z zero
            else
                f |= bit3;          // no T, gamma may set Z
            break;

        case zeta_asl_us:           // N=0
            // 1 encroaches bit 35 or input is negative?
            if (p_summ || encoded_range & 1)
                f |= bit2;          // set T, force Z zero
            else
                f |= bit3;          // no T, gamma may set Z
            break;

        case zeta_asl_su:
            // 1 encroaches bit 36 or input is negative?
            if (p_summ || encoded_range & 1)
                f |= bit2;          // set T, force Z zero
            else
                f |= bit3;          // no T, gamma may set Z
            if (encoded_range & 1)
                f |= bit0;          // N follows L35
            break;

        case zeta_asl_ss:
            // 1 encroaches bit 35 with non-negative input, or
            // 1 encroaches bit 35 with negative input?
            if (p_summ && !(encoded_range & 1)
                    || c_summ && encoded_range & 1)
                f |= bit2;          // set T, force Z zero
            else
                f |= bit3;          // no T, gamma may set Z
            if (encoded_range & 1)
                f |= bit0;          // N follows L35
            break;
    }

    // XXX 18Aug2020 Suggest testing sometime to confirm that a sequence
    // that sets R, clears it (now T is set, R is not) cannot "re-infect" R.
    if (prev_r || t_changes && f & bit2)
        f |= bit1;                  // sticky R'ange flag

    f |= invert_zero_and;

    // Sanity test for the "zeta's N or 0" OR gate
    if (f & bit0 && f & bit5)
        ng("zeta's N sanity test failed");

    return f;
}

static int zeta_calc(int subword_unused, int i)
{
    int prev_n, prev_r, prev_t, prev_z, enc_ra, c_summ, p_summ, scheme;

    prev_n = i       &  1;
    prev_r = i >>  1 &  1;
    prev_t = i >>  2 &  1;
    prev_z = i >>  3 &  1;
    enc_ra = i >>  4 & 31;
    c_summ = i >>  9 &  1;
    p_summ = i >> 10 &  1;
    scheme = i >> 11 & 31;

    return zeta_operation(prev_n, prev_r, prev_t, prev_z, enc_ra,
        c_summ, p_summ, scheme);
}
