/*
    misc.c
    General programming support, not specific to this application
*/

static int mallocs, frees;

// Send warnings to standard output. Start 'fmt' with '!' to exit program.
void ng(char *fmt, ...)
{
    va_list ap;

    fflush(stdout);                 // get last gasp of helpful text
    va_start(ap, fmt);
    fprintf(stderr, TEA);           // color highlight start
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, BAW);           // color highlight done
    fprintf(stderr, "\n");
    va_end(ap);
    if (*fmt == '!')
        exit(1);
}

/*
    Read an entire file into a single RAM buffer.
    Not suitable for special files like /dev/stdin.

    The file isn't meant to be huge; e.g., it must fit into RAM,
    and its length needs to fit a long.

    On error, don't leak resources, don't issue a message, and return NULL.

    Returns the buffer on success, with no indication as to actual length.
    There is a \0 at the end.  You should free() this buffer when you
    are through with it.
*/
char *slurp(char *filename)
{
    FILE *f = NULL;
    char *buf = NULL;
    long len;

    // Incredible number of system call dependencies in this simple task!
    if (   !(f = fopen(filename, "rb"))
        || fseek(f, 0, SEEK_END)
        || (len = ftell(f)) < 0
        || !(buf = trackmem(NULL, 1 + len))
        || fseek(f, 0, SEEK_SET)
        || len != fread(buf, 1, len, f)
        || fclose(f) )
        goto nope;

    buf[len] = 0;
    return buf;

nope:
    if (f)                          // fclose(NULL) is not legal
        fclose(f);
    trackmem(buf, 0);               // free(NULL) is legal
    return NULL;
}

/*
    calloc/realloc/free wrapper with counting

    In the realloc case only, gained memory is not initialized.
    XXX This needs more in-depth explanation.
*/
void *trackmem(void *old, size_t bytes)
{
    void *p = realloc(old, bytes);

    if (!old && bytes) {
        memset(p, 0, bytes);
        ++mallocs;
    }

    else if (old && !bytes)
        ++frees;

    if (bytes && !p)
        ng("!failed to allocate %lu bytes", (unsigned long) bytes);

    return p;
}

// Print a malloc/free count report.
void trackmem_report(void)
{
   if (frees != mallocs)
        ng("imbalance: mallocs %i frees %i", mallocs, frees);
}

// Print a tribble as bits.
void print6(char *prefix, int t, char *suffix)
{
    uint i;

    if (prefix)
        printf("%s", prefix);

    for (i=bit5; i; i >>= 1)
        printf("%s", t & i ? "1" : "0");

    if (suffix)
        printf("%s", suffix);
}

// Print a 36-bit word as bits.
void print36(char *prefix, long w, char *suffix)
{
    int i;

    if (prefix)
        printf("%s", prefix);

    for (i=5; i>=0; --i)
        print6(NULL, 63 & w >> 6 * i, " " + !i);

    if (suffix)
        printf("%s", suffix);
}

// set buffer from CSPRNG
void urandom(int len, void *buf)
{
    static FILE *uFile;                     // /dev/urandom
    static char *path = "/dev/urandom";

    if (len < 1) return;                    // no "leftover" space to randomize

    if (!uFile && !(uFile = fopen(path, "r")))
        ng("!cannot open CSPRNG %s", path);

    do {
        errno = 0;
        if (len == fread(buf, 1, len, uFile))
            return;                         // success (almost always the case)
    } while (errno == EINTR);

    ng("error reading CSPRNG %s", path);
}
