/*
    misc.h
    Various definitions, mostly not architecture-specific
*/

// Macros favored by the author
#define ifnot(x) if (!(x))              // fewer parens
#define nobreak                         // clarify 'case'; can syntax highlight
#define CARD(x) \
    (sizeof(x) / sizeof(*(x)))          // # of elements in a fixed array
#define UNUSED __attribute__((unused))  // possibly only for GCC (sorry)

// ANSI color escape seq macros, all same width before expansion
#define RED "\x1b[1;31m"
#define GRE "\x1b[0;32m"
#define YEL "\x1b[1;33m"
#define CYA "\x1b[1;34m"
#define PUR "\x1b[1;35m"
#define TEA "\x1b[1;36m"
#define WHI "\x1b[1;37m"
#define BAW "\x1b[m"

// These works like (1 << i), are clearer, and don't need parentheses.
// Unfortunately, 'long' enums are not portable, so none are here.
enum {
    bit0 = 1, bit1 = 2, bit2 = 4, bit3 = 8, bit4 = 16, bit5 = 32, bit6 = 64,
    bit7 = 128, bit8 = 256, bit9 = 512, bit10 = 1024, bit11 = 2048,
    bit12 = 4096, bit13 = 8192, bit14 = 16384, bit15 = 32768, bit16 = 65536,
    bit17 = 131072, bit18 = 262144,
};

// Typedefs favored by the author
typedef unsigned char uchr;
typedef unsigned int uint;
typedef unsigned long ulong;

// Helper functions
void ng(char *fmt, ...);                // Print error; exit if *fmt == '!'.
char *slurp(char *filename);            // Read a file into a buffer.
void *trackmem                          // Track memory (de)allocation.
    (void *old, size_t bytes);
void trackmem_report(void);             // Report on trackmem() calls.
void print6                             // Print a 6-bit subword as bits.
    (char *prefix, int t, char *suffix);
void print36                            // Print a 36-bit word as bits.
    (char *prefix, long w, char *suffix);
void urandom(int len, void *buf);       // Randomize a buffer.

/*
    Above this point is not architecture-specific.
    Below this point   is   architecture-specific.
*/

#define mask36 0xfffffffffL             // 36 ones  on right
#define comp36 (~mask36)                // 36 zeros on right
#define pole   0x800000000L             // x != 0 && x == -x
#define two36 0x1000000000L             // 2 ** 36

enum {
    a_carry_bit = 1 << 6,               // alpha_i output carry bit
    a_prop_bit = 1 << 7,                // alpha_i output propagate bit
    g_zero_bit = 1 << 12,
};

/*
    The Dauug|36 firmware offers fast, well-mixed pseudorandom number
    generation via a fast linear feedback shift register (LFSR) followed
    by a substitution-permutation network (SPN). If a period of 2**36
    is long enough, it takes just two instructions to generate a "good"
    random number. But we need a polynomial to define the LFSR taps,
    and here it is.

    The following polynomials are all primitive and use the same taps in 
    the 36 most sig. bits. Therefore, we conceivably can use the same 
    polynomial for LFSRs of 36, 72, 108, 144. This is the ONLY polynomial
    suitable for all four sizes that has 8 or fewer taps. It cannot support
    180, 216, 252, or 288 bits.

    There is no penalty for using more than 8 taps, but this requires CPU
    time and potentially a Mathematica license over that time to search.
    Finding this 8-tap solution already took 90 minutes of CPU time, and
    further solutions may require far more time.

    x^ 36 + x^ 31 + x^ 13 + x^  7 + x^  6 + x^  5 + x^  3 + x^  2 + 1 
    x^ 72 + x^ 67 + x^ 49 + x^ 43 + x^ 42 + x^ 41 + x^ 39 + x^ 38 + 1 
    x^108 + x^103 + x^ 85 + x^ 79 + x^ 78 + x^ 77 + x^ 75 + x^ 74 + 1 
    x^144 + x^139 + x^121 + x^115 + x^114 + x^113 + x^111 + x^110 + 1 

    const long lfsr_taps =
          1L << 35 | 1L << 30 | 1L << 12 | 1L <<  6
        | 1L <<  5 | 1L <<  4 | 1L <<  2 | 1L <<  1;
*/
#define lfsr_taps 0x840001076L
