#!/usr/bin/python3
from functools import reduce
from itertools import combinations, cycle
import operator

#print(
'''
-----------------
WHERE THIS STANDS
-----------------
24 Feb 2021

Timings start to look good. Some timing adjustments are needed. For example:

1G74's t_pd is <= 1.2 ns (Q) or 1.1 ns (Q#).
16374's t_h needs to be >= 0.4 ns.

So if the RAM's tKQX (clock to output invalid) is >= 2.0 ns, and all clocks are delivered equal, then the length/capacitance penalty when getting data from a RAM into a 16374 has to be less than 2.0 - 1.1 - 0.4 = 0.5 ns, and that's without any skew allowance.

Right now, ffm04flop.q# is eating 0.69 ns, which does not meet the standard. That's with hub-and-spoke wiring, and a 10 mm per endpoint allowance for routing and vias. Cutting the allowance to 5 mm brings us to 0.52 ns, which is too little allowance, doesn't meet the need, and doesn't account for skew.

All this can be solved, but the focus instead should be to see if we even have the right parts on the board. This is to say, simulate functionality (with a relaxed tKQX if need be) and get the logic working. Even at that point, timing should not be everything, because the CPU still has a few additions that will doubtless move tiles relative to one another.

When returning to timing, hook up the already-written tree generator, and have the system find minimum-delay trees for the needed flows. (We don't need to look at output-to-output delays, etc.) The tree generator is not practical beyond a certain number of pins, or maybe even a certain number of nets, but it can be selectively enabled from the netlist file.

The next step in writing the simulator is perhaps a disk format for Python to C, or a sketch of a gate simulation in C.

The above is recorded, because I am moving my attention to the March 10 talk, and I will need to find my place.

Next USAF SBIR: April 21 pre-release, May 19 open, June 17 close.

Next NSF SBIR: March 4 close (will miss), June 3 close.

Next NSF SaTC (Secure and Trustworthy Cyberspace, 21-500): does not close
    for SMALL, MEDIUM, and EDU projects. You're good up to $1.2 M,
    but to get there you need a faculty appointment.

'''
#)
#raise SystemExit

# This program reads a human-typed (and somewhat human-friendly)
# netlist and translates it into:
#
#     a) a KiCad file for making circuit boards
#     b) a human-readable bill of materials for making purchases
#     c) a netlist formatted for the [in-progress] Dauug|36 simulator
#
# It also reads a component layout file comprised of "tiles" to
# give the simulator a rough estimate of all wiring.
#
# It also can adjust a 'kicad_pcb' file created by pcbnew to place
# components where they go. KiCad's PcbNew program doesn't get this
# information from our netlist, so it puts components in bad spots
# and leaves the user with a mouse and a big job. This feature
# lives in the kicad_pcb() function and is called by uncommenting
# the line that calls it at the end of this Python module.
#
# In this usage, a "net" means a set of pins that are wired together
# via circuit board traces. A "netlist" therefore describes an
# entire circuit board's connectivity, but does not describe where
# components are placed physically, or the resistance, inductance,
# capacitance, propagation delay, or radiative effects of a board.

#==============================================================================
# USER-CONFIGURABLE PARAMETERS
#==============================================================================

# netlist input filename
# XXX could be a command line argument
in_filename = 'feb.netlist'

# tile input filename
tile_filename = 'tiles'

# output filenames for KiCad and my bill of materials
# XXX could be a command line argument, or specified via directive
pcbnew_filename = 'generated.net'
bom_filename = 'generated.bom'

# user's .kind names for supported .glue gates
# XXX could be specified via a Phase 2 directive
# XXX pin names are likewise hard-coded in this module
bool_kind_names = dict(zip(
    'and or nand nor xor not notnot'.split(),
    'and or nand nor xor inv buf'.split()))

# user's pin names for board's vdd and ground
# XXX could be specified via directive
vdd_name, gnd_name = '.vdd', '.gnd'

# user's names for simulator's vdd and ground
# XXX could be specified via directive
sim_vdd_name, sim_gnd_name = ('rail', 'hi'), ('rail', 'lo')

#==============================================================================
# OTHER GLOBAL DATA
#==============================================================================

# EXPLANATION OF MACROS
#
# The input file specifies netlists via a series of whitespace-
# delimited tokens. No tokens can contain whitespace. Everything
# in this file is case-sensitive.
#
# A macro system provides substring replacement within tokens.
# Macros can have, but are not required to have, multiple 
# replacements, causing tokens to multiply in the Cartesian
# sense. All macros are expanded from left to right.
#
# Macro names begin and end with angle brackets; e.g., <mike>.
# Like tokens, macro names can't contain whitespace or internal
# angle brackets. There are four kinds of macros:
#
# KIND 1: USER SPECIFIED
#
# If the netlist preamble contains the definition:
#
#     .macro <mike> michael lee raymer
#
# Then these two lines are equivalent:
#
#     .named <mike>'s_part
#     .named michael's_part lee's_part raymer's_part
#
# So now there are THREE parts. Yes, it's okay to use ' in names,
# because we're very flexible except concerning whitespace.
#
# This first of the four kinds of macros is the only kind that
# you use a .macro directive in the netlist preamble to define it.
#
# KIND 2: RANGES OF NUMBERS
#
# Simple up and down integer counting is offered. Zero padding
# to the width of the lesser bound is supported. So equivalent:
#
#     .pins = doom<12-08>
#     .pins = doom12 doom11 doom10 doom09 doom08
#
# KIND 3: REPETITION
#
# Maybe you want to connect your doom pins to the same Vdd:
#
#     .wire doom<12-08> = <5*>the.vdd
#
# Note the .wire directive works pairwise across =, so the above is
# like typing 5 individual .wire directives for a single connection.
#
# KIND 4: SERIES OF NAMES
#
# Use commas in macros to expand the set. Like this:
#
#     .wire <mon,tue,wed,thu,fri>.y = <tue,wed,thu,fri,mon>.a
#
# This contrived example suggests a shift register that
# spans the weekdays. Five connections are created.
#
# IMPLEMENTATION NOTE
#
# When a netlist is read, it is parsed to find occurrences of
# macro kinds 2, 3, and 4 before other processing is done.
# These macros are converted into and stored as kind 1, making later
# expansion much simpler.

# EXPLANATION OF INPUT FILE FORMAT
#
# Lines are just lines, and there is no line-splicing mechanism.
# Semicolons introduce comments to the end of the current line.
# Contiguous whitespace within a line acts like a single space.
# Blank lines are ignored. Leading and trailing whitespace is ignored.
#
# The input file has a first, a second, and a third "phase."
#
# PHASE 1: DIRECTIVES DEFINING MACROS
#
#     .macro            ; See the above EXPLANATION OF MACROS.
#
# PHASE 2: DIRECTIVES DESCRIBING AND NAMING COMPONENTS
#
# The file needs adequate descriptions of each kind of component
# used. Each of these descriptions begins with the .kind directive,
# and a short block parameters that is exported to KiCad's netlist.
# What KiCad does with these parameters is beyond my desire to
# document. Thus a new component may look like this:
#
#     .kind dual-nand
#     .footprint Package_SO:VSSOP-8_2.4x2.1mm_P0.5mm
#     .lib 74xGxx
#     .part 74AUC2G00
#     .value 74AUC2G00
#     .order Texas Instruments SN74AUC2G00DCUR
#     .mm 2.1 3.2
#     .npins 8
#
# Three keywords here are NOT exported to KiCad:
#
#     .kind    This name determines the behavior that the simulator
#              will convey on this kind of component. Pick a name
#              that you like. (This name will need to be present
#              in the simulator's source code.)
#
#     .npins   This script wants this information (a) when checking
#              for human mistakes in the file, and (b) for determining
#              if a component is "real." It's sometimes helpful to
#              define "conceptual" components or circuit sections to
#              simplify and document wiring. Just omit .npins to use
#              this feature. Conceptual components do not undergo
#              rule checking and are not exported anywhere.
#
#     .order   This is the part number that will be given in the
#              generated bill of materials. You want this *really*
#              specific, with packaging specified and such.
#
#     .mm      Maximum bounding box of this component, in mm,
#              including pins but not the soldering footprint.
#              Specify width and then height. This information is
#              used for estimating circuit board size.
#
#     .rect    Linear pin geometry. Not suitable for BGA parts.
#              Seven numbers must appear on this line, denoted
#
#              dw     separation between the north-south rows of pins
#              dh     separation between the east-west rows of pins
#              pitch  pitch between adjacent pins
#              west   number of pins on left side
#              south  number of pins on "bottom" side (often 0)
#              east   number of pins on right side
#              north  number of pins on "top" side (often 0)
#
#              All linear measurements are between nominal pin centers.
#
#              If south and north are both 0, dh is irrelevant since
#              there are no pins to separate.
#
#              Pin 1 is the northmost on the west side. From there,
#              pins are consecutively numbered going counterclockwise.
#
#              This information is not used for board layout, but for using
#              placement to estimate trace length and trace capacitance.
#              If you have an odd part with a triangle geometry or something,
#              just fabricate something here that's good enough for
#              simulation. Part placement has more effect than pin position.
#
#     .noload  Indicates that the simulator does not model propagation
#              delay associated with capacitive loads of this component.
#
#     .load    Used to adjust propagation delay based on capacitance.
#              Four parameters in this order:
#
#              in_pF        capacitance of each input pin
#              io_pF        capacitance of each i/o pin
#              delay_pF     datasheet load capacitance for p_d
#              ns_pF_slope  estimated delay when capacitance > delay_pF
#
# The .value and .order directives permit multiple-token answers.
#
# For each kind of (real) component, the relationship between physical
# pins and their functions must be given. This is done with the .pins
# (not .npins) directive. The following example uses one line, although
# it could have been split into as many as eight:
#
#     .pins <1-8> = 1a 1b 2y gnd 2a 2b 1y vdd
#
# To the left of the = sign are the PIN NUMBERS, 8 in all. The ordering
# and number convention are yours to control. For BGA packages, these
# numbers are typically alphanumeric as opposed to merely numeric.
#
# For conceptual devices (not real devices), there are no physical
# pins, so the PIN NUMBERS are omitted. But the = sign stays.
#
# To the right of the = sign are PIN NAMES corresponding pairwise to the
# pin numbers on the left. There aren't many naming rules, but spaces
# and semicolons do not work in .
#
# For each of the pin names, their correct use needs to be given.
# Continuing this simple dual-nand example:
#
#     .power vdd gnd
#     .in    1a 1b 2a 2b
#     .out   1y 2y
#
#     .oute     ; this directive is for output pins w/ output enable
#     .bidir    ; this directive is for bidirectional pins
#     .nc       ; this directive is for pins to be left disconnected
#     .pull     ; this directive marks the weak side of pull resistors
#     .pflag    ; this directive indicates /availability/ of power
#               ; (.power indicates a /requirement/ for power)
#
# These unused directives (.oute through .pflag) can either have no 
# pin names or, more commonly, be omitted from component descriptions.
#
# The .pflag directive is used for conceptual devices, not physical ones.
# Typically one VDD and one ground net are marked with .pflag in circuits.
#
# Above are the names of the Phase 2 directives that *describe* components.
# They create instances of the comp_kind class in this script, one
# instance per kind.
#
# A final Phase 2 directive, .named, /instantiates and names/
# components. It creates instances of the comp_inst class in this
# script, one instance per actual component. In this example, we set
# up 100 NAND gates in 50 dual packages:
#
#     .named nand<01-50>
#
# PHASE 3: DIRECTIVE FOR WIRING PINS
#
# The .wire directive works pairwise across the = sign. For instance,
# to connect nand01.1y to nand33.1a and nand02.1y to nand33.1b:
#
#     .wire nand01.1y nand02.1y = nand33.1a nand33.1b
#
# Macros can get a lot done without much typing. Let's power those
# 100 NAND gates from the previous section:
#
#     .wire nand<01-50>.<vdd,gnd> = <50*>circuit.<vdd,gnd>
#
# When wiring circuits, the exact pin of the exact component is
# specified, always joined with a period without internal whitespace.
# Component names may also contain periods, but the rightmost period
# within a given token always sets off name of a pin as provided via
# the .pins directive for that component's kind.
#
# PHASE 3: DIRECTIVE FOR CLOCK DISTRIBUTION
#
# The .short directive is used to connect buffers in parallel for clock
# distribution. These parallel connections increase available drive
# and (kind of) maintain a single concept of the clock waveform and
# try to minimize skew. The KiCad operator will have to work some magic
# getting these pins connected with equal-length tracks. Here's the
# syntax (note there is no = sign):
#
#     .short 1866 source.out<0-3> dest.in<0-7>
#
# It's always .short, then an integer delay in picoseconds, and then
# some number of input and output pins. Order of pins does not matter,
# because they all get connected together.
#
# On the KiCad side, .short connects a single track to all of the pins;
# that is, the multiple outputs drive some large fan-out of inputs.
# The circuit board designer will have to combine the output lines to
# a common point via equal-length paths, and then distribute to each input
# by equal-length paths, such that the total track delay from any output
# to any input is (in this example) 1866 ps.
#
# The simulation software is not intended to handle such fan-ins and
# fan-outs, and even if it could, it knows nothing of waveforms, rise and
# fall times, signal distortion, or noise. Instead, what the simulator
# receives from this script is that destination is driven by just one
# of the sources with the specified path delay (1866 ps in this example),
# with the sources chosen round-robin to keep the fan-out within the
# simulator's configured fan_out_max.
#
# For design rule checking, only one output pin per node is checked.
# This prevents display of 'out with out' error messages.
#
# PHASE 3: RPN CIRCUIT SYNTHESIS
#
# The .glue directive permits the user to write a logic function
# in reverse Polish notation, with the line being interpreted as if
# the user had typed in the requiste .named and .wire directives
# to implement the gate instances and wiring. There are a few
# contingencies:
#
#     a) The only gates are nand, nor, not, and, or, notnot, xor.
#
#     b) The SN74AUC2Gxx chips are used,
#        where xx = 00, 02, 04, 08, 32, 34, 86 for the above.
#        Note that all of these chips are dual-gate packages,
#        and the reason I use these is that they are faster than
#        their single-gate counterparts, and easier to solder by
#        hand than their leadless quad-gate counterparts.
#
#     c) These devices must be declared in Phase 2 and have a .kind
#        known to this script, which is dual-nand, dual-nor, ....
#
#     d) not and notnot have one input; the others have two.
#
#     e) I didn't want to reserve the word "buf" to mean buffer,
#        so the gate is called "notnot".
#
# Usage looks like this (and operations must be lowercase):
#
#     .glue out.0 = in.0 not
#     .glue out.1 = in.0 in.1 xor
#     .glue out.2 = in.0 in.1 and in.2 xor
#     .glue out.3 = in.0 in.1 and in.2 and in.3 xor
#
# This example has 3 xor gates, 2 and gates, and an inverter.
# This script eliminates duplicate calculations, so the 'and'
# operation for out.2 is reused in out.3 without waste.
#
# Sometimes a high-current load needs to be driven, and a
# low-delay method for this is to use two copies of the last gate
# in parallel. This is signified by prepending '2*' to the boolean
# operation. For instance:
#
#     .glue parity.y = p.a p.b xor p.c 2*xor
#
# Doubling always is done in a single physical package.
# So the p.a p.b xor will not be on the same chip as the final,
# high-current xor. There is no support for tripling, quadrupling, etc.
#
# The .cohort directive prevents glue logic chips from combining
# gates from distant parts of the circuit board. To share the same
# chip, two nand gates must be glued within the same cohort. There
# are two forms of the directive:
#
#     .cohort           ; from here forward is a unique cohort
#     .cohort mypart    ; from here forward is a cohort named mypart
#
# The first form is simple: the following .glue forms a cohort where
# its gates can share one chip. Any unused halves are abandoned when
# the next cohort is entered. The second form lets us return to a 
# cohort later in a file and use up any unused halves of gates from
# that cohort. The reason we offer the second form is that the logical
# structure of your netlist may not always correspond with the physical
# layout of your board.
#
# The .cohort directive only affects the operation of the .glue
# directive, and .cohort's use is not mandatory (although very
# advisible in large circuits).
#
# PHASE 3: VISUALIZATION
#
#     .draw pin(s)      ; draw net of each specified pin
#
# This is a diagnostic feature. Each net will be output to a separate
# SVG file named after the pin used to select that net.
#
# PHASE ANY: DIRECTIVES FOR MULTILINE COMMENTS
#
# The .off and .on directives suspend and resume parsing.
# They need not be balanced in any fashion.
#
#     .off
#     How about those Buckeyes!
#     I mean THE OHIO STATE UNIVERSITY BUCKEYES.
#     Not some other, wannabe buckeyes.
#
#     Below is a circuit inspired by the Buckeyes.
#     .on
#
# PHASE ANY: OPERATOR COMMENTS
#
# The .attn directive writes to standard output, making the run
# "look unclean." So things are not forgotten.
#
#     .attn I disconnected the branch decision pin!
#

#==============================================================================
# NOTEWORTHY MAPPINGS
#==============================================================================

# The user's perspective of a "logical pin," which is the name of
# a component and its pin according to purpose (not position),
# is a string that reads like 'nand01.vdd'.
#
# In contrast, the program's perspective of a logical pin needs
# access to the <comp_inst> object and its <comp_kind> through that,
# as well as the name of the pin. So most of this code treats pins
# as the tuple (<comp_inst>, 'vdd'), where 'vdd' whatever the name
# of the logical pin happens to be, and <comp_inst> is a Python
# object from class comp_inst.

# Map from 'nand01.vdd' -> (<comp_inst>, 'vdd')
pin_names = {}

# Circuit nets in this programs are simply sets of logical pins
# that became merged via .wire directives. Prior to the first
# .wire, these are singleton sets of every logical pin on every named
# component.

# This map takes a logical pin, and gives the set of all logical pins
# to which it is wired, including itself. It's from
# (<comp_inst>, 'vdd') -> frozenset([ (<comp_inst>, 'vdd'), ... ])
nets = {}

# List of nets set up by the .short directive.
# The format is [(<comp_inst>, 'pin', 1866) ...], with the integer
# being picoseconds delay from all outputs to all inputs.
# Remember that any <comp_inst>, 'pin' within a net can look it up.
# Each of these tuples (minus the integer) is a member of nets.
# Entries in this list may not be unique, depending on what the person
# writing the netlist has done, and delay might not be consistent
# among repeated nets.
short_pins = []

# Sanitized version of short_pins as dictionary:
# { net : picoseconds, ... }, or more verbosely,
# { frozenset([ (<comp_inst>, 'vdd'), ... ]) : picoseconds, ... }
short_nets = {}

# These identify the net that is vdd and net that is gnd.
# Both are members of nets.
vdd_net = frozenset()
gnd_net = frozenset()

#==============================================================================
# NOTEWORTHY CLASSES
#==============================================================================

# A particular kind of object, such as 'dual-nand'.
class comp_kind(object):
    kinds = {}                      # 'name' : instance of comp_kind
    current = None                  # instance of this class

    # This does NOT check for a duplicate name. Caller needs to.
    def __init__(z, name):
        z.name = name               # 'sram-64k-16-bga-48'
        z.value = name              # can override; e.g. '0.1 uF'
        z.lib = ''                  # 'Device' or other KiCad lib
        z.part = ''                 # 'R' or other KiCad part
        z.order = ''                # ordering information (important!)
        z.footprint = ''            # 'Resistor_SMD:R_0402_1005Metric'
        z.mm = 0, 0                 # should NOT be zero
        z.npins = 0                 # 48
        z.pins = {}                 # 'vcc' : 'e1' (BGA-48 pin scheme)
        z.pin_kinds = {}            # 'vcc' : 'power'
        z.named = set()             # set(instances of comp_inst)
        z.l_num = l_num             # line number where this kind starts
        z.geom = {}                 # phys_pin_num : (x_rel_mm, y_rel_mm)
        z.sml = ''                  # 's'mall 'm'edium 'l'arge (for tiling)
        z.scape = False             # normally does not rotate landscape
        z.in_pF = None              # input pin capacitance or zero
        z.io_pF = 0                 # i/o pin capacitance or zero
        z.delay_pF = 0              # datasheet test jig capacitance
        z.ns_pF_slope = 0           # time penalty in ns per extra pF
        z.price = 0                 # about how many dollars for one
        comp_kind.kinds[name] = z
        comp_kind.current = z

    def __str__(z):                 # diagnostic
        return z.name

# A specific instance, such as 'nand19', of a kind of object.
class comp_inst(object):
    insts = {}                      # 'name' : instance of comp_inst

    # Neither does this check for a duplicate name. Caller needs to.
    def __init__(z, name, kind):
        z.name = name               # 'name'
        z.kind = kind               # instance of comp_kind
        z.place = 0, 0              # center position on board
        z.orient = None             # north (normal), south, east, west
        z.geom = {}                 # phys_pin_num : (x_abs_mm, y_abs_mm)
        z.label = None              # for drawings
        kind.named.add(z)
        comp_inst.insts[name] = z

        # Every pin is initially in a net by itself.
        for logi_pin in z.kind.pins:
            # singleton net with single pin (comp_inst, 'logi_pin')
            pair = z, logi_pin
            nets[pair] = frozenset([pair])

            # map ('component_name.logi_pin') to same pin
            full_name = '%s.%s' % (name, logi_pin)
            if full_name in pin_names:
                # This could happen in a weird naming scheme where both
                # component names and logical pin names contain dots.
                ng('%s is not a unique component.pin naming' % full_name)
            pin_names[full_name] = pair

    def __str__(z):                 # diagnostic
        return z.name

#==============================================================================
# MACRO PROCESSING CODE
#==============================================================================

# master dictionary of macro substitutions
macros = {}             # { 'a,b,c' : ['a', 'b', 'c'], ... }

# Create a KIND 1 macro from a .macro directive.
# The .macro keyword is already removed from the head of token list 'l'.
def make_macro(l):
    if len(l) < 2:
        ng('missing name of macro')
    mac = l[1]
    if not mac.startswith('<') or not mac.endswith('>'):
        ng('macro name needs < >')
    elif '>' in mac[:-1] or '<' in mac[1:]:
        ng('macro may not have internal < or >')
    else:
        macros[mac] = l[2:]

# Scan a line for macros that need to be created automatically,
# which are KIND 2, KIND 3, and KIND 4.
#
# Some examples:
#
#     <3-6>     (KIND 2) becomes   ['3', '4', '5', '6']
#     <08-11>   (KIND 2) becomes   ['08', '09', '10', '11']
#     <3-0>     (KIND 2) becomes   ['3', '2', '1', '0']
#     <4*>      (KIND 3) becomes   ['', '', '', '']
#     <m,w,a>   (KIND 4) becomes   ['m', 'w', 'a']
#
# This is only the scanning phase. During expansion later:
#
#     a<3-6>        becomes   ['a3', 'a4', 'a5', 'a6']
#     a<08-11>b     becomes   ['a08b', 'a09b', 'a10b', 'a11b']
#     <3-0>b        becomes   ['3b', '2b', '1b', '0b']
#     <4*>gnd.a     becomes   ['gnd.a', 'gnd.a', 'gnd.a', 'gnd.a']
#     <m,w,a>.vdd   becomes   ['m.vdd', 'w.vdd', 'a.vdd']
#
def find_numeric_macros_in_line(l):
    while '<' in l:
        ig, l = l.split('<', 1)
        if '>' not in l:
            return ng('unclosed macro')
        mac, l = l.split('>', 1)
        bracmac = '<' + mac + '>'

        if mac.endswith('*') and mac[:-1].isdigit():
            macros[bracmac] = [''] * int(mac[:-1])
            continue

        elif ',' in mac:
            macros[bracmac] = mac.split(',')

        elif '-' in mac:
            # 'vi' and 'de' mean 'from' and 'to' respectively,
            # without infringing Python's ill-chosen reserved word.
            vi, de = mac.split('-', 1)
            if not vi.isdigit() or not de.isdigit():
                continue
            viFmt, deFmt = '%%0%ii' % len(vi), '%%0%ii' % len(de) 
            vi, de = int(vi), int(de)
            if vi > de:
                macros[bracmac] = [deFmt % i for i in range(vi, de-1, -1)]
            else:
                macros[bracmac] = [viFmt % i for i in range(vi, de+1)]
            continue

# Recursively expand all macros in token list 'l'.
#
# Macros are expanded from left to right, resulting in the leftmost
# macro in a token being the most significant in ordering. For example:
#
#      <a,b,c><x,y,z>   expands to   ax ay az bx by bz cx cy cz
#                       and NOT      ax bx cx ay by cy az bz cz
#
# This order is critical for .wire directives to work correctly.
def expand_macros(l):
    while True:
        changed_line = False
        out = []
        for tok in l:
            changed_token = False
            if '<' not in tok:
                out.append(tok)
                continue
            vi = tok.index('<')
            if '>' not in tok[vi:]:
                return ng('unclosed macro')
            de = tok.index('>')
            if de < vi:
                return ng('malformed macro')
            m = tok[vi:de+1]
            if m not in macros:
                return ng('undefined macro %s' % m)
            for expand in macros[m]:
                out.append(tok.replace(m, expand, 1))
            changed_token = True
            changed_line = True
        if not changed_line:
            break
        l.clear()
        l.extend(out)

#==============================================================================
# GLUE LOGIC GENERATION
#==============================================================================

# for getting out of deep errors
class Stop(BaseException): pass

# number of inputs for boolean functions
#
# XXX Warning: adding gates that don't have symmetric inputs would
# introduce a bug in already_derived below. It's an unlikely scenario,
# because industry generally manufacture combinational gates without
# symmetric inputs. (It's also an industry shortcoming.)
arity = { 'and' : 2, 'or' : 2, 'nand' : 2, 'nor' : 2,
    'xor' : 2, 'not' : 1, 'notnot' : 1 }

# This is a list of components that are not yet using their second half.
# It's a dict indexed by (cohort, func) where both are strings.
spare_halves = {}

# remember tandemed pins that should not offend design rule checks
tandem_pins = set()

# Remember already derived (sub-)expressions so we do not build
# them more than once. Indexed by (cohort, func, frozenset(input-pins))
# and returns the output pin.
#
# The reason for the frozenset is that all our glue logic gates
# happen to have symmetric inputs. So we don't want to consider 
# order of inputs anywhere in these tree comparisons.
already_derived = {}

# Number of glue logic chips present, keyed by ('cohort', 'func')
glue_counts = {}

# parse and perform the .glue directive after the first (.glue) token
def make_glue(l):

    if len(l) < 3 or l[1] != '=':
        return ng('.glue syntax error')

    dest = pin_names.get(l[0])
    if not dest:
        return ng('unknown component.pin %s' % l[0])

    # the next ~25 lines is simply an RPN parser
    stack = []
    for t in l[2:]:
        double = False
        if t.startswith('2*'):
            double = True
            t = t[2:]
        if t in arity:              # check that t is a logic function
            if len(stack) < arity[t]:
                return ng('RPN stack underflow')
            a = b = stack.pop()
            if arity[t] == 2:
                a = stack.pop()     # a and b are pins

            # At this point, we want to add gate 't' with input
            # pins 'a' and 'b'. But if we've already done so,
            # we can simply use that pin instead.
            ad = already_derived.get((cohort, t, frozenset((a,b))))
            if not double and ad is not None:
                stack.append(ad)
                continue

            # We have not yet built this function. Build it,
            # and keep track of where we can access it again later.
            try:
                stack.append(solder_glue(a, b, t, double))
                already_derived[(cohort, t, frozenset((a,b)))] = stack[-1]
            except Stop:
                return              # error already recorded
            continue
        if double:
            t = '2*' + t            # odd but legal pin name, not a gate
        pin = pin_names.get(t)
        if not pin:                 # XXX could scan for more bad names
            return ng('unknown component.pin %s' % t)
        stack.append(pin)

    if len(stack) > 1:
        ng('incomplete glue expression')
    connect_pins(dest, stack[0])

# helper function to find or create a gate that does 'func'
# and connect its inputs
def solder_glue(connect_to_a, connect_to_b, func, double):

    # The object here is to fetch a comp_inst with the names of the
    # a, b, and y pins to be connected. For arity 1 gates, the a and
    # b pins have the same name, so connecting b will turn out to
    # be a no-op when the nets are merged by connect_pins().
    # There are three paths to this end, as follows:
    if double:
        # Path 1 is we need both halves of the same chip,
        # so we'll allocate it here and tandem them together.
        c = make_logic_chip(func)
        connect_pins((c, '1a'), (c, '2a'))
        if arity[func] == 2:
            connect_pins((c, '1b'), (c, '2b'))
        connect_pins((c, '1y'), (c, '2y'))
        tandem_pins.add((c, '2y'))      # to aid design rule checking
        a, b, y = '1a', '1b' if arity[func] == 2 else '1a', '1y'

    elif (cohort, func) in spare_halves:
        # Path 2 is to use the second half of a chip that is
        # not fully used in the same cohort.
        c = spare_halves.pop((cohort, func))
        a, b, y = '2a', '2b' if arity[func] == 2 else '2a', '2y'

    else:
        # Path 3 is to allocate a new chip, use the first half,
        # and reserve the second half for later use.
        c = make_logic_chip(func)
        spare_halves[(cohort, func)] = c
        a, b, y = '1a', '1b' if arity[func] == 2 else '1a', '1y'

    # We achieved our object, have our chip, and know the input pins.
    connect_pins(connect_to_a, (c, a))
    connect_pins(connect_to_b, (c, b))

    # The output pin is passed back for further processing.
    return c, y

# glue logic helper routine to allocate a logic chip
def make_logic_chip(func):

    # which component kind implements this function?
    kind_name = bool_kind_names.get(func)
    kind = comp_kind.kinds.get(kind_name)
    if not kind:
        ng('input defines no kind %s for glue function %s' %
                (kind_name, func))
        raise Stop

    # the new chip needs a name
    seq = glue_counts.get((cohort, func), 0)
    name = 'glue-%s-%s-%i' % (cohort, func, seq)
    glue_counts[(cohort, func)] = 1 + seq

    # now make the component
    c = comp_inst(name, kind)

    # connect power to this chip
    vdd = pin_names.get(vdd_name)
    gnd = pin_names.get(gnd_name)
    if not vdd or not gnd:
        ng('missing board vdd/gnd pin %s/%s' % (vdd_name, gnd_name))
        raise Stop
    connect_pins((c, 'vdd'), vdd)
    connect_pins((c, 'gnd'), gnd)

    return c

# glue finalizer to leave no floating inputs on unused halves
def rail_glue():

    # This test ensures vdd_name and gnd_name are legitimate.
    if not spare_halves:
        return

    # Rail floating inputs on all idle halves.
    sh = spare_halves.copy()
    while sh:
        (cohort, func), s = sh.popitem()
        # print('railing:  %-12s %s' % (cohort, s.name))
        if '2b' in s.kind.pins:
            connect_pins((s, '2a'), (s, '2b'))
        connect_pins((s, '2a'), pin_names[vdd_name])

#==============================================================================
# MAIN CODE
#==============================================================================

# Everything starts here. Open, parse, and process the input file.
# XXX Some syntax error checking could be factored and streamlined.
def go():
    phase2kws = '''.kind .order .footprint .mm .rect .npins .pins .named
.lib .part .value .load .noload .price'''.split() + ['.' + w for w in pin_kind_words]
    phase3kws = '.cohort .glue .wire .short .draw'.split()
    global l_num, cohort, vdd_net, gnd_net
    off = False
    cohort_fmt, last_auto = '[auto cohort %i]', 0
    cohort = cohort_fmt % last_auto

    # read the file
    try:
        lines = open(in_filename).readlines()
    except:
        ng('could not read ' + in_filename)
        check_for_errs()

    # remove comments, then set up numeric macros
    lines = [l.split(';', 1)[0] for l in lines]
    for l_num, l in enumerate(lines, 1):
        find_numeric_macros_in_line(l)

    # stop if there are problems
    check_for_errs()        

    # triage the directives
    phase = 1
    for l_num, l in enumerate(lines, 1):
        l = l.split()

        # blank lines and multiline comments at any time
        if not l:
            continue
        elif l[0] == '.on':             # resume ordinary processing
            off = False
        elif off:
            continue
        elif l[0] == '.off':            # all lines hence are comments
            off = True                  # until .on

        # operator warnings at any time
        elif l[0] == '.attn':
            print(end = '(%i) ' % l_num)
            print(' '.join(l[1:]))
            continue

        # macros in phase 1 only
        elif l[0] == '.macro':
            if phase > 1:
                ng('macro too late')
            make_macro(l)

        # component definitions (except .glue) in phase 1 only
        elif l[0] in phase2kws:
            if phase > 2:
                ng('component definition too late')
            phase = max(2, phase)
            make_compo(l)

        # .wire, .glue, .cohort, .short in phase 3 only
        elif l[0] in phase3kws:
            phase = max(3, phase)
            if l[0] == '.wire':
                make_wire(l)
            elif l[0] == '.short':
                make_short(l)
            elif l[0] == '.draw':
                make_draw(l)
            elif l[0] == '.cohort':
                last_auto += 1
                l.append(cohort_fmt % last_auto)
                if len(l) > 3:
                    ng('limit one token for cohort name')
                else:
                    cohort = l[1]       # unique if not specified by name
            else:                       # this is .glue
                make_glue(l[1:])
        else:
            ng('unrecognized directive')

    # when we're done, rail unused inputs from .glue directives
    rail_glue()

    # nets[] won't change from here, so find vdd and gnd.
    try:
        vdd_net = nets[pin_names.get(vdd_name)]
    except:
        ng('net %s is required but not found' % vdd_name)
    try:
        gnd_net = nets[pin_names.get(gnd_name)]
    except:
        ng('net %s is required but not found' % gnd_name)

    # stop if there are problems
    check_for_errs()        
    check_pin_balance()
    check_pin_kinds()
    check_part_details()
    reconcile_shorts()
    check_for_errs()
    check_design_rules()
    check_for_errs()

    # get part tilings for simulator
    read_tiles()
    check_for_errs()
    svg_floorplan('floorplan.svg')
    if False:
        svg_draw_all_nets()
    else:
        svg_draw_nets()

    # compute propagation time caused by wire lengths and high capacitive loads
    time_calcs()
    time_calcs_vdd_gnd()
    time_calcs_short()

    # write output files
    save_to_pcbnew()
    save_bom()

    print('That was very nice.')

# helper routine for Phase 2 (making components)
def make_compo(l):
    expand_macros(l)
    d = l[0]
    kinds = comp_kind.kinds
    kind = comp_kind.current

    if d == '.kind':
        if len(l) != 2:
            return ng('syntax error')
        k = l[1]
        if k in kinds:
            return ng('duplicate .kind')
        # all good, here is a new kind of component
        comp_kind(k)

    elif not kind:
        return ng('no .kind as of')

    elif d == '.footprint':
        if len(l) != 2:
            return ng('syntax error')
        if kind.footprint:
            return ng('duplicate .footprint')
        kind.footprint = l[1]

    elif d == '.lib':
        if len(l) != 2:
            return ng('syntax error')
        if kind.lib:
            return ng('duplicate .lib')
        kind.lib = l[1]

    elif d == '.part':
        if len(l) != 2:
            return ng('syntax error')
        if kind.part:
            return ng('duplicate .part')
        kind.part = l[1]

    elif d == '.value':             # .value allows internal, single spaces
        if len(l) < 2:
            return ng('syntax error')
        if kind.value and kind.value != kind.name:
            return ng('duplicate .value')
        kind.value = ' '.join(l[1:])

    elif d == '.order':             # .order allows internal, single spaces
        if len(l) < 2:
            return ng('syntax error')
        if kind.order:
            return ng('duplicate .order')
        kind.order = ' '.join(l[1:])

    elif d == '.mm':
        if kind.mm[0] * kind.mm[1]:
            return ng('duplicate .mm')
        try:
            dx, dy = float(l[1]), float(l[2])
        except:
            return ng('syntax error')
        if dx <= 0 or dy <= 0:
            return ng('invalid component dimensions')
        kind.mm = [dx, dy]

    elif d == '.rect':
        return do_rect(kind, l[1:])

    elif d == '.npins':
        if len(l) != 2 or not l[1].isdigit:
            return ng('syntax error')
        if kind.npins:
            return ng('duplicate .npins')
        kind.npins = int(l[1])

    elif d == '.pins':
        if l.count('=') != 1:
            return ng('syntax error')
        x = l.index('=')
        phys_pins, logi_pins = l[1:x], l[x+1:]
        if not phys_pins:
            # This is a "conceptual" part of the circuit,
            # not an object that is soldered on the board.
            for logi_pin in logi_pins:
                if logi_pin in kind.pins:
                    ng('duplicate logical pin %s' % logi_pin)
                kind.pins[logi_pin] = None      # there is no physical pin
        elif len(phys_pins) != len(logi_pins):
            return ng('unbalanced physical = logical pin count')
        else:
            pset = set(kind.pins.values())
            for phys_pin, logi_pin in zip(phys_pins, logi_pins):
                if phys_pin in pset:
                    ng('duplicate physical pin %s' % phys_pin)
                if logi_pin in kind.pins:
                    ng('duplicate logical pin %s' % logi_pin)
                kind.pins[logi_pin] = phys_pin

    elif d == '.noload':
        if len(l) != 1:
            return ng('syntax error')
        kind.in_pF = 0                 # 

    elif d == '.load':
        if len(l) != 5:
            return ng('syntax error')
        try:
            kind.in_pF, kind.io_pF, kind.delay_pF, kind.ns_pF_slope = \
                    map(float, l[1:])
        except:
            return ng('invalid parameters in .load')

    elif d == '.price':
        if len(l) != 2:
            return ng('syntax error')
        if kind.price:
            return ng('multiple price assertions')
        try:
            kind.price = float(l[1])
        except:
            return ng('inappropriate .price')

    elif d[1:] in pin_kind_words and d[0] == '.':
        for logi_pin in l[1:]:
            if logi_pin in kind.pin_kinds:
                ng('logical pin %s given multiple modes' % logi_pin)
                continue
            kind.pin_kinds[logi_pin] = d[1:]

    elif d == '.named':
        insts = comp_inst.insts
        for name in l[1:]:
            if name in insts:
                ng('duplicate component name')
                continue
            comp_inst(name, kind)

    else:
        ng('unknown keyword %s' % d)

# helper routine for Phase 1 (defining pin geometry)
def do_rect(kind, l):
    if len(l) != 7:
        return ng('.rect needs 7 parameters dx dy pitch west south east north')
    try:
        dx, dy, pitch = map(float, l[:3])
    except:
        return ng('.rect first 3 parameters are numeric dx dy pitch')
    try:
        w, s, e, n = map(int, l[3:])
    except:
        return ng('.rect last 4 parameters are pin counts around the package')

    if w<0 or s<0 or e<0 or n<0:
        return ng('negative pin counts are senseless')
    if dx > 16.1 or dy > 22.1:
        return ng('part size is larger than supported')

    # our coordinate system has x growing west, y growing north
    pin = 1

    # west side pins
    x, y = -dx/2, pitch * (w-1) / 2
    while w:
        kind.geom[pin] = x, y
        y -= pitch
        pin += 1
        w -= 1

    # south side pins
    x, y = -pitch * (s-1) / 2, -dy/2
    while s:
        kind.geom[pin] = x, y
        x += pitch
        pin += 1
        s -= 1

    # east side pins
    x, y = dx/2, -pitch * (e-1) / 2
    while e:
        kind.geom[pin] = x, y
        y += pitch
        pin += 1
        e -= 1

    # north side pins
    x, y = pitch * (n-1) / 2, dy/2
    while n:
        kind.geom[pin] = x, y
        x -= pitch
        pin += 1
        n -= 1

# helper routine for Phase 3 (wiring pins together)
def make_wire(l):
    expand_macros(l)
    if l.count('=') != 1:
        return ng('syntax error')
    x = l.index('=')
    lhs, rhs = l[1:x], l[x+1:]
    if len(lhs) != len(rhs):
        return ng('unbalanced pin lists across =')
    for s, t in zip(lhs, rhs):
        if ':0:' in (s, t):                     # :0: means "no connection"
            continue
        # s and t are component.pin strings
        ss = pin_names.get(s)
        tt = pin_names.get(t)
        if not ss:
            ng('unknown component.pin %s' % s)
        if not tt:
            ng('unknown component.pin %s' % t)
        connect_pins(ss, tt)

# helper routine for Phase 3 (shorting pins together)
def make_short(l):
    expand_macros(l)
    if '=' in l:
        return ng('extraneous =')
    if len(l) < 4:
        return

    # After .short, need to know length of outgoing wire traces in ps.
    # This has to be manually incorporated into the board layout.
    try:
        ps = int(l[1])
        if ps < 1:
            raise Exception
    except:
        ng('expecting path delay in ps after .short')
        return

    # Grab the first (comp_inst, 'pin') on this line.
    ss = pin_names.get(l[2])
    if not ss:
        ng('unknown component.pin %s' % ss)
        return

    # Remember this first pin so we can find its shorted net later.
    # Include the wire delay. List element is (comp_inst, 'pin', int_ps)
    short_pins.append((ss[0], ss[1], ps))

    # Connect the first pin to each remaining one.
    for t in l[3:]:
        tt = pin_names.get(t)
        if not tt:
            ng('unknown component.pin %s' % t)
            continue
        connect_pins(ss, tt)

# Helper routine for Phase 3 (XXX maybe we need a Phase 4),
# reconciling short_pins into short_nets
def reconcile_shorts():
    for c, pin, ps in short_pins:
        net = nets[(c, pin)]
        if net not in short_nets:
            short_nets[net] = ps
            continue
        if ps != short_nets[net]:
            ng('conflicting .short delay in net with %s.%s' % (c.name, pin))

# diagnostic routine for Phase 3 (drawing nets via SVG)
def make_draw(l):
    batch = []
    expand_macros(l)
    for comp_pin in l[1:]:
        pin = pin_names.get(comp_pin)
        if not pin:
            ng('unknown component.pin %s' % comp_pin)
            continue
        batch.append(pin)
    if batch:
        nets_to_draw.append(batch)

# helper function to merge nets of two pins
def connect_pins(pin1, pin2):
    if not (pin1 and pin2):
        return
    new_net = nets[pin1].union(nets[pin2])  # union of component/pin pairs
    for nn in new_net:
        nets[nn] = new_net                  # now all merged!

# Count parts in the circuit so I can buy them.
# Also estimate pin counts and circuit board area.
def save_bom():
    npins = 0

    # 'margins' is potential spacing in mm between component bounding boxes,
    # 'margin_sums' is resulting estimated board size in mm2.
    margins = list(range(1, 11))            # margin 0 would short out
    margin_sums = [0] * len(margins)

    try:
        f = open(bom_filename, 'w')
    except:
        print('unable to write to %s' % bom_filename)

    # Note parts we want to buy
    bom = []
    l1, l2, l3, l4 = 0, 0, 0, 0             # XXX could refactor
    total_parts = 0
    total_price = 0
    for k in comp_kind.kinds.values():
        if not k.npins:
            continue                        # conceptual type, not real
        n = len(k.named)                    # number of such components
        if not n:
            continue
        total_parts += n
        unit_price = n * k.price
        total_price += unit_price
        bom.append((str(n), '%6.2f' % unit_price, k.name, k.order))

        # keep track of column widths
        ll1, ll2, ll3, ll4 = map(len, bom[-1])
        l1, l2, l3, l4 = max(l1, ll1), max(l2, ll2), max(l3, ll3), max(l4, ll4)

        # keep track of board area
        dx, dy = k.mm
        if not dx * dy:
            ng('component %s area unknown' %  k.name)
        for i, m in enumerate(margins):
            area_per = (dx + m) * (dy + m)
            margin_sums[i] += n * area_per

        # how many pins to solder?
        npins += n * k.npins

    # output bom with columns that fit the counts
    for w, x, y, z in bom:
        print('  %*s    %-*s    %-*s    %-*s' %
                (l1, w, l2, x, l3, y, l4, z), file = f)

    # Here is an estimate of the work required by a full
    # continuity test by a flying probe. Verifying connectedness
    # of 'j' pins requires 'j' probe contacts. Verifying
    # unconnectedness of 'k' nets requires k(k-1)/2 probe contacts.
    n_nets = len(set(nets.values()))
    print('\nBuy solder for %i pins. There are %i parts in %i distinct nets.'
            % (npins, total_parts, n_nets), file=f)
    print('A full scan with 2 flying probes takes %i touches.' %
        (npins + n_nets * (n_nets - 1) / 2), file=f)
    print('These components cost %.2f in small lots.\n' % total_price, file=f)

    # output possible board areas
    print('margin    area     side', file=f)
    print('  mm      mm*2      mm', file=f)
    print('-----------------------', file=f)
    for i, m in enumerate(margins):
        print('%4i %9i %8i' % (m, round(margin_sums[i]),
            round(margin_sums[i]**.5)), file=f)

    # Note spare halves we *might* find uses for.
    if spare_halves:
        print(file = f)
        sortme = []
        for cohort, func in spare_halves:
            sortme.append("'%s' has a spare '%s'" % (cohort, func))
        sortme.sort()
        for s in sortme:
            print(s, file=f)

    # List of components for placing in carriers.
    print('\nComponents:\n', file=f)
    comp_names = []
    for k in comp_kind.kinds.values():
        if not k.npins:
            continue                        # conceptual type, not real
        for comp in k.named:
            comp_names.append(comp.name)
    comp_names.sort(key = pin_name_key)
    for name in comp_names:
        print(name, file=f)

    f.close()

# helper function to convert (<comp_inst>, 'pin') to 'comp.pin'
def name(pin):
    return '%s.%s' % (pin[0].name, pin[1])

#==============================================================================
# TILING AND GEOMETRIC CODE
#==============================================================================

# helper function to read the tilings
# XXX no support for per-component orientation at this time
def read_tiles():
    global l_num
    col = None
    have = set()        # 'component-name'
    tiles = {}          # (col, row) : [<comp_inst>]

    try:
        lines = open(tile_filename).read().splitlines()
    except:
        return ng('unable to open tiling file %s' % tile_filename)

    for l_num, l in enumerate(lines, 1):
        
        l = l.split(';')[0].strip()
        if not l: continue
        # comments are removed! sorry, no .off / .on
        l = l.split()

        if l[0] == '.tile':
            if len(l) != 2:
                ng('syntax error')
                continue
            col, row = l[1].lower(), ''
            while col[-1].isdigit():
                row = col[-1] + row
                col = col[:-1]
            if not col.isalpha() or not row.isdigit() or len(col) > 1:
                ng('.tile format looks like like D6')
                continue
            col = ord(col) - 96
            row = int(row)          # row and col are indexed from 1
            if (col, row) not in tiles:
                tiles[(col, row)] = []
            continue

        # This provides a forcible component orientation.
        orient = None
        if l[0] in ['north:', 'south:', 'east:', 'west:']:
            orient = l[0][0]
            l = l[1:]

        cname = l[0]
        if len(l) != 1:
            ng('tile list requires one component per line')
            continue
        if col is None:
            ng('component requires .tile previously')
            continue
        if cname == ':0:':          # empty small space written as :0:
            tiles[(col, row)].append(None)
            continue
        comp = comp_inst.insts.get(cname)
        if comp is None:
            ng('unknown component %s' % cname)
            continue
        if comp in have:
            ng('component %s already tiled' % cname)
            continue
        have.add(comp)
        tiles[(col, row)].append(comp)
        comp.orient = orient

    # Did we get them all?
    oops = set(comp_inst.insts.values()) - have
    for oop in oops:
        if not oop.kind.npins:
            continue
        if oop.kind.name == 'hide-bead':
            continue            # kludge XXX
        ng('component %s missing from tiling' % oop.name)

    # parameters for tiling (use longer dimension first on tile)
    margin = 4.1                # min. req. mm component separation
    tile = 21.1, 27.1           # 5 mm larger than SRAM on both dimensions

    # We need to know how wide and tall the board is.
    # XXX Hopefully the user started numbering at A1.
    n_rows = max([row for (col, row) in tiles])
    n_cols = max([col for (col, row) in tiles])

    # Establish size limits for each kind of component.
    # This is not a very general scheme. The basic idea is that
    # we can fit one SRAM, two 16374s (or 16244s), nine small parts,
    # or a 16374 (or 16244) with five small parts in a tile.
    # The margin is for beads, bycaps, and my soldering ability.
    large = tile[0] - margin, tile[1] - margin
    medium = tile[0]*2/3 - margin, tile[1]*2/3 - margin
    small = tile[0]/3 - margin, tile[1]/3 - margin

    # assign small, medium, large, and default rotation to each kind
    for k in comp_kind.kinds.values():
        if not k.npins:
            continue
        if k.name == 'hide-bead':
            continue            # kludge XXX
        yes, scape = fits(k.mm, small)
        if yes:
            k.sml, k.scape = 's', scape
            continue
        yes, scape = fits(k.mm, medium)
        if yes:
            k.sml, k.scape = 'm', scape
            continue
        yes, scape = fits(k.mm, large)
        if yes:
            k.sml, k.scape = 'l', scape
            continue
        ng('component type %s is too big' % k.name)

    # This is where to put small components without and with a medium present.
    s9 = (-1,-1), (-1,0), (-1,1), (0,-1), (0,0), (0,1), (1,-1), (1,0), (1,1)
    s5 = (-1,1), (0,1), (1,-1), (1,0), (1,1)

    # for each tile, group and place components by s/m/l size
    for (col, row), t in tiles.items():

        # tile's x and y centers
        tx = tile[0] * (col - (n_cols + 1) / 2)
        ty = tile[1] * ((n_rows + 1) / 2 - row)

        # group parts by previously computed size
        histo = { 's' : [], 'm' : [], 'l' : [] }
        for c in t:
            if c is None:               # empty position :0: is small
                histo['s'].append(c)
                continue
            histo[c.kind.sml].append(c)
        s, m, l = map(histo.get, ['s', 'm', 'l'])

        # labels depend on space availability
        for c in t:
            if c is None:
                continue                # empty position :0:
            c.label = '`' + c.kind.name if c in s else c.name

        # tile arrangements are expressed in twelfths
        if l:
            order = (0, 0),
        elif len(m) > 1:
            order = (0, 3), (0, -3)
        elif m:
            if s:
                order = (-2, 2), (4, 4), (4, 0), (-4, -4), (0, -4), (4, -4)
            else:
                order = (0, 0),
        else:
            order = (-4, 4), (0, 4), (4, 4), (-4, 0), (0, 0), (4, 0), \
                    (-4, -4), (0, -4), (4, -4)

        work, order = list(l) + list(m) + list(s), list(order)
        while work:
            if not order:
                ng('too many components in tile %s%s' % (chr(64+col), row))
                break
            c, (dx, dy) = work.pop(0), order.pop(0)
            if c is None:
                continue                # empty position :0:
            c.place = tx + dx * tile[0] / 12, ty + dy * tile[1] / 12
            if not c.orient:
                c.orient = 'w' if c.kind.scape or len(m) == 2 else 'n'

    # Give positions to every physical pin.
    for c in comp_inst.insts.values():
        k = c.kind
        for p in range(1, k.npins+1):
            dx, dy = reorient(k.geom[p], c.orient)
            c.geom[p] = c.place[0] + dx, c.place[1] + dy

# This routine returns (fit, rotate), where 'fit' is whether the
# part fits within the limit, and 'rotate' is whether rotation is advised.
def fits(partxy, limitxy):

    limit_scape = limitxy[0] > limitxy[1]   # limit is landscape
    part_scape = partxy[0] > partxy[1]      # part is landscape

    if limit_scape ^ part_scape:
        partxy = partxy[::-1]               # give part best fit

    return partxy[0] <= limitxy[0] and partxy[1] <= limitxy[1], \
            limit_scape ^ part_scape

# Reorient a given pin from north to north, east, south, or west.
def reorient(dxdy, orient):
    dx, dy = dxdy
    if orient == 'w':
        dx, dy = -dy, dx
    elif orient == 's':
        dx, dy = -dx, -dy
    elif orient == 'e':
        dx, dy = dy, -dx
    return dx, dy

#==============================================================================
# TIME ESTIMATION CODE
#==============================================================================

more_ns = {}            # ('src_comp', 'src_pin', 'dst_comp', 'dst_pin')) : ns

def time_calcs():

    ns_per_mm = 0.006   # speed of signal on a wire trace, 6 pS / mm

    # These nets are driven by two pins, so drive is increased.
    tandem_nets = set(nets[pin] for pin in tandem_pins)

    # Loop over the nets in the system.
    # Note that 'nets' is full of dupes, as it's indexed by pin, and
    # is de-duplicated on this line.
    for s in set(nets.values()):

        # DIAGNOSTIC to see a list of large nets
        if False and len(s) > 5:
            syn = []
            for comp, pin in s:
                syn.append('%s.%s' % (comp, pin))
            syn.sort(key = pin_name_key)
            print('%3i' % len(syn), ' '.join(syn))

        # Timing for vdd, gnd, and .short nets is handled elsewhere.
        if s == vdd_net or s == gnd_net or s in short_nets:
            continue

        # If this net has double drive, remember that before overwriting s.
        double_drive = s in tandem_nets

        # Remove irrelevant pins.
        s = set((c, pin) for (c, pin) in s
                if c.kind.npins and c.kind.name != 'hide-bead')
        # s is a connected set of (comp_inst, 'pin').

        # Extract physical pin coordinates from this set.
        pin_cloud = set(pin_coords(c, pin) for (c, pin) in s)

        # Find the hub of these coordinates and the total length
        # of the spokes. All length measures in this program are mm.
        hub, hubsum = find_hub(pin_cloud)

        # Get the total capacitance on this net.
        # We'll individually back out capacitance of driving pins later.
        netcap = 0                     # unit is pF
        for (c, pin) in s:
            pkind = c.kind.pin_kinds[pin]
            if pkind == 'in':
                netcap += c.kind.in_pF
            elif pkind == 'oute' or pkind == 'bidir':
                netcap += c.kind.io_pF

        # That's the component capacitance.
        # Add an extra .08 pF for each mm of spoke length.
        netcap += .08 * hubsum

        # Pin-to-pin nanosecond delay.
        for (sc, spin) in s:
            if sc.kind.pin_kinds[spin] not in ('out', 'oute', 'bidir'):
                continue                # not a source pin

            # How much load capacitance beyond this pin's datasheet spec?
            # Don't forget that the source's output capacitance was
            # already included in the spec.
            excess_pF = max(0, netcap - sc.kind.io_pF - sc.kind.delay_pF)

            # How many ns to add for this load capacitance?
            excess_cap_ns = sc.kind.ns_pF_slope * excess_pF

            # XXX Maybe just count outputs, so that 2 isn't only possibility?
            if double_drive:
                excess_cap_ns *= 0.5

            # How many mm between this source pin and the hub?
            sdist = knight_dist(hub, pin_coords(sc, spin))

            for (dc, dpin) in s:
                if (dc, dpin) == (sc, spin):
                    continue            # same pin
                if dc.kind.pin_kinds[dpin] not in ('in', 'bidir'):
                    continue                # not a destination pin

                # How many mm total from src to here via hub?
                dist = sdist + knight_dist(hub, pin_coords(dc, dpin))

                # How much delay on account of the distance?
                dist_ns = ns_per_mm * dist

                # Finally, our ns delay over and above datasheet.
                add_ns = excess_cap_ns + dist_ns

                # We also need an upper bound on the speed a signal
                # can propagate at. We assume there is no wire
                # capacitance and use straight-line distance.
                straight_line = straight_dist(
                        pin_coords(sc, spin), pin_coords(dc, dpin))
                min_ns = ns_per_mm * straight_line                

                more_ns[(sc.name, spin, dc.name, dpin)] = (add_ns, min_ns)

                if False:               # human-readable time data
                    print('%5.2f  %5.1f  %s.%s -> %s.%s' % 
                            (add_ns, min_ns, sc.name, spin, dc.name, dpin))

# Add time "calculations" for the vdd and gnd rails.
# In reality we merely fill in 100 ps, but having this information
# causes the appropriate output for the simulator.
def time_calcs_vdd_gnd():

    two_passes = (sim_vdd_name, vdd_net), (sim_gnd_name, gnd_net)

    # Do both rails with their corresponding networks.
    for (rail_name, rail_pin), net in two_passes:

        # For each physical component that has input pins,
        # put in a timing from the rail to that pin.
        for (c, pin) in net:
            if not c.kind.npins:
                continue
            if c.kind.pin_kinds[pin] != 'in':
                continue

            # Minimum and maximum delay is 100 ps.
            more_ns[(rail_name, rail_pin, c.name, pin)] = (.1, .1)

# Add time "calculations" for the .short nets.
# In the simulator, there aren't /actually/ shorts, but inputs are
# assumed to be correctly ganged, and every input pin is driven by
# a single output pin.
def time_calcs_short():

    for net, ps in short_nets.items():
        pins_in = []
        pins_out = []

        for c, pin_name in net:
            if not c.kind.npins:
                continue
            pkind = c.kind.pin_kinds[pin_name]
            if pkind in ('out', 'oute'):
                pins_out.append((c.name, pin_name))
            elif pkind == 'in':
                pins_in.append((c.name, pin_name, ps))
            else:
                ng('unexpected pin kind %i on shorted %s.%s' %
                        (pkind, c.name, pin_name))
                continue

        # Drive every input pin with the fixed ps delay given.
        # Use a round-robin to assign output pins.
        for (c_in, p_in, ps), (c_out, p_out) in zip(pins_in, cycle(pins_out)):
            more_ns[(c_out, p_out, c_in, p_in)] = (ps / 1000, ps / 1000)

# Given a set of pin coordinates, find the "hub" that minimizes the sum
# of knight distances between the hub and each pin coordinate in the set.
# This is not a perfect model for board traces, but it's at least a
# starting point for "is this already acceptable?" modeling questions.
# Returns ((x, y), sum_of_dists).
def find_hub(pts):
    hub = 0, 0
    hubsum = float('inf')

    progress = True
    while progress:
        progress = False
        for i in range(-1, 2):
            for j in range(-1, 2):
                h = hub[0] + i, hub[1] + j
                s = sum(knight_dist(h, p) for p in pts)
                if s < hubsum:
                    hub, hubsum, progress = h, s, True
    return hub, hubsum

# turn (comp_inst, 'pin') into physical (x, y)
def pin_coords(c, pin):             # pin looks like 'vdd'
    pin = c.kind.pins[pin]          # pin looks like '14'
    pin = int(pin)                  # pin looks like 14
    return c.geom[pin]              # return value is (x_abs_mm, y_abs_mm)

# This is like Manhattan distance, but street angle granularity is 45 degrees.
# This is also where we factor in some extra mm for endpoint routing.
def knight_dist(p1, p2):
    extra_mm = 10
    l, s = abs(p2[0]-p1[0]), abs(p2[1]-p1[1])
    if l < s:
        l, s = s, l
    return l - s + 1.414 * s + extra_mm

# Here is the minimum conceivable distance between two pins.
def straight_dist(p1, p2):
    return ((p2[0]-p1[0])**2 + (p2[1]-p1[1])**2)**.5

# Generator function to return all spanning trees of connected set 's'.
# A tree's format is ( (e0, e1), (e0, e2), (e0, ...), (e1, ...) ).
# In other words, each tree is specified as an edgelist.
# Verified output against https://oeis.org/A000272
def spans(s):
    list_s = list(s)
    n = len(s)                      # size of original set
    c2i = combinations(range(n), 2) # all combinations of 2 indices
    save = {}                       # fast convert mask to combinations
    masks = []

    for i, j in c2i:                # go through edges
        mask = 1 << i | 1 << j      # make a bitmask to represent edge
        masks.append(mask)          # save mask
        save[mask] = i, j           # show how to get from mask to edge

    want = (1 << n) - 1             # desired OR of n-1 masks
    found = 0
    for some in combinations(masks, n-1):
        if reduce(operator.__or__, some, 0) == want:
            # Right number of vertices, but might not be a tree. Check.
            v = list(some)
            h, t = v[0], v[1:]      # head and tail
            conn = True             # head is connected
            while t:
                tt = [i for i in t if i & h]
                if not tt:
                    # this graph is not connected
                    conn = False
                    break
                h |= tt[0]
                t.remove(tt[0])
            if not conn:
                continue
            found += 1
            edges = [(list_s[save[mask][0]], list_s[save[mask][1]]) for mask in some]
            yield edges

#==============================================================================
# OUTPUT TIME ESTIMATES AND CONNECTIVITY TO SIMULATOR
#==============================================================================

def sim_write(fName):

    try:
        f = open(fName, 'w')
    except:
        return ng('unable to write simulator topology file %s' % fName)

    # Output a little header.
    print('; Automatically generated by run.netlist.py', file=f)
    print('; This is meant to be called within a human-written test script.',
        file=f)

    # first name the components we have
    for k in comp_kind.kinds.values():
        if not k.npins: continue
        if not k.named: continue
        print('k %s' % k.name, file=f)
        for name in k.named:
            print('n %s' % name, file=f)

    # To allow a smaller file by removing dupes, we will
    # pivot from: ('src_comp', 'src_pin', 'dst_comp', 'dst_pin') : ns
    # to: { 'src_comp' : { 'dst_comp' : [ (src_pin, dst_pin, ns) ... ] } }
    v = {}
    for (src_comp, src_pin, dst_comp, dst_pin), ns in more_ns.items():
        if src_comp not in v:
            v[src_comp] = {}
        if dst_comp not in v[src_comp]:
            v[src_comp][dst_comp] = []
        v[src_comp][dst_comp].append((src_pin, dst_pin, ns))

    # Write out the signal timings (in ps).
    # This does NOT include .short nets, because their timings were not
    # calculated. Nor the vdd and gnd nets.
    for src_comp, dests in v.items():
        print('f %s' % src_comp, file=f)
        for dst_comp, sigs in dests.items():
            print('t %s' % dst_comp, file=f)
            for src_pin, dst_pin, ns in sigs:
                print('s %s %s %s' %
                        (src_pin, dst_pin, int(.5+1000*ns[0])), file=f)

    f.close()

#==============================================================================
# VALIDATION, DESIGN RULE CHECK, AND ERROR REPORTING CODE
#==============================================================================

# list of errors found during processing
errs = []               # [ 'line 3: syntax error', ... ]

# Add a newly found error to the list.
# The line number usually comes from global variable l_num,
# but can be overridden when appropriate.
def ng(why, line_num = None):
    errs.append('line %i: %s' % (line_num if line_num else l_num, why))

# If any errors were found, print them and quit.
def check_for_errs():
    if not errs: return
    for e in errs:
        print(e)
    raise SystemExit

# Diagnostic: display all netlists
def show_nets():
    s = set(nets.values())
    for i, net in enumerate(s):
        print('\n%i' % i)
        sortme = ['%s.%s' % (compo.name, logi_pin) for compo, logi_pin in net]
        sortme.sort()
        for ok in sortme:
            print('  ' + ok)

# Make sure part information is complete.
def check_part_details():
    for k in comp_kind.kinds.values():
        if not k.lib and k.npins:
            ng('component %s needs .lib specified' % k.name, k.l_num)
        if not k.part and k.npins:
            ng('component %s needs .part specified' % k.name, k.l_num)
        if not k.footprint and k.npins:
            ng('component %s needs .footprint specified' % k.name, k.l_num)

# Make sure pin counts match pins present.
def check_pin_balance():
    for k in comp_kind.kinds.values():
        nones = list(k.pins.values()).count(None)
        if k.npins:
            if nones:
                ng('kind %s is missing some physical pins' % k.name,
                        k.l_num)
            if k.npins != len(k.pins):
                ng('kind %s .npins does not match .pins in count' % k.name,
                        k.l_num)
        else:
            if nones != len(k.pins):
                ng('kind %s .npins does not match physical pin count' % k.name,
                        k.l_num)

# Ensure a 1:1 correspondence between named logical pins and a directive
# that says what kind of pin each is. This only applies to real components,
# not conceptual ones. Also verify that a pin geometry is present.
def check_pin_kinds():
    for k in comp_kind.kinds.values():
        if k.npins != len(k.geom):
            ng('disparity between .npins and .rect for %s' % k.name)
        if not k.npins:
            continue
        diff = set(k.pins).symmetric_difference(k.pin_kinds)
        if diff:
            ng('disparity for %s between .pins and kinds for %s'
                    % (k.name, ' '.join(diff)), k.l_num)

# This is a simple design rule sanity check.
# It can't foresee all problems, but can catch some common ones.
#
# Here are the types of pairs we might find in a net, and
# how we feel about those pairs appearing:
#
#         nc    in    out   oute  pull  bidir power solo float
# nc      ok    ok    ok    ok    ok    ok    ok    ok   ok
# in      ok    ok    ok    ok    ok    ok    ok    NG   NG
# out     ok    ok    NG    NG    NG    NG    NG    ok   n/a
# oute    ok    ok    NG    ok    ok    ok    NG    ok   n/a
# pull    ok    ok    NG    ok    NG    ok    NG    ok   n/a
# bidir   ok    ok    NG    ok    ok    ok    NG    NG   n/a
# power   ok    ok    NG    NG    NG    NG    ok    NG   n/a
# solo    ok    NG    ok    ok    ok    NG    NG    n/a  ok
# float   ok    NG    n/a   n/a   n/a   n/a   n/a   ok   n/a
#
# Here 'solo' means the net has only one pin.
# Here 'float' means the net is undriven.
#
# Also, two pflags in one net means vdd and gnd might be connected,
# so we test for it.
#
def check_design_rules():
    nogood = [x.split('.') for x in '''out.out out.oute out.pull out.bidir
out.power oute.power pull.pull pull.power bidir.power in.solo in.float
bidir.solo power.solo pflag.pflag'''.split()]

    for bignet in set(nets.values()):
        d = {word : 0 for word in pin_kind_words}

        # we need to ignore conceptual components
        net = { (compo, logi_pin) for (compo, logi_pin) in bignet
                if compo.kind.npins }

        # except that we need to detect pflag
        # XXX maybe power should come from a real component?
        powernet = { (compo, logi_pin) for (compo, logi_pin) in bignet
                if compo.kind.pin_kinds.get(logi_pin) == 'pflag' }

        if not net:
            continue

        d['solo'] = 1 if len(net) == 1 else 0
        d['pflag'] = 1 if len(powernet) else 0

        pins = []
        trouble = []
        for compo, logi_pin in net:
            if (compo, logi_pin) in tandem_pins:
                continue                # ignore parallel gate second output
            k = compo.kind
            pin_kind = k.pin_kinds[logi_pin]
            pins.append('%-5s  %s.%s' % (pin_kind, compo.name, logi_pin))
            if pin_kind != 'pflag':
                d[pin_kind] += 1
        d['float'] = not (d['out'] or d['oute'] or d['pull'] 
                or d['bidir'] or d['power'])
        for i, j in nogood:
            if i == j and d[i] > 1 or i != j and d[i] and d[j]:
                trouble.append('%s with %s' % (i, j))
        if d['power'] and not d['pflag']:
            trouble.append('power is not connected (via pflag)')
        if not trouble:
            continue
        if len(trouble) == 1 and trouble[0] == 'out with out' \
                    and bignet in short_nets:
            continue                    # ordinary effect of .short
        ng('design rule fail (%s):' % ', '.join(trouble))
        for p in pins:
            ng('    ' + p)

#==============================================================================
# SVG DRAWING (diagnostic feature)
#==============================================================================

# The xlink thing is because Safari and certain editors have been
# slow to adopt plain href.
head_t = '''
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="100%%" height="100%%" viewBox="%.3f %.3f %.3f %.3f" style="background: #000">
<g text-anchor="middle" stroke="none" font-family="monospace">
<rect x="%.3f" y="%.3f" width="%.3f" height="%.3f" fill="#fff"/>
'''.strip()

foot_t = '''
</g>
</svg>
'''.strip()

text_t = '''
<text x="%.3f" y="%.3f" fill="#000" font-size="%s">%s</text>
'''.strip()

rect_t = '''
<rect x="%.3f" y="%.3f" width="%.3f" height="%.3f" fill="#0ff"/>
'''.strip()

# A provision to repeat geometries that have already been drawn.
# Href serial numbers derive from the size of this dict.
gate_reuser = {}        # (orient, frozenset(pin_geoms)) : 'href_target'

# This is a list of batches, where a batch is a list of (comp_inst, 'pin').
# The pins of a batch select the nets that are drawn, with all of that
# batch's nets being drawn to the same SVG file. The next batch is drawn into
# a different file. XXX Ideally the batch filename's pin is the same as
# the pin specified in the batch, but I'm not sure if that's the case now.
nets_to_draw = []

line_colors = cycle(
        '#a6cee3 #1f78b4 #b2df8a #33a02c #fb9a99 #e31a1c #fdbf6f #ff7f00 #cab2d6 #6a3d9a #eeee88 #b15928'.split()
        ).__next__

# Start drawing an SVG file.
# Our bounding box spec is different from SVG's.
# We use west, south, east, north with y increasing to the north.
def svg_begin(fname, west, south, east, north):
    global gf           # graphics output file

    try:
        gf = open(fname, 'w')
    except:
        return ng('unable to open %s for svg write' % fname)

    # we negate all y values passing to SVG
    xmin = west
    width = east - west
    ymin = -north
    height = north - south
    print(head_t % (xmin, ymin, width, height, xmin, ymin, width, height),
            file=gf)

# Finish drawing an SVG file.
def svg_end():
    print(foot_t, file=gf)
    gf.close()

# Insert a comment. Not "safe" (robust with respect to quoting).
def svg_comment(t):
    print('<!-- %s -->' % t, file=gf)

# Insert a gap.
def svg_nl(t):
    print(file=gf)

# Insert a circle.
def svg_circle(x, y, r, color = '#800'):
    print('<circle cx="%.3f" cy="%.3f" r="%.3f" fill="%s"/>' %
            (x, -y, r, color), file=gf)

# Insert a line.
def svg_line(x0, y0, x1, y1, color = '#080'):
    print('<line x1="%.3f" y1="%.3f" x2="%.3f" y2="%.3f" stroke="%s" stroke-width=".50"/>' % (x0, -y0, x1, -y1, color), file=gf)

# Insert a text label.
def svg_text(text, x, y, size = 2):
    move_down = size/3
    if text.startswith('`'):
        move_down += 3.5
        text = text[1:]
    print(text_t % (x, -y + move_down, size, text), file=gf)

# Insert a centered rectangle.
def svg_center_rect(x, y, w, h, landscape):
    if landscape:
        w, h = h, w
    x -= w/2
    y += h/2
    print(rect_t % (x, -y, w, h), file=gf)

# Draw a single component. We'll have a "long" and "short" way.
# The long way's purpose is to validate how the short one works.
def svg_gate_long(c):
    w, h = c.kind.mm
    for pin in range(1, c.kind.npins + 1):
        color, diam = ('blue', .25) if pin == 1 else ('black', .15)
        svg_circle(c.geom[pin][0], c.geom[pin][1], diam, color)

# This version has more code, but reuses components within drawings.
# Output drops from 398265 to 40038 bytes (and can gzip to 3493).
def svg_gate_short(c):
    k = c.kind
    key = c.orient, frozenset(k.geom.values())
    name = gate_reuser.get(key)
    if not name:
        name = 'q%s' % len(gate_reuser)
        print('<defs><g id="%s">' % name, file=gf)
        gate_reuser[key] = name

        for pin in range(1, k.npins + 1):
            color, diam = ('red', .25) if pin == 1 else ('black', .15)
            dx, dy = reorient(k.geom[pin], c.orient)
            svg_circle(dx, dy, diam, color)

        print('</g></defs>', file=gf)

    print('<use xlink:href="#%s" x="%.3f" y="%.3f"/>' %
            (name, c.place[0], -c.place[1]), file=gf)

# Draw the gate floorplan. If a net is specified (by pin), highlight it.
# 'batch' is a list of (comp_inst, 'pin').
def svg_floorplan(svg_fname, batch = None):
    gate_reuser.clear()
    svg_begin(svg_fname, -110, -110, 110, 110)

    if batch:
        for net in batch:
            svg_net(net)

    for c in comp_inst.insts.values():
        if not c.kind.npins:
            continue
        if c.kind.name == 'hide-bead':
            continue                # kludge xxx
        svg_gate_short(c)
        if c.label:
            svg_text(c.label, c.place[0], c.place[1])

    svg_end()

# Draw a net using the hub/spoke model.
# 'net_pin' identifies the net and is a (comp_inst, 'pin').
def svg_net(net_pin):
    # Try to change color between nets.
    color = line_colors()

    # Expand net_pin from (comp, 'pin') to all physical pins of this net.
    net = set((c, pin) for (c, pin) in nets[net_pin] if c.kind.npins)

    # Get their physical coordinates. We will not label them,
    # but only draw the computed tracks.
    pin_cloud = set(pin_coords(c, pin) for (c, pin) in net)

    # Locate the hub. XXX drawing suppressed
    (hx, hy), hubsum = find_hub(pin_cloud)
#   svg_circle(hx, hy, .6, color)

    # Draw from hub to each pin, adhering to 45-degree track granularity
    # (knight steps).
    for (x, y) in pin_cloud:
#       svg_circle(x, y, .6, 'blue')        # XXX circle suppressed
        xl, yl = abs(x - hx), abs(y - hy)
        if xl > yl:
            # x is the long dimension; calculate where turn is
            tx, ty = x + ((xl - yl) * (-1 if x > hx else 1)), y
        else:
            # y is the long dimension; calculate where turn is
            tx, ty = x, y + ((yl - xl) * (-1 if y > hy else 1))
        svg_line(x, y, tx, ty, color)
        svg_line(tx, ty, hx, hy, color)

# Draw the requested nets.
def svg_draw_nets():
    for batch in nets_to_draw:              # batch: [(comp_inst, 'pin'), ... ]
        comp, pin = batch[0]                # name file after first pin
        name = '%s.%s' % (comp.name, pin)
        name = name.replace('#', '!')       # easier on bash
        name = 'net-%s.svg' % name
        svg_floorplan(name, batch)

# Draw all the nets. This function is enabled via source code modification,
# and is used for special presentations. It interferes with regular drawing.
# XXX This could be refactored, as it forces svg_draw_nets to find something
# we already found and disguised.
def svg_draw_all_nets(with_power = False):

    # Grab one pin to identify each net.
    bunches = []
    for net in set(nets.values()):
        # net is frozenset([(comp_inst, 'pin'), ...]).
        for comp, pin in net:               # arbitrarily grab a pin
            break
        bunches.append((comp, pin))

    # Overwrite whatever drawing work would ordinarily be done.
    # XXX Huge nets are not drawn, due to the 'count' limit.
    del nets_to_draw[:]                     # interference w/ reg. drawing
    more_pins = []
    for comp, pin in bunches:
        if not comp.kind.npins:
            continue
        net = nets[(comp, pin)]
        if not with_power and (net == vdd_net or net == gnd_net):
            continue
        more_pins.append((comp, pin))
    nets_to_draw.append(more_pins)

    svg_draw_nets()

# SVG rounding helper routine
def R(x):
    return round(x, 3)

#==============================================================================
# KICAD OUTPUT SECTION
#==============================================================================

# This part of the script has to honor KiCad's Pcbnew file format
# as well as the behavior of the KiCad 5.0.2 source code. As a
# result, this section may look cryptic. If you want to browse the
# KiCad 5.0.2 source in relation to what is here, the following
# files are where I suggest turning:
#
#     common/pcb.keywords
#     include/pcb_lexer.h
#     include/netlist_lexer.h
#     pcbnew/netlist_reader.h
#     pcbnew/kicad_netlist_reader.cpp
# 

# This is a map from the terms I chose for our netlist file into
# Pcbnew's more verbose file format:
pin_kind_words = dict(zip('in bidir out oute nc power pull pflag'.split(),
        'input BiDi output 3state NotConnected power_in passive n/a'.split()))

# This routine writes out our netlist in the format that KiCad's
# Eeschema (schematic editor) writes and KiCad's Pcbnew (board layout
# editor) reads. We omit some bloat that Pcbnew ignores, specifically
# the (design) and (libraries) sections, as we wouldn't know how to
# correctly write these unspecified elements.
#
# May be important to know: this file format is a lot like S-expressions
# from the Scheme and Lisp world.
def save_to_pcbnew():

    try:
        f = open(pcbnew_filename, 'w')
    except:
        print('could not write %s' % pcbnew_filename)
        return

    print('(export (version D)', end='', file=f)

    # (components block
    print('\n  (components', end='', file=f)
    for c in comp_inst.insts.values():
        k = c.kind
        if not k.npins:
            continue                        # not a physical component
        print('\n    (comp (ref %s)' % quoth(c.name), end='', file=f)
        print('\n      (value %s)' % quoth(k.value), end='', file=f)
        print('\n      (footprint %s)' % quoth(k.footprint), end='', file=f)
        print('\n      (libsource (lib %s) (part %s))' %
                (quoth(k.lib), quoth(k.part)), end='', file=f)
        print('\n      (sheetpath (names /) (tstamps /))', end='', file=f)
        print('\n      (tstamp %s)' % quoth('ts.' + c.name), end='', file=f)
        print(')', end='', file=f)          # close (comp
    print(')', end='', file=f)              # close (components

    # (libparts block
    print('\n  (libparts', end='', file=f)
    for k in comp_kind.kinds.values():
        if not k.npins:
            continue                        # not a physical kind
        print('\n    (libpart (lib %s) (part %s))' %
                (quoth(k.lib), quoth(k.part)), end='', file=f)
        print('\n      (pins', end='', file=f)
        logi2phys = list(k.pins.items())
        logi2phys.sort(key = pin_name_key)
        for logi_pin, phys_pin in logi2phys:
            print('\n        (pin (num %s) (name %s) (type %s))'
                    % (phys_pin, logi_pin,
                    pin_kind_words[k.pin_kinds[logi_pin]]), end='', file=f)
        print(')', end='', file=f)          # close (pins
    print(')', end='', file=f)              # close (libparts

    # (nets block
    print('\n  (nets', end='', file=f)
    s = set(nets.values())
    i = 0
    for net_ in s:
        net_name, net = physical_only(net_)
        if not net:
            continue
        i = i + 1                           # only enumerate non-empty nets
        print('\n    (net (code %i) (name %s)' % (i, quoth(net_name)),
                end='', file=f)
        for compo, logi_pin in net:
            phys_pin = compo.kind.pins[logi_pin]
            print('\n      (node (ref %s) (pin %s))' % 
                    (quoth(compo.name), quoth(phys_pin)), end='', file=f)
        print(')', end='', file=f)          # close (net
    print(')', end='', file=f)              # close (nets
    print(')', file=f)                      # close (export

    # KiCad's format doesn't have a newline at the end of the last
    # line, but I'm a maverick and wrote one anyway.

# Helper routine for KiCad export.
# It adds quotation marks to string values that call for them.
# See also common/richio.cpp in the KiCad 5.0.2 source tree.
def quoth(s):
    d = dict(zip('"_(_)_#_\\_\t_\n_\r_ '.split('_'),
        '\"_(_)_#_\\\\_\\t_\\n_\\r_ '.split('_')))
    set1 = set(c for c in s)
    set2 = set(c for c in d)
    if s and not set1.intersection(set2):
        return s
    return '"' + ''.join(d.get(c, c) for c in s) + '"'

# Helper routine for KiCad export. It does two things.
#
# 1. This takes a net (set of connected pins), and removes any that
#    are merely conceptual, because non-physical parts are not relevant
#    to KiCad's circuit board layout.
#
# 2. A unique name is automatically generated for the net being
#    returned. This name is a guess at best, and I don't really 
#    care (at least yet) what the name is. This guess, however, DOES
#    consider non-physical parts attached to the net, because these
#    at least represent some kind of human annotation.
def physical_only(net):
    outnet = set()
    name = None
    for compo, logi_pin in net:
        if not compo.kind.npins:
            # If more than one non-phys, only one will prevail in name
            name = '/%s/%s' % (compo.kind.name, logi_pin)
            continue
        outnet.add((compo, logi_pin))
    if not name:
        # outnet guaranteed to have at least one component in this case
        compo, logi_pin = list(outnet)[0]
        name = 'Net-(%s-Pad-%s)' % (compo.name, compo.kind.pins[logi_pin])
    return name, outnet

# My KiCad file output has its pins sorted by physical pin number.
# This is probably non-essential. But doing this sorting is a little
# tricky, because (a) widths of numbers may change as counts increase,
# and (b) BGA pins are letter-number pairs, such as 'F13'. This
# function is a simple derived key to make such sorts work. The
# names are assumed to have the form LETTERSopt + DIGITSopt.
def old_pin_name_key(x):
    x = x[1]
    nondigits = ''.join(c for c in x if not x.isdigit())
    digits = '0' + ''.join(c for c in x if x.isdigit())
    return (nondigits, int(digits))

# Names that end with numbers probably reflect a numeric sequence
# that should be reflected in sort order.
def pin_name_key(x):
    if type(x) is tuple:            # (comp, 'pin')
        x = x[1]                    # otherwise, just 'pin'
    digits = ''
    while x[-1:].isdigit():
        digits = x[-1] + digits
        x = x[:-1]
    if not digits:
        digits = '0'
    return (x, int(digits))

#==============================================================================
# KICAD PCB ADJUSTMENT
#==============================================================================

# The workflow for KiCad is first to run this program to make
# the netlist, then use KiCad's PcbNew program to fetch the
# netlist (menu Tools ... Load Netlist), and save it as a
# board file. The board will be reasonable, but the netlist
# file did not know of any component placement. At this point,
# call this magic function below to load Kicad's board file
# (file extension .kicad_pcb) and update its preliminary
# component positions and rotations with those from this 
# script's board tiling input.
#
# The editing done by this function is written to be idempotent;
# e.g., once it has altered a file, further applications will
# not corrupt the initial alteration.
def kicad_pcb(in_fname, out_fname):

    # Order in which components appear. ['name', 'name2', ...]
    order = []
    rotates = { 'n' : '0', 'e' : '90', 's' : '180', 'w' : '270' }

    if in_fname == out_fname:
        return ng('kicad_pcb() requires a separate output file')
    try:
        f = open(in_fname)
    except:
        return ng('unable to open PCB file %s for read' % in_fname)

    # The S-expressions of this file are not parsed as such,
    # as all we seek is to update a few positions and rotations.
    # What's here might not require adaptation for KiCad 
    # versions other than 5.0.2.

    # The file gives the position of each component prior to
    # identifying the component. Okay, this first read of the
    # file will determine component order. We're parsing a
    # line that looks like:
    # '    (fp_text reference alpha0 (at 0 -12.95) (layer F.SilkS)'
    prefix = '    (fp_text reference '
    for l in f.readlines():
        if l.startswith(prefix):
            order.append(l[len(prefix):].split(' ', 1)[0])

    # Rewind and go through again. Write output as we go.
    f.seek(0)
    try:
        of = open(out_fname, 'w')
    except:
        return ng('unable to open PCB file %s for write' % in_fname)
    alter = False
    for l in f.readlines():

        # The first line of each component to change looks like:
        # '    (at 95.905001 126.145001)'
        # Dimensions mercifully are already in mm. We negate the
        # y value. For rotation, we add a third value which is
        # 0 (may omit) for north, 90 east, 180 south, 270 west.
        if l.startswith('    (at '):
            c = comp_inst.insts[order.pop(0)]
            x, y = c.place[0], -c.place[1]
            r = rotates[c.orient]
            l = '    (at %.2f %.2f %s)\n' % (x, y, r)

            # Begin altering rotations that follow.
            alter = True

        # Once we see a line that looks like this, we cease
        # changing rotations in the file:
        # '    (model ${KISYS3DMOD}/Package_QFP.3dshapes/PQFP-100_14x20mm_P0.65mm.wrl'
        elif l.startswith('    (model '):
            alter = False

        # The rotation change is to replace anything that looks like
        # ' (at -5.525 -11.125) ' to include the needed rotation.
        # But only within our range of interest.
        elif alter:
            if ' (at ' in l:
                # extract digits following ' (at '
                lpos = l.index(' (at ') + 5
                rpos = l.index(')', lpos)
                nums = l[lpos:rpos].split()
                if len(nums) == 2:
                    nums.append('0')
                nums[2] = r
                l = l[:lpos] + ' '.join(nums) + l[rpos:]

        # Write the line as it now reads.
        print(l, file = of, end='')

    f.close()
    of.close()

#==============================================================================
# END OF SOURCE
#==============================================================================

go()
#kicad_pcb('../pcb/noname.kicad_pcb.in', '../pcb/noname.kicad_pcb')
#show_nets()
sim_write('../netsim/connectivity.ns')
check_for_errs()                    # one last time, for kicad_pcb()
