/*
    compo.h
    This module provides building blocks for simulated pins and components.
*/

/*
    TODO fan_out_max was bumped from 9 to 14 so that work on logical
         functioning could continue. That was 9 June 2021, and the
         offending signals came from the initialization subsystem.
         Consider rework for a more reasonable fan-out, but wait
         until better guidance for timing is available.
*/
enum {
    fan_out_max = 14,       // maximum fan-out
    fan_in_max = 5,         // maximum fan-in

    // bitmask for driving in_mux.in
    hiz = 0,                // offline
    lo = 1,                 // logical zero
    hi = 2,                 // logical one
    uncert = 4,             // transition pending
};

typedef long time;          // simulation elapsed time in picoseconds

/*
    A "fire" happens when a pin might be driven high and low simultaneously.
    Fires necessarily happen, but they have to be very short so that no
    components are damaged. Here is a structure for keeping track.
*/
struct fire {
    int flame;              // if nonzero, a fire is in progress
    time when;              // time (ps) of fire start / fire quenched
    time longest;           // length of longest fire (ps)
    time sep;               // shortest separation observed between two fires
};

/*
    Input pins have a lot going on. Two highlights:

    (a) Because input pins can be driven by tristate or bidirectional
        devices, they can have several sources. The logic values and
        impedances are tabulated from each source and combined into
        an aggregate called 'val'.

    (b) Of the thousands of input pins in a simulation, very few
        change at a time. To speed computation, pins that need to
        be checked are moved to a singly-linked "dirty" list. The
        end of the list is marked with a self-reference instead of
        NULL, because NULL is used instead to indicate a pin is clean
        (and therefore not on the list).

        The order of the dirty list is irrelevant, because all its
        pins are processed at the same picosecond. This does not
        result in a race condition, because all simulator operations
        take at least one picosecond. So there is no circumstance
        where two pins can observe each other's behavior while they're
        on the dirty list together.
*/
typedef struct in_pin {
    char *name;             // pin name, with component
    struct in_pin *dirty;   // NULL if clean, next pin if dirty, self if last
    struct Prefix *comp;    // component this pin belongs to
    time last_change;       // last time pin changed value
    int n_in;               // number of sources driving this pin
    int val;                // aggregate input value from sources
    int changed;            // set when 'val' changes; user clears
    char in[fan_in_max];    // signals from this pin's sources
    char zin[fan_in_max];   // sources' impedance; 1 = high, 2 = low
    struct fire fire;       // short circuit statistics
} in_pin;

/*
    This struct adds functionality to an input pin to make it a
    clock input pin.

    When initializing an in_namer (see c_374.c, c_1g74.c, c_sram.c),
    pass the address of this struct's .pin, not the struct itself.
*/
typedef struct {
    time last_rise;
    time last_fall;
    int rising;             // set when clock rises; user clears
    in_pin pin;             // associated input pin
} in_clock;

/*
    Output pins are essentially lists of the input pins they drive.

    Because input pins need separate 1-byte buffers for every output
    that drives them, the exact bytes to modify are stored with the
    output pins for speed reasons.

    XXX Please standardize spelling of high-Z for easier searches.
*/
typedef struct {
    char *name;                 // pin name, with component
    int n_out;                  // number of sinks this pin drives
    int disabled;               // pin is high-Z if true
    int val;                    // immediate output value when enabled
    int path_ps[fan_out_max];   // picoseconds to reach each sink
    char *out[fan_out_max];     // point within 'in' on each sink
    char *zout[fan_out_max];    // point within 'z' on each sink
    in_pin *in[fan_out_max];    // point to each sink
} out_pin;

/*
    Here are two structs to associate names to pins within a
    component type. They are built into arrays within the
    comp/c_*.c modules, with NULL pointers marking the end of
    each array.

    The 'pin' pointers here are in fact offsets into the
    component instance where the pins are. Although this gets the
    job done, the resulting initializations in comps/c_*.c as well
    as numerous pointer arithmetic cases in connect.c are untidy
    and oppressive to read. The offsetof() macro could have helped,
    but it isn't available on older compilers. Macros for helper
    functions written for this application could also have helped.
*/
typedef struct {
    char *name;
    in_pin *pin;
} in_namer;

typedef struct {
    char *name;
    out_pin *pin;
} out_namer;

/*
    An array of these in connect.c sets up the kinds of components
    the simulation comprises. 

    init() takes an integer parameter that an customize a component,
    such as differentiate between AND gates and XOR gates in c_2gxx.c,
    or specify the width and size of an SRAM in c_sram.c.

    be() simulates the component's behavior if the component is dirty
    (that is, any of its inputs have changed).

    tweak() implements the 'tweak' keyword in a *.ns test script to
    do something to a component (or have the component do something)
    in an already-running simulation.
*/
typedef struct {
    char *name;                     // name of component kind
    int size;                       // memory to alloc per instance
    int init_parm;                  // to customize initialization
    void (*init)(void *, int);      // call to initialize
    void (*tweak)(void *, char *);  // call to modify a live parameter
    in_namer *in_names;             // array defining input pins
    out_namer *out_names;           // array defining output pins
    void (*be)(void *);             // call to simulate
} Kind;

/*
    All component structs in comps/c_*.c begin with this prefix
    to provide a little polymorphism.

    Much like a pin's dirty list, the end of a component's dirty 
    list is marked with a self-reference instead of NULL, so that
    NULL can be used instead to indicate a component is clean
    (and therefore not on the list).
*/
typedef struct Prefix {
    char *name;                     // name of component
    Kind *kind;                     // kind of component
    int trace;                      // individually trace this component
    struct Prefix *next;            // linked list of all components
    struct Prefix *dirty;           // linked list of dirty components
} Prefix;
