/*
    c_2gxx.c

    This is the simulation's implementation of several glue logic gates
    from Texas Instruments' SN74AUC2Gxx family, providing AND, OR,
    NAND, NOR, INV, BUF, and XOR. These devices have 2 gates per IC.
*/

enum {                      // available logic functions
    glue_buf, glue_inv, glue_and, glue_or, glue_nand, glue_nor, glue_xor,
};

typedef struct {
    in_pin a, b;            // input pins for a single gate
    out_pin y;              // output pin a single gate
} c_2gxx_half;

// dual logic gate (NAND, AND, NOR, OR, XOR, BUF, INV)
typedef struct {
    Prefix pre;             // fields common to all components
    int logic_fn;           // one of glue_and, glue_xor, etc.
    int t_pd_min;           // XXX want variable delay someday
    int t_pd_max;           // propagation delay
    c_2gxx_half h[2];       // the two gates on the IC
} c_2gxx;

// Here are names and byte offsets of input pins within the struct.
// The 1b and 2b pins are not actually present on BUF and NOT gates.
in_namer in_namer_2gxx[] = {
    "1a", &((c_2gxx *) NULL)->h[0].a,
    "1b", &((c_2gxx *) NULL)->h[0].b,
    "2a", &((c_2gxx *) NULL)->h[1].a,
    "2b", &((c_2gxx *) NULL)->h[1].b,
    NULL, NULL,
};

// Here are names and byte offsets of output pins within the struct.
out_namer out_namer_2gxx[] = {
    "1y", &((c_2gxx *) NULL)->h[0].y,
    "2y", &((c_2gxx *) NULL)->h[1].y,
    NULL, NULL,
};

// Here is the initialization for component instances.
void init_2gxx(void *v, int logic_fn)
{
    int i;

    static struct {
        int logic_fn;
        int t_pd_min;       // ps at 15 pF load
        int t_pd_max;
    } T[] = {
        glue_buf, 500, 1200,
        glue_inv, 500, 1200,
        glue_and, 500, 1000,
        glue_nand, 500, 1000,
        glue_or, 400, 1100,
        glue_nor, 500, 1300,
        glue_xor, 300, 1300,
    };

    c_2gxx *g = v;

    g->logic_fn = logic_fn;

    // For buffers and inverters, force unused B inputs to valid logic level.
    if (logic_fn == glue_inv || logic_fn == glue_buf) {
        rail_in_pin(&g->h[0].b, lo);
        rail_in_pin(&g->h[1].b, lo);
    }

    // Supply propagation delay range from manufacturer datasheets.
    for (i = 0; i < sizeof(T) / sizeof(*T); i++)
        if (logic_fn == T[i].logic_fn) {
            g->t_pd_min = T[i].t_pd_min;
            g->t_pd_max = T[i].t_pd_max;
            return;
        }

    ng("init_2gxx t_pd overrun");
}

/*
    Here is a diagnostic mechanism for creating gates with propagation
    delays that can be adjusted during simulation. Such adjustable delays
    can be used to set up setup/hold/removal/recovery time tests for flip-
    flops, temperature-dependent behavior, debugging the event simulator
    proper, and so on.
*/
void tweak_2gxx(void *v, char *parm)
{
    c_2gxx *g = v;

    if (!parm) return;

    if (*parm == 'd')           // change propagation delay on BOTH GATES
        g->t_pd_max = number(1 + parm, "");
    else
        ng("unknown c_2gxx tweak %s", parm);
}

/*
    This performs basic AND-NAND-OR-NOR-INV-BUF-XOR with propagation
    delay following a change to one or more inputs. If the inputs
    are not sufficiently known to determine the output, the output
    will be 'uncert' instead of 'lo' or 'hi'.
*/
void be_2gxx_half(
    c_2gxx_half *h,
    int logic_fn,
    int t_pd,
    int trace
) {
    int y = uncert;                     // most frequent case
    int a, b;

    ifnot (h->a.changed || h->b.changed)
        return;
    h->a.changed = h->b.changed = 0;
    a = h->a.val, b = h->b.val;

    // These simple gates have 3 output states: hi, lo, uncert.
    switch (logic_fn) {

        case glue_buf:
            if (a == lo || a == hi)
                y = a;
            break;

        case glue_inv:
            if (a == lo)
                y = hi;
            else if (a == hi)
                y = lo;
            break;

        case glue_and:
            if (a == lo || b == lo)
                y = lo;
            else if (a == hi && b == hi)
                y = hi;
            break;

        case glue_or:
            if (a == hi || b == hi)
                y = hi;
            else if (a == lo && b == lo)
                y = lo;
            break;

        case glue_nand:
            if (a == lo || b == lo)
                y = hi;
            else if (a == hi && b == hi)
                y = lo;
            break;

        case glue_nor:
            if (a == hi || b == hi)
                y = lo;
            else if (a == lo && b == lo)
                y = hi;
            break;

        case glue_xor:
            if (a == lo) {
                if (b == lo)
                    y = lo;
                else if (b == hi)
                    y = hi;
            }
            else if (a == hi) {
                if (b == lo)
                    y = hi;
                else if (b == hi)
                    y = lo;
            }
            break;

        default:
            ng("!strange gate %i", logic_fn);
    }

    // Do propagation delay and drive output.
    pin_output_after_delay(&h->y, y, t_pd);

    // Note this trace output message appears early, because
    // the signal has to pass through the event queue.
    if (trace)
        printf("%s going %s\n", h->y.name, pin_val_text(y));
}

// Simulate the action of one of those dual logic gate devices
// after a change to one or more of its inputs.
void be_2gxx(void *v)
{
    c_2gxx *g = v;
    int i;

    for (i = 0; i < 2; i++)
        be_2gxx_half(&g->h[i], g->logic_fn, g->t_pd_max, g->pre.trace);
}
