/*
    c_374.c

    This is the simulation's implementation of a Texas Instruments
    SN74AUC16374 16-bit flip-flop. It's grouped as two 8-bit sections.
*/

// Here is an eight-bit slice of the part.
typedef struct {
    char *name;
    in_clock clk;           // clock input
    in_pin oe_;             // active-low output enable
    in_pin d[8];            // input pins
    out_pin q[8];           // output pins
} c_374_half;

// Here is the entire part.
typedef struct {
    Prefix pre;             // fields common to all components
    c_374_half flop[2];     // two 8-bit slices
} c_374;

// Here are names and byte offsets of input pins within the struct.
in_namer in_namer_374[] = {
    "1oe#", &((c_374 *) NULL)->flop[0].oe_,
    "1clk", &((c_374 *) NULL)->flop[0].clk.pin,
    "1d0", &((c_374 *) NULL)->flop[0].d[0],
    "1d1", &((c_374 *) NULL)->flop[0].d[1],
    "1d2", &((c_374 *) NULL)->flop[0].d[2],
    "1d3", &((c_374 *) NULL)->flop[0].d[3],
    "1d4", &((c_374 *) NULL)->flop[0].d[4],
    "1d5", &((c_374 *) NULL)->flop[0].d[5],
    "1d6", &((c_374 *) NULL)->flop[0].d[6],
    "1d7", &((c_374 *) NULL)->flop[0].d[7],
    "2oe#", &((c_374 *) NULL)->flop[1].oe_,
    "2clk", &((c_374 *) NULL)->flop[1].clk.pin,
    "2d0", &((c_374 *) NULL)->flop[1].d[0],
    "2d1", &((c_374 *) NULL)->flop[1].d[1],
    "2d2", &((c_374 *) NULL)->flop[1].d[2],
    "2d3", &((c_374 *) NULL)->flop[1].d[3],
    "2d4", &((c_374 *) NULL)->flop[1].d[4],
    "2d5", &((c_374 *) NULL)->flop[1].d[5],
    "2d6", &((c_374 *) NULL)->flop[1].d[6],
    "2d7", &((c_374 *) NULL)->flop[1].d[7],
    NULL, NULL,
};

// Here are names and byte offsets of output pins within the struct.
out_namer out_namer_374[] = {
    "1q0", &((c_374 *) NULL)->flop[0].q[0],
    "1q1", &((c_374 *) NULL)->flop[0].q[1],
    "1q2", &((c_374 *) NULL)->flop[0].q[2],
    "1q3", &((c_374 *) NULL)->flop[0].q[3],
    "1q4", &((c_374 *) NULL)->flop[0].q[4],
    "1q5", &((c_374 *) NULL)->flop[0].q[5],
    "1q6", &((c_374 *) NULL)->flop[0].q[6],
    "1q7", &((c_374 *) NULL)->flop[0].q[7],
    "2q0", &((c_374 *) NULL)->flop[1].q[0],
    "2q1", &((c_374 *) NULL)->flop[1].q[1],
    "2q2", &((c_374 *) NULL)->flop[1].q[2],
    "2q3", &((c_374 *) NULL)->flop[1].q[3],
    "2q4", &((c_374 *) NULL)->flop[1].q[4],
    "2q5", &((c_374 *) NULL)->flop[1].q[5],
    "2q6", &((c_374 *) NULL)->flop[1].q[6],
    "2q7", &((c_374 *) NULL)->flop[1].q[7],
    NULL, NULL,
};

// Here are timing parameters for the SN74AUC16374, in picoseconds.
// While not used for this module, the minimum propagation delay, 
// enable time, and disable time are 700, 700, and 500 ps respectively.
enum {
    t_w_374 = 1900,     // min clock high or low
    t_su_374 = 600,     // setup time
    t_h_374 = 400,      // hold time
    t_pd_374 = 2200,    // max propagation delay
    t_en_374 = 2200,    // max enable time
    t_dis_374 = 2200,   // max max disable time
};

// The simulation's power-up state for this component is not
// guaranteed to match the real world.
void init_16374(UNUSED void *v, UNUSED int z)
{
    // There is nothing to do.
}

// Clock a forced value to the output pins.
// The format for 'parm' is "a=255" or "b=255", where any byte value
// from 0 to 255 may be specified.
void tweak_16374(void *v, char *parm)
{
    c_374 *g = v;
    c_374_half *h;
    int i, byte;

    if (*parm == 'a')
        h = &g->flop[0];
    
    else if (*parm == 'b')
        h = &g->flop[1];

    else
        goto bad;

    if (parm[1] != '=')
        goto bad;

    byte = number(parm + 2, "0");
    if (byte & ~255)
        goto bad;
    for (i = 0; i < 8; i++)
        pin_output_after_delay
                (&h->q[i], byte & 1 << i ? hi : lo, t_pd_374 - t_h_374);
    return;

bad:
    ng("!bad 16374 tweak '%s' on '%s'\n", parm, g->pre.name);
}

// Called one hold time after a clock rise, this queues either the
// input pin states (if stable) or 'uncert' (if not stable)
// to the output pins.
static void hold_time_callback_374(c_374_half *h, void *unused, int trace)
{
    int i;
    int val;

    for (i = 0; i < 8; i++) {
        if (h->d[i].last_change + t_su_374 <= h->clk.last_rise)
            val = h->d[i].val;
        else {
            val = uncert;
            ng("setup/hold time violation on %s", h->d[i].name);
        }
        pin_output_after_delay(&h->q[i], val, t_pd_374 - t_h_374);
    }

    if (trace)
        trace_input_pins(h->d, 8);
}

// This simulates either of the chip's 8-bit flip-flops.
void be_16374_half(c_374_half *h, int trace, int unit)
{
    int i;

    // Note time of any changes on d0 ... d7.
    for (i = 0; i < 8; i++) {
        if (!h->d[i].changed)
            continue;
        h->d[i].changed = 0;
    }

    // Change outputs to reflect output enable (#) pin.
    // There may be a race condition between propagation delay and 
    // enable/disable. The latter is scheduled 1 ps earlier so as not
    // to accidently change the output pins to an old value.
    if (h->oe_.changed) {
        h->oe_.changed = 0;
        if (trace)
            trace_input_pins(&h->oe_, 1);

        // Eight bits to enable or disable
        output_impedance_after_delay(h->q, h->oe_.val == lo ? -8 : 8, t_en_374);
    }

    // On rising clock, take action one hold time later.
    do_in_clock(&h->clk, t_w_374);
    if (h->clk.rising) {
        h->clk.rising = 0;
        queue_add(t_h_374, (qcall) hold_time_callback_374, h, NULL, trace);
    }
}

// This simulates the entire chip.
void be_16374(void *v)
{
    c_374 *f = v;
    int i;

    // There are two units of eight bits each.
    for (i = 0; i < 2; i++)
        be_16374_half(&f->flop[i], f->pre.trace, i);
}
