/*
    c_io.c

    This simulates a non-existent device that can read and/or write
    up to 36 pins as an integer. It's used to provide the simulator
    a test hook into an otherwise incomplete circuit.

    Although this tool is only a test hook and doesn't bear on the
    Dauug|36 architecture itself, it would be a better tool if it
    had a way to specify how many input pins are in use. This is
    because uncertain and undriven inputs are regarded as low, so
    there is no way of observing signal errors.
*/

typedef struct {
    Prefix pre;             // fields common to all components
    in_pin a[36];           // 'a' is displayed and does not go to 'y'
    out_pin y[36];          // 'y' comes from 'y_set' below

    // Two output disable mechanisms work in series. Either connect
    // oe# and set it 'hi', or tweak "off" to set alt_disable.
    // Designed for convenience, with the expectation that you don't
    // want to use both. Note that output is enabled by default.
    in_pin oe_;
    int alt_disable;
    int is_off;             // 'pin_disable' was sent to all outputs

    // This is what 'y' gets set to when output is enabled.
    long y_set;
} c_io;

// Here are names and byte offsets of input pins within the struct.
in_namer in_namer_io[] = {
    "oe#", &((c_io *) NULL)->oe_,
    "a0", &((c_io *) NULL)->a[0],
    "a1", &((c_io *) NULL)->a[1],
    "a2", &((c_io *) NULL)->a[2],
    "a3", &((c_io *) NULL)->a[3],
    "a4", &((c_io *) NULL)->a[4],
    "a5", &((c_io *) NULL)->a[5],
    "a6", &((c_io *) NULL)->a[6],
    "a7", &((c_io *) NULL)->a[7],
    "a8", &((c_io *) NULL)->a[8],
    "a9", &((c_io *) NULL)->a[9],
    "a10", &((c_io *) NULL)->a[10],
    "a11", &((c_io *) NULL)->a[11],
    "a12", &((c_io *) NULL)->a[12],
    "a13", &((c_io *) NULL)->a[13],
    "a14", &((c_io *) NULL)->a[14],
    "a15", &((c_io *) NULL)->a[15],
    "a16", &((c_io *) NULL)->a[16],
    "a17", &((c_io *) NULL)->a[17],
    "a18", &((c_io *) NULL)->a[18],
    "a19", &((c_io *) NULL)->a[19],
    "a20", &((c_io *) NULL)->a[20],
    "a21", &((c_io *) NULL)->a[21],
    "a22", &((c_io *) NULL)->a[22],
    "a23", &((c_io *) NULL)->a[23],
    "a24", &((c_io *) NULL)->a[24],
    "a25", &((c_io *) NULL)->a[25],
    "a26", &((c_io *) NULL)->a[26],
    "a27", &((c_io *) NULL)->a[27],
    "a28", &((c_io *) NULL)->a[28],
    "a29", &((c_io *) NULL)->a[29],
    "a30", &((c_io *) NULL)->a[30],
    "a31", &((c_io *) NULL)->a[31],
    "a32", &((c_io *) NULL)->a[32],
    "a33", &((c_io *) NULL)->a[33],
    "a34", &((c_io *) NULL)->a[34],
    "a35", &((c_io *) NULL)->a[35],
    NULL, NULL,
};

// Here are names and byte offsets of output pins within the struct.
out_namer out_namer_io[] = {
    "y0", &((c_io *) NULL)->y[0],
    "y1", &((c_io *) NULL)->y[1],
    "y2", &((c_io *) NULL)->y[2],
    "y3", &((c_io *) NULL)->y[3],
    "y4", &((c_io *) NULL)->y[4],
    "y5", &((c_io *) NULL)->y[5],
    "y6", &((c_io *) NULL)->y[6],
    "y7", &((c_io *) NULL)->y[7],
    "y8", &((c_io *) NULL)->y[8],
    "y9", &((c_io *) NULL)->y[9],
    "y10", &((c_io *) NULL)->y[10],
    "y11", &((c_io *) NULL)->y[11],
    "y12", &((c_io *) NULL)->y[12],
    "y13", &((c_io *) NULL)->y[13],
    "y14", &((c_io *) NULL)->y[14],
    "y15", &((c_io *) NULL)->y[15],
    "y16", &((c_io *) NULL)->y[16],
    "y17", &((c_io *) NULL)->y[17],
    "y18", &((c_io *) NULL)->y[18],
    "y19", &((c_io *) NULL)->y[19],
    "y20", &((c_io *) NULL)->y[20],
    "y21", &((c_io *) NULL)->y[21],
    "y22", &((c_io *) NULL)->y[22],
    "y23", &((c_io *) NULL)->y[23],
    "y24", &((c_io *) NULL)->y[24],
    "y25", &((c_io *) NULL)->y[25],
    "y26", &((c_io *) NULL)->y[26],
    "y27", &((c_io *) NULL)->y[27],
    "y28", &((c_io *) NULL)->y[28],
    "y29", &((c_io *) NULL)->y[29],
    "y30", &((c_io *) NULL)->y[30],
    "y31", &((c_io *) NULL)->y[31],
    "y32", &((c_io *) NULL)->y[32],
    "y33", &((c_io *) NULL)->y[33],
    "y34", &((c_io *) NULL)->y[34],
    "y35", &((c_io *) NULL)->y[35],
    NULL, NULL,
};

void init_io(UNUSED void *v, UNUSED int parm)
{
    // Output impedance is low by default and requires no attention here.
    // In fact, if we did initialize AND tweak at 0 ps, the two settings
    // might get mixed up in the non-stable priority queue and cause
    // unexpected results.
}

/*
    Input tweaks:

    ?           print out whatever the input is
    ?424        confirms (in color) that the input is 0b110101000 (424 decimal)

    For convenience, floating and uncertain inputs count as low.

    Output tweaks:

    2320        writes 0b100100010000 (2320 decimal) to the output lines
    on          enables output (default condition)
    off         disables output

    The "on" and "off" tweaks enable and disable output by connecting
    the lo or hi rails to the input wiring of oe#. It's a total kludge.
    So if you want to enable and disable output, you can provide your
    own input to oe#, or use the on/off tweaks, but not both.
*/
void tweak_io(void *v, char *parm)
{
    c_io *g = v;
    long want, got;
    int i;

    // Compute input word. Not always used.
    for (got = 0, i = 0; i < 36; i++)
        if (g->a[i].val == hi)
            got |= 1L << i;

    if (!strcmp(parm, "on"))
        g->alt_disable = 0;

    else if (!strcmp(parm, "off"))
        g->alt_disable = 1;

    else if (!strcmp(parm, "?"))
        printf("%s now reads %li\n", g->pre.name, got);

    else if (*parm == '?') {
        want = number(parm+1, "0");
        if (want == got)
            printf("%s - %s reads %li\n", good(), g->pre.name, want);
        else
            printf("%s - %s reads %li instead of %li\n",
                    fail(), g->pre.name, got, want);
        return;
    }

    else {
        g->y_set = number(parm, "0");
        for (i = 0; i < 36; i++)
            pin_output_after_delay(&g->y[i], g->y_set & 1L << i ? hi : lo, 2);
    }
}

// Here is the simulation for this "virtual" part.
void be_io(void *v)
{
    c_io *g = v;
    int i, changed = 0, dis;
    long n;

    // Enable or disable output via a control pin or tweak.
    dis = g->oe_.val == hi || g->alt_disable;
    if (dis != g->is_off) {
        g->is_off = dis;
        output_impedance_after_delay(g->y, dis ? 36 : -36, 100);
    }

    // Check the inputs.
    for (i = 0; i < 36; i++) {
        if (g->a[i].changed) {
            g->a[i].changed = 0;
            changed = 1;
        }
    }

    // Evaluate inputs and display result.
    // For convenience, uncertain/hiz inputs are counted 'lo'.
    if (changed && g->pre.trace) {
        for (n = 0, i = 0; i < 36; i++)
            if (g->a[i].val == hi)
                n |= 1L << i;
        printf("%s is now %li\n", g->pre.name, n);
    }
}
