/*
    c_io.c

    This is the simulation's implementation of a free-running oscillator.
    Unlike real oscillators, this one starts up instantaneously.
    The default frequency is 80 MHz, as that is what we think we'll use
    for the Dauug|36 architecture.

    There are three tweaks available:

    tweak the_osc run       sets oscillator to free-running (default)
    tweak the_osc stop      causes osc. to quit simulation at every clock rise
    tweak the_osc 12345     sets half-period to 12356 (or whatever) ps
*/

typedef struct {
    Prefix pre;             // all components have this
    int val;                // what y should "become"
    int half_ps;            // duration of half-cycle
    int stopped;            // if true, quit simulation every clock rise
    in_pin internal;        // convenient for timing
    out_pin out;            // square-wave output
} c_osc;

enum {
    osc_default_ps = 6250,  // 80 MHz (IMPORTANT - board expects this)
};

// Here is the name and byte offset of the struct's one input pin.
// This "internal" input pin is fed from the output pin via a time
// delay through the event queue, causing the desired oscillation.
in_namer in_namer_osc[] = {
    "internal", &((c_osc *) NULL)->internal,
    NULL, NULL,
};

// Here is the name and byte offset of the struct's one output pin.
out_namer out_namer_osc[] = {
    "out", &((c_osc *) NULL)->out,
    NULL, NULL,
};

// The initialization has an internal connection to make so that
// oscillation can occur.
void init_osc(void *v, int truth)
{
    c_osc *g = v;
    g->half_ps = osc_default_ps;
    connect_pin_to_pin(&g->out, &g->internal, 1);

    // Clock rises at dawn of time.
    g->internal.changed = 1;
}

// This lets us change the half-period of the oscillator.
// 'parm' is a decimal integer string with units of picoseconds.
void tweak_osc(void *v, char *parm)
{
    c_osc *g = v;

    if (!strcmp(parm, "run")) {
        g->stopped = 0;
        return;
    }

    if (!strcmp(parm, "stop")) {
        g->stopped = 1;
        return;
    }

    g->half_ps = number(parm, NULL);
}

// Here is the simulation itself.
void be_osc(void *v)
{
    static int clicks;
    c_osc *g = v;

    if (!g->internal.changed)
        return;
    g->internal.changed = 0;

    // The first time here, g->internal.val == uncert, so g->val will be 
    // 'hi' on the first pass. Thus rises our clock at the dawn of time.
    g->val = g->internal.val == hi ? lo : hi;

    if (g->val != g->out.val)
        pin_output_after_delay(&g->out, g->val, g->half_ps - (now ? 1 : 0));

    if (g->val == hi) {
        if (g->pre.trace)
            printf(YEL "****** click %i ******" BAW "\n", clicks % 4);
        ++clicks;
        if (g->stopped)
            quit_at = now;
    }
}
