/*
    c_reset.c

    This is the simulation's implementation of a voltage monitor that
    provides an active-low system reset signal. A real-world chip for
    this function has not been selected for the architecture yet as of
    11 May 2011.
*/

typedef struct {
    Prefix pre;             // all components have this
    time when_up;           // time when y should go high
    out_pin y;              // output
} c_reset;

enum {                      // 100 ns
    reset_default_ps = 100000,
};

// The struct has no input pins.
in_namer in_namer_reset[] = {
    NULL, NULL,
};

// Here is the name and byte offset of the struct's one output pin.
out_namer out_namer_reset[] = {
    "y", &((c_reset *) NULL)->y,
    NULL, NULL,
};

// When this does nothing, component recalculation will alter reset.
void do_almost_nothing(void *v, void *unused, int i)
{
    c_reset *g = v;

    if (!g->pre.dirty) {
        g->pre.dirty = dirty_comps ? dirty_comps : &g->pre;
        dirty_comps = &g->pre;
    }
}

void go_low_for(c_reset *g, int ps)
{
    if (ps < 1)
        ng("unsupported reset duration %i", ps);
    pin_output_after_delay(&g->y, lo, 1);
    g->when_up = now + ps;
    queue_add(ps, (qcall) do_almost_nothing, g, NULL, 0);

    if (g->pre.trace)
        printf("%s.y going low\n", g->pre.name);
}

// There are no voltages in the simulation to monitor. This initialization
// mimics a power-up outcome by driving the output low and starting the
// timer.
void init_reset(void *v, int truth)
{
    go_low_for(v, reset_default_ps);
}

// A tweak to this component sets its timeout to the number of picoseconds
// indicated in "parm" as a decimal integer, then drives the output low
// and starts the timer.
void tweak_reset(void *v, char *parm)
{
    c_reset *g = v;
    int ps;

    ps = number(parm, NULL);
    go_low_for(g, ps);
}

// Simulate the voltage monitor (system reset chip).
void be_reset(void *v)
{
    c_reset *g = v;

    // not enabled / not time yet?
    if (!g->when_up || now < g->when_up)
        return;

    // bring the pin up 
    pin_output_after_delay(&g->y, hi, 1);

    // disable
    g->when_up = 0;

    if (g->pre.trace)
        printf("%s.y going high\n", g->pre.name);
}
