/*
    connect.c
    This module hooks components together.

    So the implementation hierarchy looks like:

    connect.c
        comps/c_*.c
            pin.c
                compo.h

    This module's linked lists for dirty pins and dirty components
    have unusual pointer semantics; compo.h provides an explanation.

    This module also has some hard-to-read pointer arithmetic;
    see the in_namer typedef comment in compo.h for details.
*/

/*
    Here is a linked list of all components in the simulation.
*/
Prefix *everyComponent = NULL;

/*
    "rail" is a pseudo-component with pseudo-pins named "lo" and "hi".
    The reason it's not an actual component is to provide unlimited
    fanout. To save memory, the non-rail components only have a fanout
    of 'fan_out_max', which is quite small.
*/
const char *rail_name ="rail";

/*
    This table sets up various types of components that the simuator supports,
    names the types, sets initialization parameters, and provides functions to
    manipulate instances. 'Kind' is defined in compo.h.
*/
Kind kinds[] = {

    "sram-sync-code-data",
    sizeof(c_sram),
    23,                         // 2**23 x 36 bits
    init_sram,
    tweak_sram,
    in_namer_sram,
    out_namer_sram,
    be_sram,

    "sram-sync-decoder",
    sizeof(c_sram),
    17,                         // 2**17 x 36 bits
    init_sram,
    tweak_sram,
    in_namer_sram,
    out_namer_sram,
    be_sram,

    "sram-sync-page-table",
    sizeof(c_sram),
    -19,                        // 2**19 x 18 bits
    init_sram,
    tweak_sram,
    in_namer_sram,
    out_namer_sram,
    be_sram,

    "sram-sync-computing",
    sizeof(c_sram),
    -18,                        // 2**18 x 18 bits
    init_sram,
    tweak_sram,
    in_namer_sram,
    out_namer_sram,
    be_sram,

    "16374",
    sizeof(c_374),
    0,
    init_16374,
    tweak_16374,
    in_namer_374,
    out_namer_374,
    be_16374,

    "16244",
    sizeof(c_244),
    0,
    init_16244,
    NULL,
    in_namer_244,
    out_namer_244,
    be_16244,

    "1g74",
    sizeof(c_1g74),
    0,
    init_1g74,
    tweak_1g74,
    in_namer_1g74,
    out_namer_1g74,
    be_1g74,

    "flop",                     // synonym for 1g74
    sizeof(c_1g74),
    0,
    init_1g74,
    tweak_1g74,
    in_namer_1g74,
    out_namer_1g74,
    be_1g74,

    "or",
    sizeof(c_2gxx),
    glue_or,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "and",
    sizeof(c_2gxx),
    glue_and,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "nor",
    sizeof(c_2gxx),
    glue_nor,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "nand",
    sizeof(c_2gxx),
    glue_nand,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "xor",
    sizeof(c_2gxx),
    glue_xor,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "buf",
    sizeof(c_2gxx),
    glue_buf,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "inv",
    sizeof(c_2gxx),
    glue_inv,
    init_2gxx,
    tweak_2gxx,
    in_namer_2gxx,
    out_namer_2gxx,
    be_2gxx,

    "switch",
    sizeof(c_switch),
    0,
    init_switch,
    tweak_switch,
    in_namer_switch,
    out_namer_switch,
    be_switch,

    "io",
    sizeof(c_io),
    0,
    init_io,
    tweak_io,
    in_namer_io,
    out_namer_io,
    be_io,

    "osc",
    sizeof(c_osc),
    0,
    init_osc,
    tweak_osc,
    in_namer_osc,
    out_namer_osc,
    be_osc,

    "80mhz",                    // synonym for osc
    sizeof(c_osc),
    0,
    init_osc,
    tweak_osc,
    in_namer_osc,
    out_namer_osc,
    be_osc,

    "reset",
    sizeof(c_reset),
    0,
    init_reset,
    tweak_reset,
    in_namer_reset,
    out_namer_reset,
    be_reset,

    "pull",
    sizeof(c_pull),
    0,
    init_pull,
    tweak_pull,
    in_namer_pull,
    out_namer_pull,
    be_pull,

    "r10k",                     // synonym for pull
    sizeof(c_pull),
    0,
    init_pull,
    tweak_pull,
    in_namer_pull,
    out_namer_pull,
    be_pull,
};

/*
    Allocate and initialize a new component of indicated 'kind'.
    There is no support for free().
*/
void *new_component(char *name, char *kind)
{
    int i;
    Prefix *c;
    in_namer *inam;
    out_namer *onam;
    in_pin *ip;

    // 'name' may not be "rail", because that's a reserved pseudo-component.
    if (!strcmp(name, rail_name))
        ng("!may not create a component named '%s'", rail_name);

    // lookup characteristics of 'kind'
    for (i = 0; i < sizeof(kinds) / sizeof(*kinds); i++)
        if (!strcmp(kind, kinds[i].name))
            goto ok;
    ng("!unknown kind %s", kind);

ok:
    // allocate and zero memory, copy name, copy kind, chain to list
    c = trackmem(NULL, kinds[i].size);
    c->name = copy(name, -1);
    c->kind = kinds + i;
    c->next = everyComponent;
    everyComponent = c;

    // write names on all input pins
    // also, write a pointer back to the component
    for (inam = c->kind->in_names; inam->name; ++inam) {
        ip = (in_pin *) ((char *) c + ((char *) inam->pin - (char *) NULL));
        ip->name = cat(c->name, ".", inam->name);
        ip->comp = c;
    }

    // write names on all output pins
    for (onam = c->kind->out_names; onam->name; ++onam)
        ((out_pin *) ((char *) c + ((char *) onam->pin - (char *) NULL)))
                ->name = cat(c->name, ".", onam->name);

    // component-specific initialization
    if (c->kind->init)
        c->kind->init(c, c->kind->init_parm);

    return c;
}

// Look up a component by name.
Prefix *find_component(char *name)
{
    Prefix *c;

    for (c = everyComponent; c; c = c->next)
        if (!strcmp(c->name, name))
            return c;
    return NULL;
}

// Request that a component output diagnostic (tracing) information
// when it is in use.
void enable_trace(char *name)
{
    Prefix *c = find_component(name);

    if (!c)
        ng("!unknown component %s cannot be traced", name);
    c->trace = 1;
}

// Whenever the queue has nothing left to do for a given picosecond,
// this routine is called to give the components a chance to reflect
// on what occurred during that picosecond.
void all_components_respond(void)
{
    Prefix *next;

    // All pins with incoming signals will be updated here.
    do_dirty_pins();

    // All components on the dirty list have some business that needs
    // to happen. Each is popped from the list here and serviced.
    if (dirty_comps) {
        while (next = dirty_comps->dirty) {
            dirty_comps->dirty = NULL;
            dirty_comps->kind->be(dirty_comps);
            dirty_comps = next;
        }
    }

    dirty_comps = NULL;
}

// Put every input pin and every component on dirty lists.
// This is an initial condition to bootstrap the simulation.
void every_pin_is_dirty(void)
{
    Prefix *c;
    in_namer *list;
    in_pin *p;

    for (c = everyComponent; c; c = c->next) {

        // The pins are dirty.
        for (list = c->kind->in_names; list->name; ++list) {
            p = (in_pin *) ((char *) c + 
                    ((char *) list->pin - (char *) NULL));
            if (p->dirty)
                continue;       // pin might have multiple names
            p->dirty = dirty_pins ? dirty_pins : p;
            dirty_pins = p;
        }

        // Component c is also dirty.
        ifnot (c->dirty) {
            c->dirty = dirty_comps ? dirty_comps : c;
            dirty_comps = c;
        }
    }
}

// Given a component and name of an input pin, get a pointer to the pin.
in_pin *find_input_pin(Prefix *comp, char *name)
{
    in_namer *list = comp->kind->in_names;

    for (; list->name; ++list)
        if (!strcmp(name, list->name))
            return (in_pin *) ((char *) comp + 
                    ((char *) list->pin - (char *) NULL));
    return NULL;
}

// Given a component and name of an output pin, get a pointer to the pin.
out_pin *find_output_pin(Prefix *comp, char *name)
{
    out_namer *list = comp->kind->out_names;

    for (; list->name; ++list)
        if (!strcmp(name, list->name))
            return (out_pin *) ((char *) comp + 
                    ((char *) list->pin - (char *) NULL));
    return NULL;
}

// Wire pin "cFrom.pFrom" to pin "cTo.pTo" with delay "ps" ps.
void connect_pins_by_name(
    char *cFrom, char *pFrom, char *cTo, char *pTo, int ps
) {
    in_pin *ip;
    out_pin *op;
    Prefix *cf, *ct;

    // Work out the destination pin.
    ifnot (ct = find_component(cTo)) {
        ng("unfound %s", cTo);
        return;
    }

    ifnot (ip = find_input_pin(ct, pTo)) {
        ng("unfound input %s on %s", pTo, cTo);
        return;
    }

    // If 'cFrom' is pseudo-component "rail", it's a special case.
    if (!strcmp(cFrom, rail_name)) {
        if (!strcmp(pFrom, "hi"))
            rail_in_pin(ip, hi);
        else if (!strcmp(pFrom, "lo"))
            rail_in_pin(ip, lo);
        else
            ng("unknown '%s' on pseudo-component '%s'", 
                    pFrom, rail_name);
        return;
    }

    // Back to main case, coming from a normal output pin on a normal part.
    ifnot (cf = find_component(cFrom)) {
        ng("unfound %s", cFrom);
        return;
    }

    ifnot (op = find_output_pin(cf, pFrom)) {
        ng("unfound output %s on %s", pFrom, cFrom);
        return;
    }

    connect_pin_to_pin(op, ip, ps);
}

/*
    The 'check' keyword permits a *.ns script to verify that an output
    pin has a particular state during a regression test. The function
    below tests to see whether the state is or is not as expected and
    reports the outcome. Counters in util.c keep track of how many
    good() and fail() outcomes occurred.
*/
void check_output_pin_state(char *comp, char *outpin, char *want_text)
{
    Prefix *c;
    out_pin *p;
    int want, got;
    char *got_text;

    ifnot (c = find_component(comp)) {
        ng("unfound %s", comp);
        return;
    }

    ifnot (p = find_output_pin(c, outpin)) {
        ng("unfound output %s on %s", outpin, comp);
        return;
    }

    want = pin_text_val(want_text);
    got = p->disabled ? hiz : p->val;

    if (got == want)
        printf("%s - %s.%s is %s\n", good(), comp, outpin, want_text);
    else {
        got_text = pin_val_text(got);
        printf("%s - %s.%s is %s (%i = %iv %id) instead of %s (%i)\n",
                fail(), comp, outpin, got_text, got, p->val,
                p->disabled, want_text, want);
    }
}

/*
    The 'tweak' keyword permits a *.ns script to tell a component to
    do something or change its configuration within a test. Here is
    the implementation. The supplied parameter's interpretation is
    component-specific as provided via the kinds[] table.
*/
void tweak_component(char *comp, char *parm)
{
    Prefix *c;

    ifnot (c = find_component(comp)) {
        ng("unfound %s", comp);
        return;
    }

    if (!c->kind->tweak) {
        ng("'%s' has no tweak interface", comp);
        return;                         // not invalid, but no operation
    }

    c->kind->tweak(c, parm);
}

/*
    This provides a written report of short circuits (input pins being
    driven high and low simultaneously) and possible short circuits
    (input pins being driven from more than one source, with at least
    one source's level is not known). Short circuits are normal, but
    we require a report to help establish that they are brief enough
    and infrequent enough that no damage to components will occur.
*/
void fire_report(void)
{
    static char *fname = "fire.report";
    Prefix *c;
    in_namer *n;
    in_pin *p;
    time elapsed;
    char *comp_name = NULL;
    FILE *f;

    f = fopen(fname, "w");

    // Loop over every pin of every component.
    for (c = everyComponent; c; c = c->next)
        for (n = c->kind->in_names; n->name; ++n) {
            p = (in_pin *) ((char *) c + 
                    ((char *) n->pin - (char *) NULL));

            // If a fire is happening at this pin, 'longest' might
            // need updated in light of this fire.
            if (p->fire.flame) {
                elapsed = now - p->fire.when;
                if (elapsed > p->fire.longest)
                    p->fire.longest = elapsed;
            }

            if (!p->fire.longest)
                continue;

            if (!comp_name)
                fprintf(f, "component // pin, longest (ps), min. sep (ps), "
                        "duty upper bound\n\n");
            if (comp_name != c->name)
                fprintf(f, "%s\n", comp_name = c->name);
            fprintf(f, "  %-5s %12li %12li    %.3f\n",
                    n->name, p->fire.longest, p->fire.sep,
                    (double) p->fire.longest / p->fire.sep);
        }

    if (!comp_name)
        fprintf(f, "No fires.\n");
    else
        ng("see %s for fire information", fname);
    fclose(f);
}

// Free all component heap memory following use.
void discard_netlist(void)
{
    Prefix *c, *next;
    in_namer *inam;
    out_namer *onam;
    in_pin *ip;
    out_pin *op;

    for (c = everyComponent; c; c = next) {
        next = c->next;

        // Free the input pin names.
        for (inam = c->kind->in_names; inam->name; ++inam) {
            ip = (in_pin *) ((char *) c + ((char *) inam->pin - (char *) NULL));
            trackmem(ip->name, 0);
        }

        // Free the output pin names.
        for (onam = c->kind->out_names; onam->name; ++onam) {
            op = (out_pin *) ((char *) c + ((char *) onam->pin - (char *) NULL));
            trackmem(op->name, 0);
        }

        trackmem(c->name, 0);           // Free the component's name.
        trackmem(c, 0);                 // Free the component.
    }

    everyComponent = NULL;
}
