/*
    load.c
    This is the parser for netsim (netlist simulation) scripts.
*/

static void *asm_ctx;                   // assembler context (limit one)

void load(char *fName)
{
    enum { max_bytes = 80, max_args = 5 };

    static char *cmds[] = {
        "k k", "k kind",
        "n n", "n name",
        "? ?", "? trace",
        "f f", "f from",
        "t t", "t to",
        "s s", "s signal",
        "E settle",
        "C check",
        "O note",
        "T tweak",
        "Q quit",
        "L load",
        "a asm",
        "- off",
        "+ on",
        "c c", "c cls",
        NULL
    };

    int n_args;                         // correct number of arguments
    int n_have;                         // number of arguments scanned
    int lineNo = 0;                     // for error reports
    int off = 0;                        // in multi-line comment

    char buf[max_bytes];                // small line buffer
    char *blanks = "\n\t ";             // token delimiters
    char *loading;                      // strtok_r scratch memory

    char *cmd, *arg[max_args];          // command and argument pointers
    int oper;                           // operation specified by command

    // These have to be preserved from line to line. Be sure to init.
    char cKind[max_bytes];              // component kind
    char cFrom[max_bytes];              // from-component name
    char cTo[max_bytes];                // to-component name

    // These are consumed immediately and don't need copied.
    char *pFrom, *pTo;                  // from-pin and to-pin names
    char *note;                         // text of note
    int ps;                             // from-to delay in picoseconds

    FILE *f;                            // stream being loaded

    long *image;                        // code RAM contents
    int n_image;                        // code RAM size

    *cKind = *cFrom = *cTo = '\0';      // sane init

    f = fopen(fName, "r");
    if (!f)
        goto h;

    while (fgets(buf, max_bytes, f)) {
        ++lineNo;

        // delete anything from comment symbol ';' rightward
        buf[strcspn(buf, ";")] = '\0';

        // get command from line; only first char matters
        cmd = strtok_r(buf, blanks, &loading);
        if (!cmd)
            continue;                   // blank line or comment

        // fold command from keyword to a single character
        oper = *wesKeyword(cmd, cmds, "2");

        // multi-line comment?
        if (off && oper != '+')
            continue;

        // There is a free-form "note" command to display test info.
        // We just grab the rest of the input string (without breaking).
        if (oper == 'O') {
            note = strtok_r(NULL, "\n", &loading);
            printf(GRE "=== %s ===" BAW "\n", note);
            continue;
        }

        // Get args after command up to some limit.
        for (n_have = 0; n_have < max_args; ++n_have)
            ifnot (arg[n_have] = strtok_r(NULL, blanks, &loading))
                break;

        // see how many args the command calls for, and verify we have that
        n_args = wesCharSub(oper, "f1t1s3k1n1?1E0C3T2Q1-0+0c0L1a1f0") - '0';
        if (n_args != n_have) {
            ng("%s : %i wrong # of args", fName, lineNo);
            continue;
        }

        switch (oper) {

            default:
                ng("%s : %i unknown command", fName, lineNo);
                continue;

            case 'k':                   // k <component-kind>
                strcpy(cKind, arg[0]);
                continue;

            case 'n':                   // n <new-component-name>
                new_component(arg[0], cKind);
                continue;

            case '?':                   // ? <component-name>
                enable_trace(arg[0]);
                continue;

            case 'f':                   // f <from-component-name>
                strcpy(cFrom, arg[0]);
                continue;

            case 't':                   // t <to-component-name>
                strcpy(cTo, arg[0]);
                continue;

            case 's':                   // s <from-pin> <to-pin> <delay-ps>
                pFrom = arg[0];
                pTo = arg[1];
                if (strlen(arg[2]) != strspn(arg[2], "0123456789")) {
                    ng("%s : %i invalid delay", fName, lineNo);
                    continue;
                }
                ps = atoi(arg[2]);
                if (ps < 1) {
                    ng("%s : %i positive delay required", fName, lineNo);
                    continue;
                }
                connect_pins_by_name(cFrom, pFrom, cTo, pTo, ps);
                continue;

            case 'E':                   // settle
                queue_run();
                continue;

            case 'C':                   // check <comp> <out-pin> <state>
                check_output_pin_state(arg[0], arg[1], arg[2]);
                continue;

            case 'T':                   // tweak
                tweak_component(arg[0], arg[1]);
                continue;

            case 'Q':                   // quit (at ps)
                quit_at = number(arg[0], "l");
                continue;

            case '-':                   // comment out
                off = 1;
                continue;

            case '+':                   // comment in
                off = 0;
                continue;

            case 'c':                   // clear display and scrollback
                printf("\x1b" "c");
                fflush(stdout);
                continue;

            case 'L':                   // 'load' does NOT reinitialize,
                load(arg[0]);           // but is a continuation of
                continue;               // what already is happening.

            case 'a':                   // asm
                // Free whatever was assembled previously (if anything)
                // and assemble a new code RAM from the file specified.
                // Then find the code RAM's memory (you have to have
                // already created the code RAM in the script) and load
                // it with what the assembler did.
                asm_dealloc(asm_ctx);
                asm_ctx = asm_assemble(arg[0], "fbl", "assembling");
                get_firmware("C", NULL, &n_image, (void **) &image);
                asm_fetch_code(asm_ctx, image, n_image);
                continue;
        }
    }

    if (ferror(f) || fclose(f))
        goto h;

    return;
h:
    ng("error '%s' reading '%s'", strerror(errno), fName);
}

// This cleans up memory after use of the assembler.
void discard_assembly(void)
{
    asm_dealloc(asm_ctx);
    asm_ctx = NULL;
}
