/*
    ns.c
    This is the root source file for the netlist simulation (netsim).

    To simplify reading this source code by newcomers and finding
    things by gurus, this netsim is divided into many files according
    to plausible but maybe not optimal concept boundaries.

    Those files are catenated into one translation unit, which speeds
    compilation and eliminates much need for dependency checking. I
    recognize that some maintainers would prefer a more "traditional"
    build scenario. Every approach has its benefits and drawbacks.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>                 // symlink()
#include <locale.h>                 // setlocale()

#include "../consts/opcode.h"       // c_sram.c uses oc_HALT
#include "../asm/asm.h"             // how to access assembler
#include "../misc/misc.h"           // shared with VM and assembler
#include "../firmware/firmware.h"   // get_firmware()
#include "compo.h"                  // building blocks for pins and components

#include "../misc/misc.c"           // shared with VM and assembler
#include "pin.c"                    // simulates pins
#include "util.c"                   // netsim-specific tidbit functions
#include "comps/c_1g74.c"           // part: single D flip-flop
#include "comps/c_244.c"            // part: 16-bit (quad quad) buffer
#include "comps/c_2gxx.c"           // part: dual BUF INV AND OR NAND NOR XOR
#include "comps/c_374.c"            // part: 16-bit (dual octal) D flip-flop
#include "comps/c_io.c"             // part: bus I/O for simulations only
#include "comps/c_osc.c"            // part: 80 MHz crystal oscillator
#include "comps/c_pull.c"           // part: 10k resistor (not used or sim'd)
#include "comps/c_reset.c"          // part: generic voltage monitor
#include "comps/c_sram.c"           // part: synchronous static RAM
#include "comps/c_switch.c"         // part: switch for simulations only
#include "connect.c"                // code to hook components together
#include "queue.c"                  // container for priority queue
#include "quser.c"                  // simulator's interface to priority queue
#include "load.c"                   // read and perform a *.ns script

/*
    The netsim executable 'ns' takes at most one argument, which is
    the name of a script to run. If no script is specified, the
    most recently run script is re-run, as determined by a symlink
    'last-test.ns' that is managed by this program.
*/
int main(int argc, char **argv)
{
    char *fname = "last-test.ns";

    if (argc == 2) {                // keep link to most recently run script
        unlink(fname);
        symlink(argv[1], fname);
        fname = argv[1];
    }

    setlocale(LC_NUMERIC, "");      // printf ' modifier in queue_run()

    queue_setup();                  // create queue for simulation's events
    load(fname);                    // start reading the script
    queue_run();                    // allow queue to finish
    fire_report();                  // write out short circuit durations

    discard_queue();                // free the event queue's memory
    discard_assembly();             // free memory used by assembler
    discard_all_firmware();         // free the firmware's memory
    discard_netlist();              // free the netlist's memory
    trackmem_report();              // account for all heap memory

    // Some scripts test for expected outcomes. Were there failures?
    if (n_fail) {
        printf("%s - ", fail());    // exaggerates n_fail by 1
        printf("%i good and %i failed tests\n", n_good, n_fail - 1);
        return 1;
    }

    // Some scripts test for expected outcomes. Were there *no* failures?
    if (n_good) {
        printf("%s - ", good());    // exaggerates n_good by 1
        printf("all %i tests were successful\n", n_good - 1);
    }

    return 0;
}
