/*
    pin.c
    This code simulates input, output, and clock input pins on components.
*/

// This permits simplified casting of callback function pointers in the queue.
typedef void (*qcall)(void *, void *, int);

// We need this prototype from quser.c.
void queue_add(int ps_delay, qcall callback, void *v, void *w, int i);

// Here is the simulator's elapsed time, measured as whole picoseconds,
// and a provision to end the simulation.
time now;
time quit_at = LONG_MAX;

/*
    This is a list of "dirty" pins that need the do_in_pin() operation
    applied. Pins in this list have non-NULL "dirty" pointers, making
    it easy to see whether a pin needs to be placed on this list (at
    the head, as it's singly linked). The pin at the end of the list,
    which is also the first pin placed on the list, has its "dirty"
    pointer point to itself as an end-of-list marker.
*/
in_pin *dirty_pins;

// Same idea for dirty components.
Prefix *dirty_comps;

/*
    This connects an input pin to a power rail. It's a special function
    because power is a pseudo-component that isn't limited by the
    simulation's fan_out_max limit.
*/
void rail_in_pin(in_pin *p, int rail)
{
    if (p->n_in)
        ng("rail_in_pin on already-connected %s", p->name);
    p->n_in = 1;
    p->in[0] = rail;
    p->zin[0] = 0;
    p->last_change = now;

    if (!p->dirty) {
        p->dirty = dirty_pins ? dirty_pins : p;
        dirty_pins = p;
        if (!p->comp->dirty) {
            p->comp->dirty = dirty_comps ? dirty_comps : p->comp;
            dirty_comps = p->comp;
        }
    }
}

// Translate an input pin's aggregate value to text for diagnostic output.
char *pin_val_text(int val)
{
    switch (val) {
        case hiz:
            return "hi_z";
        case lo:
            return "lo";
        case hi:
            return "hi";
        case lo | hi:
            return "conflict";
        case uncert:
            return "uncert";
        default:
            return "combo";
            ng("(combo was %i)", val);
    }
}

/*
    Make input pin p's value "up to date" with respect to its inputs.
    If not a definite 'hi' or 'lo', this aggregate value will be 'invalid'.
    Also trip a 'changed' flag on the pin. Lastly, record duration of
    potential short-circuits.
*/
static void do_in_pin(in_pin *p)
{
    int i, v = 0;               // loop counter, latest value
    int fire = 0;
    time elapsed;

    if (0) {
        printf("in_pin %s\n", p->name);
        for (i = 0; i < p->n_in; i++)
            printf("  %iv %id\n", p->in[i], p->zin[i]);
    }

    for (i = 0; i < p->n_in; i++)
        if (!p->zin[i])         // High-impedance sources don't count!
            v |= p->in[i];

    if (v != hi && v != lo) {   // unclean, conflicting, floating->uncertain
        if (v & (hi | lo))
            fire = 1;           // potentially or actually conflicting
        v = uncert;
    }

    if (fire != p->fire.flame) {                // short circuit accounting
        elapsed = now - p->fire.when;
        p->fire.when = now;
        if (fire) {                             // new fire
            if (!p->fire.sep)
                p->fire.sep = LONG_MAX;         // very first fire
            if (elapsed < p->fire.sep)          // shortest time with no fire
                p->fire.sep = elapsed ? elapsed : 1;
        }
        else if (elapsed > p->fire.longest)
            p->fire.longest = elapsed;          // longest duration of fire
        p->fire.flame = fire;
    }

    // Set flag so user can notice changes. User clears flag.
    // At 0 ps, we also set the flag so user can get initial value.
    if (p->val != v || !now) {
        p->val = v;
        p->changed = 1;
        p->last_change = now;
    }

    if (0)
        printf("in_pin %s aggregate value %s changed %i\n",
                p->name, pin_val_text(v), p->changed);
}

/*
    Pop each dirty pin from the list, and call do_in_pin on it.
    Note that end of list is marked using a self-reference instead of NULL.
*/
void do_dirty_pins(void)
{
    in_pin *next;

    if (dirty_pins)
        while (next = dirty_pins->dirty) {
            dirty_pins->dirty = NULL;
            do_in_pin(dirty_pins);
            dirty_pins = next;
        }

    dirty_pins = NULL;
}

/*
    Make input clock pin's value "up to date" with respect to its inputs.
    Its 'state' will be 'lo' or 'hi', and 'rising' will trip if we rise.
    Up to the user to clear 'rising'.

    Because the pulse width of the clock signal is checked against
    datasheet limits, this function issues a lot of warnings when the
    simulation is first powered up and everything is settling.
*/
void do_in_clock(in_clock *ic, int min_width)
{
    if (!ic->pin.changed)
        return;
    ic->pin.changed = 0;

    switch (ic->pin.val) {
        case lo:
            ic->last_fall = now;
            if (ic->last_rise + min_width > now)
                ng("clock fall too early for %s", ic->pin.name);
            break;
        case hi:
            ic->rising = 1;
            ic->last_rise = now;
            if (ic->last_fall + min_width > now)
                ng("clock rise too early for %s", ic->pin.name);
            break;
        default:
            ng("unclean clock %i on %s", ic->pin.val, ic->pin.name);
    }
}

/*
    This is a callback function from the event queue to change a
    source's contribution to an input pin's value. It's capable of
    updating either the source's logic value or the source's impedance.
*/
static void incoming_pin_callback(char *in, in_pin *p, int byte)
{
    *in = byte;

    if (!in || !p)
        ng("!sanity incoming_pin_callback");

    // Put 'p' on the 'dirty_pins' list if not already.
    // Note end-of-list is indicated by a self-reference.
    if (!p->dirty) {
        p->dirty = dirty_pins ? dirty_pins : p;
        dirty_pins = p;
        if (!p->comp->dirty) {
            p->comp->dirty = dirty_comps ? dirty_comps : p->comp;
            dirty_comps = p->comp;
        }
    }
}

/*
    This is a callback function from the event queue to set the
    impedance on an array of output pins, such as the I/O pins
    on an SRAM.
*/
void impedance_callback(out_pin *pins, void *unused, int n_pins)
{
    int i, j;
    int z = n_pins >= 0;        // 1 high, 0 low

    if (n_pins < 0)
        n_pins = -n_pins;       // no longer negative

    for (i = 0; i < n_pins; i++)
        if (pins[i].disabled != z) {
            pins[i].disabled = z;
            for (j = 0; j < pins[i].n_out; j++)
                queue_add(pins[i].path_ps[j], (qcall) incoming_pin_callback,
                    pins[i].zout[j], pins[i].in[j], z);
        }
}

/*
    This is a callback function from the event queue to set an output
    pin's logic value to 'hi', 'lo', or 'uncert'.

    To speed up the simulation, the last value set on an output pin
    is remembered for the purpose of removing duplicate events.

    This function works irrespective of output impedance. You can set
    a value while the pin is hi-Z, and the value will appear once the
    pin is lo-Z. See also impedance_callback().

    A danger of this function is that the event queue isn't stable,
    so multiple changes to an output pin during the same picosecond
    will not have reproducible results. We've tried to avoid this.
*/
static void pin_outgoing_callback(out_pin *p, void *unused, int val)
{
    int i;

    p->val = val;
    for (i = 0; i < p->n_out; i++) {
        queue_add(p->path_ps[i], (qcall) incoming_pin_callback,
                p->out[i], p->in[i], val);
    }
}

/*
    This adds an entry to the event queue to set an output pin's value
    to 'hi', 'lo', or 'uncert' after a specified gate delay. After the
    delay time has passed, pin_outgoing_callback() will update the
    output pin as requested here.
*/
void pin_output_after_delay(out_pin *p, int val, int ps)
{
    queue_add(ps, (qcall) pin_outgoing_callback, p, NULL, val);
}

/*
    This adds an entry to the event queue to set the impedance of
    an array of output pins after a delay. After the delay time has
    passed, impedance_callback() will make the impedance adjustments.

    The 'n_pins' argument encodes both the array size (via magnitude)
    and the impedance (positive is high, negative is low).
*/
void output_impedance_after_delay(out_pin *p, int n_pins, int ps)
{
    queue_add(ps, (qcall) impedance_callback, p, NULL, n_pins);
}

/*
    This is called to connect an output pin to an input pin.

    It can also be called to create a "bidirectional" pin such as
    an SRAM I/O pin by connecting an out_pin and in_pin together
    within the same component.
*/
void connect_pin_to_pin(out_pin *from, in_pin *to, int ps)
{
    int f;

    // If path is already connected, we don't want a second one,
    // and let the programmer know.
    for (f = 0; f < from->n_out; ++f)
        if (from->out[f] >= to->in && from->out[f] < to->in + fan_in_max) {
            ng("pins already connected");
            return;
        }

    // If there isn't room at the source or destination, that's an error.
    if (from->n_out >= fan_out_max) {
        ng("fan-out too high on %s", from->name);
        return;
    }
    if (to->n_in >= fan_in_max) {
        ng("fan-in too high on %s", to->name);
        return;
    }

    // Make the connection.
    from->zout[from->n_out] = to->zin + to->n_in;
    from->out[from->n_out] = to->in + to->n_in++;
    from->in[from->n_out] = to;
    from->path_ps[from->n_out++] = ps;
}

/*
    This converts a text representation of an output pin's condition
    to its one-byte representation. It's used by the 'check' keyword
    of a *.ns script.
*/
int pin_text_val(char *t)
{
    if (!strcmp(t, "hi"))
        return hi;
    else if (!strcmp(t, "lo"))
        return lo;
    else if (!strcmp(t, "hi_z"))
        return hiz;
    else if (!strcmp(t, "uncert"))
        return uncert;
    ng("unknown pin value %s", t);
    return uncert;
}

/*
    This is a diagnostic routine for printing the condition of
    an array of input pins as their integer representation.
*/
void trace_input_pins(in_pin *p, int n)
{
    long w = 0;
    int i;

    for (i = n - 1; i >= 0; i--) {
        if (p[i].val == lo)
            w *= 2;
        else if (p[i].val == hi)
            w = 2 * w + 1;
        else {
            printf("%s (%i-bit) uncertain\n", p->name, n);
            return;
        }
    }

    printf("%s (%i-bit) value %li\n", p->name, n, w);
}
