/*
    queue.c
	This is a heap-based generic priority queue, and is adapted from
    Marc Wesley Abel's "weslib" library.

	TO CREATE THE QUEUE

		wesQueue q;
		YourQueueEntryType et;
		wesQueueInit(&q, sizeof(et), YourQueueEntryType_LessThanFn);

	TO PUT SOMETHING IN THE QUEUE

		et = some_item_for_the_queue;
		wesQueueInsert(&q, &et);	// NOTE: et is COPIED into the queue

	TO EXAMINE THE FIRST ELEMENT

		YourQueueEntryType *e;
		if (q.n) {					// WARNING: must ensure q.n is nonzero
			e = q.first;			// WARNING: wesQueueInsert() moves q.first
			do_something_with(e);
		}
		else
			queue_is_empty();
		
	TO REMOVE THE FIRST ELEMENT*

		wesQueueRemoveHead(&q);

	TO REMOVE ALL ELEMENTS*

		q->n = 0;

	TO REMOVE ENTIRE QUEUE*

		wesFree(q->first);
		if (appropriate)
			wesFree(q)

	*provided that no special element deallocation is necessary
*/

typedef struct {
    int bytesPer;       // bytes per element
    int (*less)         // user-provided "less than" function for elements
        (void *, void *);
    int n;              // number of elements in queue
    int arraySize;      // maximum number of elements before expansion
    void *first;        // first element in queue (the array itself)
} wesQueue;

void wesQueueInit(wesQueue *q, int bytesPer, int (*less)(void *, void *))
{
	memset(q, 0, sizeof(*q));
	q->bytesPer = bytesPer;
    q->arraySize = 128;
	q->less = less;
	q->first = trackmem(NULL, q->arraySize * q->bytesPer);
}

void wesQueueInsert(wesQueue *q, void *element)
{
	int i, p;
	uchr *f;

	// ensure the array is big enough
	if (q->n >= q->arraySize) {
		q->arraySize *= 2;
		q->first = trackmem(q->first, q->arraySize * q->bytesPer);
	}

	// place new element at end of array
	f = q->first;				// q->first, indexable by bytes
	i = q->n++;				    // index of insertion point
	memcpy(f + i * q->bytesPer, element, q->bytesPer);

	// stably restore the heap condition if now violated
	while (i) {
		p = (i-1)/2;
		ifnot (q->less(f + i * q->bytesPer, f + p * q->bytesPer))
			return;				// heap condition is fine
		// need to swap i with its parent
		memcpy(f + i * q->bytesPer, f + p * q->bytesPer, q->bytesPer);
		i = p;
		memcpy(f + i * q->bytesPer, element, q->bytesPer);
	}
}

void wesQueueRemoveHead(wesQueue *q)
{
	int i = 0, c;				// present Index and least Child
	uchr *f = q->first;		    // q->first, indexable by bytes

	if (!q->n--)				// replace top with last; shorten array
		ng("!wesErrorQueueEmpty");
	memcpy(f, f + q->n * q->bytesPer, q->bytesPer);

	// stably restore the heap condition if now violated
	while ((c = 2*i + 1) < q->n) {
		if (c+1 < q->n &&		// left or right child smaller?
				q->less(f + (c+1) * q->bytesPer, f + c * q->bytesPer))
			++c;				// right child [else left child]
		if (q->less(f + c * q->bytesPer, f + i * q->bytesPer)) {
			// this child must come before this parent; swap them
			uchr tmp[q->bytesPer];
			memcpy(tmp, f + i * q->bytesPer, q->bytesPer);
			memcpy(f + i * q->bytesPer, f + c * q->bytesPer, q->bytesPer);
			memcpy(f + c * q->bytesPer, tmp, q->bytesPer);
			i = c;
		}
		else break;				// parent-child relationship already healthy
	}
}

#if 0

/*
    This is the original test code for wesQueue*, which was run 256 times.
    It's only here for reference purposes, as it has minor (and easily
    worked around) dependencies on other sections of weslib.
*/

enum { len = 5, n = 65536, total = len * n };

int ltTest(void *x, void *y)
{
	return memcmp(x, y, len) < 0;
}

int sorter(const void *x, const void *y)
{
	return memcmp(x, y, len);
}

void qTest(void)
{
	int i;
	uchr *q = wesMalloc(total), *a = wesMalloc(total);
	wesQueue Q;

	urandom(q, total);			// q is unsorted gibberish
	memcpy(a, q, total);
	qsort(a, n, len, sorter);	// a is sorted gibberish

	wesQueueInit(&Q, len, ltTest);

	for (i=0; i<n; i++)
		wesQueueInsert(&Q, q + i * len);

	for (i=0; i<n; i++) {
		memcpy(q + i * len, Q . first, len);
		wesQueueRemoveHead(&Q);
	}

	printf("memcmp returns (should be zero):  %i\n", memcmp(q, a, total));
	printf("left in queue (should be zero):  %i\n", Q . n);
}

#endif
