/*
    quser.c

    This provides a time-based priority queue for the netlist simulation
    with picosecond granularity. Everything transpires using a single
    queue, with callbacks and arguments determining what happens as
    entries expire out of the queue. The underlying container is in
    queue.c, which was adapted from weslib.
*/

wesQueue Q;                 // the queue itself

// Here is a single priority queue entry for the time element of simulation.
// See also queue.c's documentation.
typedef struct q_ent {
    time when;              // when item comes out of queue
    qcall callback;         // function to call when we come out
    void *v;                // pointer data for called function
    void *w;                // more pointer data for called function
    int i;                  // integer data for called function
} q_ent;

// This function determines time order within the priority queue.
// See also queue.c's documentation.
static int q_ent_less_than(void *vl, void *vr)
{
    q_ent *l = vl, *r = vr;

    return l->when < r->when;
}

// Initialize the simulation queue.
void queue_setup(void)
{
    wesQueueInit(&Q, sizeof(q_ent), q_ent_less_than);
}

// Free the simulation queue.
void discard_queue(void)
{
    trackmem(Q.first, 0);
    memset(&Q, 0, sizeof(Q));
}

// Schedule something to happen some (simulated) picoseconds from now.
void queue_add(int ps_delay, qcall callback, void *v, void *w, int i)
{
    q_ent e;

    if (ps_delay < 1)
        ng("!queue request for non-future action at %i ps", ps_delay);

    e.when = now + ps_delay;
    e.callback = callback;
    e.v = v;
    e.w = w;
    e.i = i;
    wesQueueInsert(&Q, &e);
}

/*
    Run the queue by popping events out in their time sequence and
    simulating them. Stop when the queue runs out of events, because
    nothing will ever happen again. Also stop when the user's time
    limit is reached, as specified with the 'quit' keyword in the
    *.ns script.
*/
void queue_run(void)
{
    q_ent *e;

    every_pin_is_dirty();

    while (1) {                         // loop over timesteps
        // The escape sequence is cursor up one row.
        printf("%80s%'li ps\n\x1b[1A", "", now);

        all_components_respond();

        if (!Q.n) {
            printf("simulation queue empty\n");
            break;
        }

        // Stop here if user requested early exit.
        if (now >= quit_at) {
            ng("quit at %li ps", now);
            quit_at = LONG_MAX;         // reset
            return;
        }

        // Advance to NEXT timestep.
        e = Q.first;
        if (e->when < now)              // XXX remove sanity check someday
            ng("!time is not monotonic");
        if (e->when == now)             // XXX remove sanity check someday
            ng("!time did not advance");
        now = e->when;

        while (Q.n) {                   // loop over events within timestep

            e = Q.first;

            // stop after this timestep is finished
            if (e->when != now)
                break;

            // do whatever the first element says
            e->callback(e->v, e->w, e->i);

            // remove the first element (there should be more)
            wesQueueRemoveHead(&Q);
        }
    }
}
