/*
    util.c
    These minor tidbit functions are used only by netsim.
*/

char *watching = "";                    // not watching anything yet

static int n_good, n_fail;              // regression test counters

/*
    Works a little like "assq" in Scheme.  The list is an alternating
    series of "from" and "to" bytes. Zeros may appear anywhere.

    IF you care about catching overruns when 'from' does not appear in the
    list, terminate the list using whatever you used first for 'from', and
    put the default value in the 'to'.

    Example:  wesCharSub('b', "a#b&c)a?") returns '&' and would have had
    default value '?'.
*/
int wesCharSub(int from, const char *associationList)
{
    const char *s = associationList;
    int stopAt = *s;

    while (1) {
        if (from == *s || *s == stopAt && s != associationList)
            return s[1];
        s += 2;
    }
}

/*
	Find 'word' in 'list', which must end with NULL.  Example:

		char **list = { "@ amp", "# hash", "^ carat", NULL };

		switch (*wesKeyword("HASH", list, "2i")) {
			case '\0':		// word is not in list
			case '@':		// word is "amp"
			case '#':		// word is "hash"
			case '^':		// word is "carat"
			default:		// word is something you added to list
		}

	'options' is a string of character flags, may be:

	'i'  make comparison case-insensitive
	'n'  return NULL instead of "" if there is no match
	'0'
	'1'
	...  any digit (default 1) indicates byte offset of search string;
	     caller may use prior bytes to specify behavior when matched
*/
char *wesKeyword(char *word, char **list, char *options)
{
	int prefixBytes = 1;	// bytes to skip in each list item
	int null = 0;			// return NULL instead of "" if not found?
	int (*cmp)				// comparison function
		(const char *, const char *) = strcmp;

	for (; *options; ++options) {
		if (isdigit(*options))
			prefixBytes = *options - '0';
		else if (*options == 'i')
			cmp = strcasecmp;
		else if (*options == 'n')
			null = 1;
	}

	for (; *list; ++list)
		if (!cmp(word, prefixBytes + *list))
			return *list;

	return null ? NULL : "";
}

/*
    Allocate and copy some memory.
    Pass -1 to get length from strlen().
*/
void *copy(void *mem, int len)
{
    void *buf;

    if (len < 0)
        len = 1 + strlen(mem);

    buf = trackmem(NULL, len);
    memcpy(buf, mem, len);
    return buf;
}

/*
    Catenate and copy three strings. You may not free this.
*/
char *cat(char *s, char *t, char *u)
{
    int len;
    char *buf;

    if (!s) s = "";
    if (!t) t = "";
    if (!u) u = "";

    len = strlen(s) + strlen(t) + strlen(u) + 1;
    buf = trackmem(NULL, len);

    strcpy(buf, s);
    strcat(buf, t);
    strcat(buf, u);

    return buf;
}

/*
    Watch a byte to see when it changes.
    It's like a memory breakpoint, for callback debugging.
*/
void watch(char *alarm)
{
    static int old;

    if (*watching != old) {
        old = *watching;
        ng("%li ps: %s", now, alarm);
    }
}

/*
    This is a range-checking convenience function for parsing a string
    that represents an integer. The string must be \0 terminated.
    
    If 'parms' is empty, the number must be positive and fit in an int.
    If 'parms' contains '-', the number may be negative.
    If 'parms' contains '0', the number may be zero.
    If 'parms' contains 'l', the number may fit in a long.

    This does not detect an out-of-range long, where strtol()
    sets 'errno' to ERANGE. Providing such a test is not a priority.
*/
long number(char *s, char *parms)
{
    long l;
    char *endp;

    if (!parms)
        parms = "";

    l = strtol(s, &endp, 10);
    if (endp == s || *endp)
        ng("!invalid number '%s'", s);

    if (!strchr(parms, '-') && l < 0)
        ng("!negative number not allowed '%s'", s);

    if (!strchr(parms, '0') && l < 0)
        ng("!zero number not allowed '%s'", s);

    if (!strchr(parms, 'l') && l != (int) l)
        ng("!number does not fit in int '%s'", s);

    return l;
}

/*
    This is a constant "OK" string within a green ANSI terminal
    control sequence, with a side effect of incrementing a counter.
    It's a convenience function for diagnostic printfs.
*/
char *good(void)
{
    ++n_good;
    return GRE "OK" BAW;
}

/*
    This is a constant "NG" string within a red ANSI terminal
    control sequence, with a side effect of incrementing a counter.
    It's a convenience function for diagnostic printfs.
*/
char *fail(void)
{
    ++n_fail;
    return RED "NG" BAW;
}
