/*
    stats.c
    Statistical tests for MIX and UNMIX

    TODO Documentation should explain this module.
*/

static long lfsr72_hi, lfsr72_lo, lfsr36_r;     /* PRNG state */

/* Randomize a 36-bit counter, but never to zero. */
static void preset(long *l)
{
    if (!l)
        return;
    do {
        urandom(sizeof *l, l);
        *l &= mask36;
    } while (!*l);
}

/*
 *  For a random word and random key, try flipping each of the 72 bits
 *  one at a time, and see how many output bits flip. The parameters are
 *  added to, so they need to be initialized.
 *
 *  Some tests on 4 cores (these do NOT overlap) 2 May 2020:
 *
 *  tests 3600000 rounds 1 word 15.375993 key 16.474688
 *  tests 3700000 rounds 1 word 15.375296 key 16.475027
 *  tests 3500000 rounds 1 word 15.375396 key 16.475030
 *  tests 3500000 rounds 1 word 15.376074 key 16.474962
 *
 *  tests 126200000 rounds 2 word 17.999696 key 17.999772
 *  tests  90800000 rounds 2 word 17.999606 key 17.999775
 *  tests  90500000 rounds 2 word 17.999684 key 17.999819
 *  tests  90500000 rounds 2 word 17.999684 key 17.999748
 *
 *  tests 56200000 rounds 3 word 17.999876 key 17.999900
 *  tests 55000000 rounds 3 word 17.999632 key 17.999934
 *  tests 55200000 rounds 3 word 17.999733 key 17.999886
 *  tests 54600000 rounds 3 word 17.999695 key 17.999816
 *
 *  3-round results suggest that holding the key constant is bad.
 */
static void one_pass(int rounds, long *word_flips, long *key_flips)
{
    long l, r, ll, rr, mix1, flipper;
    int i;

    preset(&l);
    preset(&r);

    mix1 = alu(oc_MIX, l, r, NULL);
    for (flipper = pole; flipper; flipper >>= 1) {
        ll = l ^ flipper;
        rr = r ^ flipper;

        /* hold key constant, do so many rounds from one flipped input bit */
        for (i=0; i<rounds; i++)
            ll = alu(oc_MIX, ll, r, NULL);
        *word_flips += __builtin_popcountl(mix1 ^ ll);

        /* don't change input, do so many rounds from one flipped key bit */
        ll = l;
        for (i=0; i<rounds; i++)
            ll = alu(oc_MIX, ll, rr, NULL);

        *key_flips += __builtin_popcountl(mix1 ^ ll);
    }
}

static void mix_stats(void)
{
    long tests = 0, word_flips = 0, key_flips = 0;
    enum { rounds = 1, epochs = 100000 };

    while (1) {
        one_pass(rounds, &word_flips, &key_flips);
        ++tests;

        ifnot (tests % epochs)
            printf("tests %li rounds %i word %.6f key %.6f\n", tests, rounds,
                    (double) word_flips / (36 * tests),
                    (double) key_flips / (36 * tests));
    }
}

/*
 *  This uses MIX as a one-opcode PRNG, using all 36 bits to produce
 *  a stream of bytes on stdout. It is NOT a suitable PRNG and does
 *  NOT pass any good statistical test package like dieharder. The
 *  purpose of this function is to demonstrate that this is bad.
 *  These bytes can be piped straight into dieharder, like:

 *  $ ./vm -s c | dieharder -g 200 -d [TEST]
 *  $ ./vm -s f | dieharder -g 200 -d [TEST]
 * 
 *  'mode' is 'c' for a counter, or 'f' for feedback. Both work poorly.
 */
static void write_rands(int mode, int rounds)
{
    long key = 40777292090;
    long r = 0, buf = 0, count = 0;
    int i, buf_bits = 0;

    ng("Friend, this is not a suitable PRNG.\n");

    while (1) {

        /* Mix requested number of times */
        if (mode == 'c')
            r = count++;            /* XXX not enough birthdays */
        else if (mode == 'f')
            ;                       /* XXX may have short period */
        else
            ng("!write_rands");
        for (i=0; i<rounds; i++)
            r = alu(oc_MIX, r, key, NULL);

        /* Stuf in the buf */
        buf |= r << buf_bits;
        buf_bits += 36;

        /* Enuf in the buf */
        while (buf_bits >= 8) {
            putchar(buf & 0xff);
            buf >>= 8;
            buf_bits -= 8;
        }
    }
}

/*
 *  36-bit LFSR. The taps chosen are my own magical innovation; see above.
 *  Note this is not a PRNG, but a stepping function into a MIX key.
 */
static long lfsr36(void)
{
    int i = lfsr36_r & 1;

    lfsr36_r >>= 1;
    if (i)
        lfsr36_r ^= lfsr_taps;

    return lfsr36_r;
}

/*
 *  72-bit LFSR. We only want or return the 36 low bits.
 *  Note this is not a PRNG, but a stepping function into a MIX key.
 */
static long lfsr72(void)
{
    int i = lfsr72_lo & 1;

    lfsr72_lo >>= 1;
    if (lfsr72_hi & 1)
        lfsr72_lo |= 1L << 35;

    lfsr72_hi >>= 1;
    if (i)
        lfsr72_hi ^= lfsr_taps;

    return lfsr72_lo;
}

/*
 *  Feed this into 'dieharder -g 200 -d [TEST]'.
 *  Writes random bytes to stdout, using all 36 bits.
 *
 *  This PRNG loops its output through a single MIX, using a
 *  36-bit LFSR as the key. So it's two machine instructions,
 *  and has a minimum period of 2**36.
 *
 *  Call with lfsr36 or lfsr72
 */
static void test36(long (*prng)(void))
{
    long r = 0, buf = 0;        /* seed */
    int buf_bits = 0;

    preset(&r);

    while (1) {

        r = alu(oc_MIX, r, prng(), NULL);

        /* Stuf in the buf */
        buf |= r << buf_bits;
        buf_bits += 36;

        /* Enuf in the buf */
        while (buf_bits >= 8) {
            putchar(buf & 0xff);
            buf >>= 8;
            buf_bits -= 8;
        }
    }
}

/* Comparison function for sorting in birthdays(). */
static int longcmp(const void *l, const void *r)
{
    const long *ll = l, *rr = r;
    if (*ll < *rr)
        return -1;
    if (*ll > *rr)
        return 1;
    return 0;
}

/*
 *  Make sure we're getting repeated words in our stream.
 */
static void birthdays(void)
{
    enum { samps = 1 << 18 };

    long *set, r = 0;
    int n, equal;

    set = trackmem(NULL, samps * sizeof(*set));

    preset(&r);

    while (1) {
        /* lots of samples */
        for (n=0; n<samps; ++n)
            set[n] = r = alu(oc_MIX, r, lfsr36(), NULL);

        /* sort them */
        qsort(set, samps, sizeof(*set), longcmp);

        /* count dups */
        for (equal = 0, n=1; n<samps; ++n)
            if (set[n-1] == set[n])
                ++equal;

        printf("dups this pass %i\n", equal);
    }

    trackmem(set, 0);
}

/*
 *  Check period of lfsr_36. Compares single words.
 */
static void check_period_36(void)
{
    long first, period;
    int i;

    printf("testing lfsr poly: x** {0");
    for (i=0; i<36; i++) {
        if (1L << i & lfsr_taps)
            printf(" %i", i+1);
    }
    first = lfsr36();
    printf("} from starting point: %li\n", first);

    for (period=1; ; ++period) {
        ifnot (period % 100000000)
            fprintf(stderr, ".");
        ifnot (period % 1000000000)
            printf(" %li\n", period);
        if (first == lfsr36())
            printf("\nperiod recycle at %li\n", period);
    }
}

/*
 *  Check period of lfsr_72. Compares two words.
 */
static void check_period_72(void)
{
    long hi0, lo0, period;
    int i;

    printf("testing lfsr poly: x** {0");
    for (i=0; i<36; i++) {
        if (1L << i & lfsr_taps)
            printf(" %i", i+1+36);
    }
    printf("}\n");

    lfsr72();
    lo0 = lfsr72_lo;
    hi0 = lfsr72_hi;

    for (period=1; ; ++period) {
        ifnot (period % 100000000)
            fprintf(stderr, ".");
        ifnot (period % 1000000000)
            printf(" %li\n", period);
        lfsr72();
        if (lfsr72_lo == lo0 && lfsr72_hi == hi0)
            printf("\nperiod recycle at %li\n", period);
    }
}

/* This module XXX REQUIRES XXX this initialization. */
void setup_stats(void)
{
    preset(&lfsr36_r);
    preset(&lfsr72_lo);
    preset(&lfsr72_hi);
}

/* Test hook with argv[1] */
void some_stats(char *what)
{
    int len = strlen(what);
    long huh;

    switch (*what) {
        case 'b':
            birthdays();
            break;
        case 'm':
            mix_stats();
            break;
        case 'c':       /* obsolete */
        case 'f':       /* obsolete */
            write_rands(*what, len);
            break;
        case 't':
            test36(lfsr36);
            break;
        case 'u':
            test36(lfsr72);
            break;
        case 'p':
            test_perms();   // elsewhere
            break;
        case 'Q':
            // We check this against the VM's stacked unary.
            lfsr36_r = 0745632631626L;
            huh = lfsr36();
            if (huh == 0362715314713L)
                puts("OK  lfsr36(0745632631626) = 0362715314713");
            else
                printf("NG  lfsr36(0745632631626) = 0%lo", huh);
            break;
        case 'R':
            // Check against VM stacked unaries.
            lfsr72_hi = 0736461314331L;
            lfsr72_lo = 0330445777424L;
            lfsr72();
            printf("lfsr72_lo 0%lo %li\n", lfsr72_lo, lfsr72_lo);
            printf("lfsr72_hi 0%lo %li\n", lfsr72_hi, lfsr72_hi);
            break;
        case '0':
            check_period_36();
            break;
        case '1':
            check_period_72();
            break;

        default:
            printf("I don't know how to %s.\n", what);
    }
}
