/*
    vm.c
    Command line utility for testing the assembler, firmware, and software
*/

#include <unistd.h>         // nice()
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#include "../consts/opcode.h"
#include "../misc/misc.h"
#include "../asm/asm.h"
#include "../firmware/firmware.h"

#include "../misc/misc.c"
#include "tests.c"
#include "stats.c"

int main(int argc, char *argv[])
{
    char *e, *yes_stats = NULL;
    long l, r, t;
    int opt, opcode, flags = 0;
    int yes_swiz = 0, yes_regress = 0, yes_help = 0, yes_flip = 0;

    nice(9999);
    if (sizeof(int) < 4 || sizeof(long) < 8)
        ng("!unanticipated architecture");

    setup_stats();
    alu_setup();

    while ((opt = getopt(argc, argv, "NRTZa:cdfhrs:w")) != -1)
        switch (opt) {

            // set flags going into calculator mode
            case 'N': flags |= bit0; break;
            case 'R': flags |= bit1; break;
            case 'T': flags |= bit2; break;
            case 'Z': flags |= bit3; break;

            // assembler
            case 'a': asm_assemble(optarg, "flrcd", NULL); break;

            // clear screen at startup
            case 'c': fprintf(stderr, "\x1b""c"); break;

            // more verbose diagnostics
            case 'd': alu_trace = 2; break;

            // swap left and right operands (more convenient manual tests)
            case 'f': yes_flip = 1; break;

            // swizzle right operand
            case 'w': yes_swiz = 1; break;

            // call into stats module; takes a character to specify test
            case 's': yes_stats = optarg; break;

            // regression tests
            case 'r': yes_regress = 1; break;

            // help
            default: yes_help = 1; break;
        }

    if (yes_help) {
        printf("%s", 
"Usage: vm [OPTION]... [EXPRESSION]\n\n"
"If EXPRESSION is present, like \"34 a.sss -12345\", calculates it.\n\n"
"Other notable uses and options:\n\n"
"  -a FILE   run assembler on FILE\n"
"  -c        clear screen at startup\n"
"  -d        show more diagnostics\n"
"  -f        flip left and right operands to EXPRESSION\n"
"  -h        display this help and exit\n"
"  -r        run regression tests\n"
"  -s OPT    call experimental stats module with option OPT\n"
"  -w        swizzle right arg of EXPRESSION across tribbles\n"
"  -N        set N(egate) flag to evaluate EXPRESSION\n"
"  -R        set R(ange) flag to evaluate EXPRESSION\n"
"  -T        set T(emporal) flag to evaluate EXPRESSION\n"
"  -Z        set Z(ero) flag to evaluate EXPRESSION\n"
);
        return 0;
    }

    if (yes_regress)
        alu_regression_test_suite();

    // calculator mode takes three non-flag arguments: L, opcode, R
    if (optind + 3 == argc) {

        opcode = find_opcode(argv[optind+1]);
        if (opcode < 0)
            ng("!invalid opcode");

        l = strtol(argv[optind], &e, 0);
        ifnot ((l & comp36) == 0 || (l | mask36) == -1)
            ng("!left argument out of range");
        if (*e)
            ng("!left argument bad syntax");

        r = strtol(argv[optind+2], &e, 0);
        ifnot ((r & comp36) == 0 || (r | mask36) == -1)
            ng("!right argument out of range");
        if (*e)
            ng("!right argument bad syntax");

        if (yes_flip)
            t = l, l = r, r = t;

        if (yes_swiz) {
            printf("swizzling r\n");
            r *= 010101010101L;
        }

        l &= mask36;                    // clamp -ve sign at bit 35
        r &= mask36;

        if (!alu_trace)
            alu_trace = 1;
        alu(opcode, l, r, &flags);
    }
    else if (optind != argc)
        ng("!extraneous stuff on command line");

    if (yes_stats)
        some_stats(yes_stats);

    discard_all_firmware();
    trackmem_report();

    return 0;
}
